/****************************************************************
 Matrix.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
****************************************************************/
#include <iostream>
#include <fstream>
#include "Matrix.H"
#include "Vector.H"
#include "BOOM/String.H"
#include "BOOM/VectorSorter.H"
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_cblas.h>
#include "../TNT/jama_eig.h"
using namespace std;
using namespace GSL;


Matrix::Matrix(int rows,int cols)
{
  M=gsl_matrix_alloc(rows,cols);
}



Matrix::Matrix(const GSL::Matrix &other)
{
  M=gsl_matrix_alloc(other.getNumRows(),other.getNumColumns());
  gsl_matrix_memcpy(M,other.M);
}



Matrix::Matrix(const TNT::Array2D<double> &t)
{
  int rows=t.dim1(), cols=t.dim2();
  M=gsl_matrix_alloc(rows,cols);
  for(int i=0 ; i<rows ; ++i)
    for(int j=0 ; j<cols ; ++j)
      gsl_matrix_set(M,i,j,t[i][j]);
}



Matrix::Matrix(const BOOM::String &filename)
    : M(NULL)
{
    load(filename);
}



Matrix::Matrix(istream &is)
    : M(NULL)
{
    load(is);
}



Matrix::~Matrix()
{
    if(M) gsl_matrix_free(M);
}



double Matrix::operator()(int row,int col) const
{
  return gsl_matrix_get(M,row,col);
}



double &Matrix::operator()(int row,int col)
{
  return *gsl_matrix_ptr(M,row,col);
}



const GSL::Matrix &Matrix::operator=(const GSL::Matrix &other)
{
  if(M->size1 != other.M->size1 ||
     M->size2 != other.M->size2)
    resize(other.M->size1,other.M->size2);
  gsl_matrix_memcpy(M,other.M);
  return *this;
}



bool Matrix::invert(GSL::Matrix &resultingMatrix) const
{
  gsl_matrix *LU=gsl_matrix_alloc(M->size1,M->size2);
  gsl_matrix_memcpy(LU,M);
  gsl_permutation *P=gsl_permutation_alloc(M->size1);
  int signum;
  gsl_linalg_LU_decomp(LU,P,&signum);
  if(resultingMatrix.M->size1 != M->size1 ||
     resultingMatrix.M->size2 != M->size2)
    resultingMatrix.resize(M->size1,M->size2);
  gsl_linalg_LU_invert(LU,P,resultingMatrix.M);
  gsl_matrix_free(LU);
  gsl_permutation_free(P);
}



void Matrix::getColumn(int column,GSL::Vector &into) const
{
  if(into.getDim()!=M->size1) into.resize(M->size1);
  gsl_matrix_get_col(into.peek(),M,column);
}



void Matrix::getRow(int row,GSL::Vector &into) const
{
  if(into.getDim()!=M->size2) into.resize(M->size2);
  gsl_matrix_get_row(into.peek(),M,row);
}



void Matrix::setColumn(int column,const GSL::Vector &v)
{
  gsl_matrix_set_col(M,column,v.peek());
}



void Matrix::setRow(int row,const GSL::Vector &v)
{
  gsl_matrix_set_row(M,row,v.peek());
}



void Matrix::printOn(ostream &os) const
{
  for(int i=0 ; i<M->size1 ; ++i)
    {
      for(int j=0 ; j<M->size2 ; ++j)
	os<<gsl_matrix_get(M,i,j)<<"\t";
      os<<endl;
    }
}



void Matrix::setAllTo(double d)
{
  gsl_matrix_set_all(M,d);
}



void Matrix::times(const GSL::Matrix &other,GSL::Matrix &result) const
{
  if(result.M->size1 != M->size1 ||
     result.M->size2 != other.M->size2)
    result.resize(M->size1,other.M->size2);
  gsl_blas_dgemm(CblasNoTrans,CblasNoTrans,1.0,M,other.M,0.0,result.M);
}



void Matrix::transpose(GSL::Matrix &Mt) const
{
  int rows=M->size1, cols=M->size2;
  if(Mt.M->size1!=cols || Mt.M->size2!=rows)
    Mt.resize(cols,rows);
  gsl_matrix_transpose_memcpy(Mt.M,M);
}



void Matrix::transpose()
{
  if(M->size1==M->size2)
    gsl_matrix_transpose(M);
  else
    {
      gsl_matrix *m=gsl_matrix_alloc(M->size2,M->size1);
      gsl_matrix_transpose_memcpy(m,M);
      gsl_matrix_free(M);
      M=m;
    }
}



void Matrix::resize(int nRows,int nCols)
{
    if(M) gsl_matrix_free(M);
    M=gsl_matrix_alloc(nRows,nCols);
}



ostream &GSL::operator<<(ostream &os,const GSL::Matrix &M)
{
  M.printOn(os);
  return os;
}



void Matrix::multiply(const GSL::Vector &by,GSL::Vector &result) const
{
  if(result.getDim()!=M->size1) result.resize(M->size1);
  gsl_blas_dgemv(CblasNoTrans,1.0,M,by.peek(),0.0,result.peek());
}



double Matrix::determinant() const
{
  gsl_matrix *LU=gsl_matrix_alloc(M->size1,M->size2);
  gsl_matrix_memcpy(LU,M);
  gsl_permutation *P=gsl_permutation_alloc(M->size1);
  int signum;
  gsl_linalg_LU_decomp(LU,P,&signum);
  double det=gsl_linalg_LU_det(LU,signum);
  gsl_matrix_free(LU);
  return det;
}



void Matrix::becomeIdentity()
{
  gsl_matrix_set_identity(M);
}



bool Matrix::isZero(double epsilon) const
{
  //return gsl_matrix_isnull(M);
  int rows=M->size1, cols=M->size2;
  for(int i=0 ; i<rows ; ++i)
    for(int j=0 ; j<cols ; ++j)
      {
	double entry=gsl_matrix_get(M,i,j);
	if(fabs(entry)>epsilon) return false;
      }
  return true;
}



bool Matrix::isIdentity(double epsilon) const
{
  int rows=M->size1, cols=M->size2;
  for(int i=0 ; i<rows ; ++i)
    for(int j=0 ; j<cols ; ++j)
      {
	double entry=gsl_matrix_get(M,i,j);
	if(i==j)
	  {if(fabs(entry)-1.0>epsilon) return false;}
	else
	  {if(fabs(entry)>epsilon) return false;}
      }
  return true;
}



bool Matrix::sameSize(const GSL::Matrix &R) const
{
  return R.M->size1==M->size1 && R.M->size2==M->size2;
}



void Matrix::getEigenVectors(GSL::Matrix &vectors,
			     BOOM::Vector< complex<double> > &eigenvalues)
{
  GSL::Vector real, imag;
  getEigenVectors(vectors,real,imag);
  int n=M->size1;
  for(int i=0 ; i<n ; ++i)
    eigenvalues.push_back(complex<double>(real[i],imag[i]));
}



void Matrix::getEigenVectors(GSL::Matrix &vectors,GSL::Vector &values,
			     GSL::Vector &imagParts) const
{
  // Convert to TNT format
  int n=M->size1;
  TNT::Array2D<double> *tA=convertToTNT(), tVectors(n,n);
  TNT::Array1D<double> tRealValues(n), tImagValues(n);

  // Compute eigenvalues and eigenvectors using TNT/JAMA
  JAMA::Eigenvalue<double> E(*tA);

  // Convert back to our preferred classes
  E.getV(tVectors);
  E.getRealEigenvalues(tRealValues);
  E.getImagEigenvalues(tImagValues);
  delete tA;
  vectors=tVectors;
  values=tRealValues;
  imagParts=tImagValues;
}



void Matrix::sortEigen(GSL::Matrix &vectors,GSL::Vector &values)
{
  // First, compute a permutation array using QuickSort
  BOOM::Vector<double> v;
  int n=values.getDim();
  v.resize(n);
  for(int i=0 ; i<n ; ++i) v[i]=values[i];
  BOOM::DirectComparator<double> cmp;
  BOOM::VectorSorter<double> sorter(v,cmp);
  BOOM::Vector<int> &permutation=*sorter.sortDescendByIndex();

  // Apply the permutation to the eigenvalues
  GSL::Vector t=values;
  for(int i=0 ; i<n ; ++i) 
    values[i]=t[permutation[i]];

  // Apply the permutation to the eigenvectors
  GSL::Matrix m=vectors;
  GSL::Vector ev(n);
  for(int i=0 ; i<n ; ++i)
    {
      m.getColumn(permutation[i],ev);
      vectors.setColumn(i,ev);
    }

  // Clean up
  delete &permutation;
}



void Matrix::sortEigenAbs(GSL::Matrix &vectors,GSL::Vector &values)
{
  throw "sortEigenAbs -- under construction";
}


 
TNT::Array2D<double> *Matrix::convertToTNT() const
{
  int rows=M->size1, cols=M->size2;
  TNT::Array2D<double> &t=*new TNT::Array2D<double>(rows,cols);
  for(int i=0 ; i<rows ; ++i)
    for(int j=0 ; j<cols ; ++j)
      t[i][j]=gsl_matrix_get(M,i,j);
  return &t;
}



GSL::Matrix &Matrix::operator=(const TNT::Array2D<double> &t)
{
  int rows=t.dim1(), cols=t.dim2();
  resize(rows,cols);
  for(int i=0 ; i<rows ; ++i)
    for(int j=0 ; j<cols ; ++j)
      gsl_matrix_set(M,i,j,t[i][j]);
  return *this;
}



void Matrix::applyThreshold(double epsilon)
{
  Matrix &self=*this;
  int rows=M->size1, cols=M->size2;
  for(int i=0 ; i<rows ; ++i)
    for(int j=0 ; j<cols ; ++j)
      if(fabs(self(i,j))<epsilon) self(i,j)=0.0;
}



void Matrix::add(const GSL::Matrix &other,GSL::Matrix &result) const
{
  const Matrix &self=*this;
  int rows=M->size1, cols=M->size2;
  if(result.M->size1!=rows || result.M->size2!=cols)
    result.resize(rows,cols);
  for(int i=0 ; i<rows ; ++i)
    for(int j=0 ; j<cols ; ++j)
      result(i,j)=self(i,j)+other(i,j);
}



void Matrix::subtract(const GSL::Matrix &other,GSL::Matrix &result) const
{
  const Matrix &self=*this;
  int rows=M->size1, cols=M->size2;
  if(result.M->size1!=rows || result.M->size2!=cols)
    result.resize(rows,cols);
  for(int i=0 ; i<rows ; ++i)
    for(int j=0 ; j<cols ; ++j)
      result(i,j)=self(i,j)-other(i,j);
}



void Matrix::scale(double factor)
{
  Matrix &self=*this;
  int rows=M->size1, cols=M->size2;
  for(int i=0 ; i<rows ; ++i)
    for(int j=0 ; j<cols ; ++j)
      self(i,j)*=factor;
}



void Matrix::power(int power,GSL::Matrix &result)
{
    if(power<1) throw "Matrix::power() : power must be greater than zero";
    if(power==1) result=*this;
    else
    {
        GSL::Matrix temp=*this;
        int n=power-2;
        for(int i=0 ; i<n ; ++i)
        {
            times(temp,result);
            temp=result;
        }
        times(temp,result);
    }
}



void Matrix::load(istream &is)
{
    Matrix &self=*this;
    int rows, columns;
    is>>rows>>columns;
    resize(rows,columns);
    for(int row=0 ; row<rows ; ++row)
        for(int col=0 ; col<columns ; ++col)
            is>>self(row,col);
}



void Matrix::load(const BOOM::String &filename)
{
    ifstream is(filename.c_str());
    if(!is.good()) throw BOOM::String("Error opening file ")+filename;
    load(is);
}



istream &operator>>(istream &is,GSL::Matrix &m)
{
    m.load(is);
    return is;
}


