/****************************************************************
 Optimizer.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include <iostream>
#include "Optimizer.H"
#include "BOOM/Progress.H"
#include <gsl/gsl_errno.h>
using namespace std;
using namespace GSL;


double private_f(const gsl_vector *v,void *params)
{
  ObjectiveFunction *f=reinterpret_cast<ObjectiveFunction*>(params);
  return f->f(GSL::Vector(v));
}



void private_df(const gsl_vector *v,void *params,gsl_vector *df)
{
  ObjectiveFunction *f=reinterpret_cast<ObjectiveFunction*>(params);
  GSL::Vector vWrapper(v), dfWrapper(df);
  f->gradient(vWrapper,dfWrapper);
}



void private_fdf(const gsl_vector *x,void *params,double *f,gsl_vector *df)
{
  *f=private_f(x,params);
  private_df(x,params,df);
}



Optimizer::Optimizer(OptimizerType type,
		     ObjectiveFunction &f,
		     const GSL::Vector &initialPoint,
		     double stepsize,
		     StoppingCriterion criterion,
		     double threshold,
		     int maxIterations)
  : type(type),
    f(f),
    criterion(criterion),
    stepSize(stepsize),
    threshold(threshold),
    maxIterations(maxIterations),
    algorithm(NULL),
    initialPoint(initialPoint),
    iterations(0)
{
  dimension=initialPoint.getDim();
  setAlgorithm();
}



void Optimizer::setAlgorithm()
{
  switch(type)
    {
    case FLETCHER_REEVES:
      algorithm=gsl_multimin_fdfminimizer_conjugate_fr;
      break;
    case POLAK_RIBIERE:
      algorithm=gsl_multimin_fdfminimizer_conjugate_pr;
      break;
    case BFGS:
      algorithm=gsl_multimin_fdfminimizer_vector_bfgs;
      break;
    case STEEPEST_DESCENT:
      algorithm=gsl_multimin_fdfminimizer_steepest_descent;
      break;
    case SIMPLEX:
      throw "SIMPLEX is not implemented yet; sorry about that.";
      //algorithm=gsl_multimin_fminimizer_nmsimplex;
      break;
    }
}



bool Optimizer::run()
{
  bool success=true;
  gsl_multimin_function_fdf func;
  func.f=&private_f;
  func.df=&private_df;
  func.fdf=&private_fdf;
  func.n=dimension;
  func.params=reinterpret_cast<void*>(&f);

  gsl_multimin_fdfminimizer *s=
    gsl_multimin_fdfminimizer_alloc(algorithm,dimension);

  gsl_multimin_fdfminimizer_set(s,&func,initialPoint.peek(),stepSize,
				threshold);

  iterations=0;
  bool keepGoing=true;
  BOOM::Progress progress;
  if(criterion==BY_ITERATION_COUNT) progress.start(maxIterations);
  while(keepGoing)
    {
      int status=gsl_multimin_fdfminimizer_iterate(s);
      //cout<<"STATUS="<<gsl_strerror(status)<<endl;
      ++iterations;
      switch(status)
      {
          case GSL_SUCCESS:
              break;
          case GSL_CONTINUE:
              break;
          case GSL_EUNDRFLW:
          case GSL_EOVRFLW:
              success=false;
              keepGoing=false;
              continue;
          case GSL_ENOPROG: // no progress (means convergence...?)
              success=true;
              keepGoing=false;
              continue;
          default:
              cout<<"GSL error: "<<gsl_strerror(status)<<endl;
              continue;
      }

      switch(criterion)
	{
	case BY_THRESHOLD:
	  status=gsl_multimin_test_gradient(s->gradient,threshold);
	  if(status==GSL_SUCCESS) keepGoing=false;
	  break;
	case BY_ITERATION_COUNT:
	  if(iterations>=maxIterations) keepGoing=false;
	  //cerr<<progress.getProgress(iterations)<<endl;
	  break;
	case BY_EITHER:
	  if(iterations>=maxIterations) keepGoing=false;
	  else
	    {
	      status=gsl_multimin_test_gradient(s->gradient,threshold);
	      if(status==GSL_SUCCESS) keepGoing=false;
	    }
	  break;
	}
    }
  optimalPoint=GSL::Vector(s->x);
  gsl_multimin_fdfminimizer_free(s);
  return success;
}



const GSL::Vector &Optimizer::getOptimalPoint()
{
  return optimalPoint;
}



ostream &operator<<(ostream &os,OptimizerType type)
{
  switch(type)
    {
    case FLETCHER_REEVES:
      os<<"FLETCHER_REEVES";
      break;
    case POLAK_RIBIERE:
      os<<"POLAK_RIBIERE";
      break;
    case BFGS:
      os<<"BFGS";
      break;
    case STEEPEST_DESCENT:
      os<<"STEEPEST_DESCENT";
      break;
    case SIMPLEX:
      os<<"SIMPLEX";
      break;
    }
}



ostream &operator<<(ostream &os,StoppingCriterion criterion)
{
  switch(criterion)
    {
    case BY_THRESHOLD:
      os<<"BY_THRESHOLD";
      break;
    case BY_ITERATION_COUNT:
      os<<"BY_ITERATION_COUNT";
      break;
    case BY_EITHER:
      os<<"BY_EITHER";
      break;
    }
}



StoppingCriterion GSL::stringToStoppingCriterion(const BOOM::String &s)
{
    if(s=="BY_THRESHOLD") return BY_THRESHOLD;
    if(s=="BY_ITERATION_COUNT") return BY_ITERATION_COUNT;
    if(s=="BY_EITHER") return BY_EITHER;
    throw BOOM::String("Invalid stopping criterion: ")+s;
}



size_t Optimizer::iterationsUsed()
{
  return iterations;
}



OptimizerType GSL::stringToOptimizerType(const BOOM::String &s)
{
  if(s=="FLETCHER_REEVES") return FLETCHER_REEVES;
  if(s=="POLAK_RIBIERE") return POLAK_RIBIERE;
  if(s=="BFGS") return BFGS;
  if(s=="STEEPEST_DESCENT") return STEEPEST_DESCENT;
  if(s=="SIMPLEX") return SIMPLEX;
  throw BOOM::String("Unknown OptimizerType: ")+s;
}


