/****************************************************************
 Vector.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include <iostream>
#include <math.h>
#include "Vector.H"
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>
using namespace std;
using namespace GSL;


Vector::Vector(int dim)
  : shouldDelete(true)
{
  V=gsl_vector_alloc(dim);
}



Vector::Vector(const GSL::Vector &other)
  : shouldDelete(true)
{
  V=gsl_vector_alloc(other.V->size);
  gsl_vector_memcpy(V,other.V);
}



Vector::Vector(TNT::Array1D<double> &t)
  : shouldDelete(true)
{
  int n=t.dim();
  V=gsl_vector_alloc(n);
  for(int i=0 ; i<n ; ++i)
    gsl_vector_set(V,i,t[i]);
}



Vector::Vector(gsl_vector *V,bool shouldDelete)
  : V(V),
    shouldDelete(shouldDelete)
{
}



Vector::Vector(const gsl_vector *V)
  : V(const_cast<gsl_vector*>(V)),
    shouldDelete(false)
{
}



Vector::~Vector()
{
  if(shouldDelete)
    gsl_vector_free(V);
}



void Vector::resize(int dim)
{
  gsl_vector_free(V);
  V=gsl_vector_alloc(dim);
}



int Vector::getDim() const
{
  return V->size;
}



double &Vector::operator[](int i)
{
  return *gsl_vector_ptr(V,i);
}



double Vector::operator[](int i) const
{
  return gsl_vector_get(V,i);
}



void Vector::setAllTo(double d)
{
  gsl_vector_set_all(V,d);
}



void Vector::add(const GSL::Vector &other,GSL::Vector &result) const
{
  result=*this;
  gsl_vector_add(result.V,other.V);
}



void Vector::sub(const GSL::Vector &other,GSL::Vector &result) const
{
  result=*this;
  gsl_vector_sub(result.V,other.V);
}



void Vector::scale(double factor)
{
  gsl_vector_scale(V,factor);
}



double Vector::norm() const
{
  return gsl_blas_dnrm2(V);
}



bool Vector::isNull() const
{
  return gsl_vector_isnull(V);
}



void Vector::addConstant(double d,GSL::Vector &result) const
{
  result=*this;
  gsl_vector_add_constant(result.V,d);
}



double Vector::dotProduct(const GSL::Vector &other) const
{
  double d;
  gsl_blas_ddot(V,other.V,&d);
  return d;
}



GSL::Vector &Vector::operator=(const GSL::Vector &other)
{
  if(V->size!=other.V->size)
    {
      gsl_vector_free(V);
      V=gsl_vector_alloc(other.V->size);
    }
  gsl_vector_memcpy(V,other.V);
}



gsl_vector *Vector::peek()
{
  return V;
}



const gsl_vector *Vector::peek() const
{
  return V;
}



void GSL::Vector::printOn(ostream &os) const
{
  int n=V->size;
  os<<n<<" ";
  for(int i=0 ; i<n ; ++i)
    os<<gsl_vector_get(V,i)<<" ";
}



ostream &GSL::operator<<(ostream &os,const GSL::Vector &v)
{
  v.printOn(os);
  return os;
}



GSL::Vector &Vector::operator=(const TNT::Array1D<double> &t)
{
  int n=t.dim();
  resize(n);
  for(int i=0 ; i<n ; ++i)
    gsl_vector_set(V,i,t[i]);
  return *this;
}



void Vector::applyThreshold(double epsilon)
{
  Vector &self=*this;
  int n=V->size;
  for(int i=0 ; i<n ; ++i)
    if(fabs(self[i])<epsilon) self[i]=0.0;
}



void Vector::loadFrom(istream &is)
{
  Vector &self=*this;
  int n;
  is>>n;
  resize(n);
  for(int i=0 ; i<n ; ++i)
    is>>self[i];
}



istream &GSL::operator>>(istream &is,GSL::Vector &v)
{
  v.loadFrom(is);
  return is;
}



void Vector::accumulate(const GSL::Vector &other)
{
  gsl_vector_add(V,other.V);
}


