/***********************************************************************
 MultSeqAlignment.C
 bmajoros@tigr.org

 BOOM : Bioinformatics Object Oriented Modules

 This software is open source, and is distributed under the terms of the
 ARTISTIC LICENSE, as described at www.opensource.org.
 ***********************************************************************/
#include "MultSeqAlignment.H"
#include "ProteinTrans.H"
#include "DnaAlphabet.H"
#include <iostream>
#include <fstream>
using namespace std;
using namespace BOOM;


MultSeqAlignment::MultSeqAlignment(const MultiAlignment &strAlign,
				   Alphabet &alphabet,Symbol gapSymbol)
  : length(0),
    alphabet(alphabet),
    gapSymbol(gapSymbol),
    phase(strAlign.getPhase())
{
  convert(strAlign,gapSymbol);
}



MultSeqAlignment::MultSeqAlignment(Alphabet &alphabet,Symbol gapSymbol)
  : alphabet(alphabet),
    gapSymbol(gapSymbol)
{
  // default ctor
}



Symbol MultSeqAlignment::getGapSymbol() const {
  return gapSymbol;
}



MultSeqAlignment::~MultSeqAlignment()
{
  int n=tracks.size();
  for(int i=0 ; i<n ; ++i)
    delete tracks[i];
}



void MultSeqAlignment::convert(const MultiAlignment &alignment,Symbol gap)
{
  int n=alignment.getNumTracks();
  for(int i=0 ; i<n ; ++i) {
    AlignmentTrack &oldTrack=alignment.getIthTrack(i);
    BOOM::String seqStr=oldTrack.getSeq();
    seqStr.toupper();
    BOOM::String name=oldTrack.getName();
    int ID=oldTrack.getName();
    int begin=oldTrack.getBegin(), length=oldTrack.getLength();
    AlignmentSeq *newTrack=new AlignmentSeq(name,ID,gap,begin);
    Sequence seq(seqStr,alphabet);
    int L=seq.getLength();
    for(int i=0 ; i<L ; ++i)
      if(seq[i]==INVALID_SYMBOL) seq[i]=gap;
    newTrack->getSeq()=seq;
    if(oldTrack.isTarget()) newTrack->setTarget();
    addTrack(newTrack);
  }
}




void MultSeqAlignment::addTrack(AlignmentSeq *track)
{
  int trackID=tracks.size();
  tracks.push_back(track);
  nameToTrack[track->getName()]=trackID;
  track->setID(trackID);
  if(tracks.size()==1) length=track->getLength();
}



void MultSeqAlignment::releaseTrack(int i)
{
    tracks[i]=NULL;
}



AlignmentSeq &MultSeqAlignment::getTrackByName(const BOOM::String 
					        &trackName)
{
  int trackID=nameToTrack[trackName];
  return *tracks[trackID];
}



AlignmentSeq &MultSeqAlignment::findOrCreateTrack(const BOOM::String &
						  trackName)
{
  Symbol N=DnaAlphabet::global.lookup('N');
  if(!nameToTrack.isDefined(trackName))
    {
      int ID=tracks.size();
      nameToTrack[trackName]=ID;
      AlignmentSeq *track=new AlignmentSeq(trackName,ID,N);
      tracks.push_back(track);
      return *track;
    }
  return *tracks[nameToTrack[trackName]];
}



AlignmentSeq &MultSeqAlignment::getIthTrack(int i)
{
  return *tracks[i];
}



int MultSeqAlignment::getNumTracks() const
{
  return tracks.size();
}



int MultSeqAlignment::getLength() const
{
  return length;
}



void MultSeqAlignment::deleteSequences()
{
  int n=tracks.size();
  for(int i=0 ; i<n ; ++i)
      if(tracks[i])
          tracks[i]->deleteSequence();
}



void MultSeqAlignment::save(const BOOM::String &filename) {
    ofstream os(filename.c_str());
    if(!os.good()) throw BOOM::String("Error writing to file: ")+filename;
    printSlice(os,0,length,'+',length);
}



void MultSeqAlignment::save(ostream &os) {
    printSlice(os,0,length,'+',length);
}



void MultSeqAlignment::printSlice(ostream &os,int begin,int length,
                                  char strand,int lineLength) const
{
    int n=tracks.size();
    int end=begin+length;
    for(int pos=begin ; pos<end ; pos+=lineLength)
    {
        int L=end-pos;
        if(L>lineLength) L=lineLength;
        for(int i=0 ; i<n ; ++i)
	{
            AlignmentSeq *track=tracks[i];
            if(!track) continue;
            BOOM::String label=track->getName();
            os<<"s "<<label<<" "<<pos<<" "<<L<<" + . ";
            BOOM::Sequence trackSeq=track->getSeq();
            if(pos<trackSeq.getLength())
            {
                BOOM::Sequence subseq;
                trackSeq.getSubsequence(pos,L,subseq);
                String &S=*subseq.toString(alphabet);
                if(strand=='-') S=ProteinTrans::reverseComplement(S);
                os<<S;
                delete &S;
            }
            os<<endl;
	}
    }
}



int MultSeqAlignment::rightmostGapPos(const Sequence &S,Symbol gapSymbol) {
  int L=S.getLength();
  for(int i=L-1 ; i>=0 ; --i)
    if(S[i]==gapSymbol) return i;
  return -1;
}



MultSeqAlignment *MultSeqAlignment::getSlice(int begin,int length)
{
  MultSeqAlignment &A=*new MultSeqAlignment(alphabet,gapSymbol);
  A.nameToTrack=nameToTrack;
  /*
  BOOM::Vector<AlignmentSeq*> tracks;
  BOOM::Map<BOOM::String,int> nameToTrack;
  int length, phase;
   */

  int newBegin=tracks[0]->getBegin()+begin;
  int numTracks=tracks.size();
  for(int i=0 ; i<numTracks ; ++i) {
    AlignmentSeq *oldTrack=tracks[i];
    AlignmentSeq *newTrack=new AlignmentSeq(oldTrack->getName(),
					    oldTrack->getID(),gapSymbol,
					    oldTrack->getBegin());
    if(oldTrack->isTarget()) newTrack->setTarget();
    Sequence subseq;
    oldTrack->getSeq().getSubsequence(begin,length,subseq);
    newTrack->addHSP(newBegin,subseq);
    A.addTrack(newTrack);
  }
  return &A;
}



bool MultSeqAlignment::containsGap()
{
  BOOM::Vector<AlignmentSeq*>::iterator cur=tracks.begin(), end=tracks.end();
  for(; cur!=end ; ++cur) {
    AlignmentSeq *track=*cur;
    if(track->getSeq().containsSymbol(gapSymbol)) return true;
  }
  return false;
}




void MultSeqAlignment::deleteTargetGaps(int targetTrackID)
{
  // First, find out how many non-gap columns there are
  AlignmentSeq &targetTrack=*tracks[targetTrackID];
  int nonGaps=0;
  for(int pos=0 ; pos<length ; ++pos)
    if(targetTrack[pos]!=gapSymbol)
      ++nonGaps;

  // Next, splice the non-target tracks, using the target track's gap
  // patterns as a template
  int n=tracks.size();
  for(int trackID=0 ; trackID<n ; ++trackID) {
    if(trackID==targetTrackID) continue;
    AlignmentSeq &track=*tracks[trackID];
    if(track.getLength()==0) continue;
    AlignmentSeq &newTrack=
      *new AlignmentSeq(track.getName(),trackID,gapSymbol,track.getBegin());
    newTrack.extendToLength(nonGaps);
    int to=0;
    for(int pos=0 ; pos<length ; ++pos)
      if(targetTrack[pos]!=gapSymbol)
	newTrack[to++]=track[pos];
    tracks[trackID]=&newTrack;
    delete &track;
  }

  // Now splice the target track
  AlignmentSeq &newTrack=
    *new AlignmentSeq(targetTrack.getName(),targetTrackID,gapSymbol,
		      targetTrack.getBegin());
  newTrack.extendToLength(nonGaps);
  int to=0;
  for(int pos=0 ; pos<length ; ++pos)
    if(targetTrack[pos]!=gapSymbol)
      newTrack[to++]=targetTrack[pos];
  tracks[targetTrackID]=&newTrack;
  delete &targetTrack;

  // Update object fields
  length=nonGaps;
}


