/***********************************************************************
 MultiHAF.H
 bmajoros@tigr.org

 BOOM : Bioinformatics Object Oriented Modules

 This software is open source, and is distributed under the terms of the
 ARTISTIC LICENSE, as described at www.opensource.org.
 ***********************************************************************/
#ifndef INCL_MultiHAF_H
#define INCL_MultiHAF_H
#include "Sequence.H"
#include "String.H"
#include "Alphabet.H"
#include "Vector.H"
#include "ProteinTrans.H"
#include "SubstitutionMatrix.H"
#include "Alignment.H"
#include "BandedSmithWaterman.H"
#include "HAF.H"
#include <iostream>
using namespace std;


/****************************************************************
                       class MultiHAFelem
 ****************************************************************/
namespace BOOM {
class MultiHAFelem
{
  const Sequence *sequence;
  Alignment *alignment;
  String track;
  int seqBookmark;       // index into sequence
  int alignmentBookmark; // index into (pairwise) alignment 
public:
  MultiHAFelem(const Sequence *s);
  virtual ~MultiHAFelem();
  const Sequence *getSequence() const;
  const String &getTrack() const;
  void emitGap();
  void emitBase();
  MatchType getMatchType() const;
  void advanceAlignmentBookmark();
  void advanceSeqBookmark();
  bool isFullyEmitted() const;
  void setAlignment(Alignment *);
  void replaceSequence(const Sequence *);
};



/****************************************************************
                         class MultiHAF
 ****************************************************************/
template<class FLT_TYP>
class MultiHAF
{
  Vector<MultiHAFelem*> elements;
  const SubstitutionMatrix<FLT_TYP> &M;
  FLT_TYP gapPenalty, frameshiftPenalty;
  int bandWidth;

  bool needToInsertIntoRef();
  bool finishedEmitting();
public:
  MultiHAF(const SubstitutionMatrix<FLT_TYP> &,
	   FLT_TYP gapPenalty,int bandWidth,
	   FLT_TYP frameshiftPenalty);
  virtual ~MultiHAF();
  void reset();
  void addSequence(const Sequence &);
  void performAlignment();
  int numTracks() const;
  const String &getIthTrack(int i) const;
};
}
using namespace BOOM;


/****************************************************************
                       MultiHAF methods
 ****************************************************************/

template<class FLT_TYP>
MultiHAF<FLT_TYP>::MultiHAF(const SubstitutionMatrix<FLT_TYP> &M,
			    FLT_TYP gapPenalty,
			    int bandWidth,
			    FLT_TYP frameshiftPenalty)
  : M(M),
    gapPenalty(gapPenalty),
    bandWidth(bandWidth),
    frameshiftPenalty(frameshiftPenalty)
{
  // ctor
}



template<class FLT_TYP>
MultiHAF<FLT_TYP>::~MultiHAF()
{
  // dtor
  
  reset();
}



template<class FLT_TYP>
void MultiHAF<FLT_TYP>::reset()
{
  int n=elements.size();
  for(int i=0 ; i<n ; ++i)
    delete elements[i];
  elements.clear();
}



template<class FLT_TYP>
void MultiHAF<FLT_TYP>::addSequence(const Sequence &sequence)
{
  MultiHAFelem *newElem=new MultiHAFelem(&sequence);
  elements.push_back(newElem);
}



template<class FLT_TYP>
bool MultiHAF<FLT_TYP>::finishedEmitting()
{
  Vector<MultiHAFelem*>::iterator cur=elements.begin(), 
    end=elements.end();
  for(; cur!=end ; ++cur)
    {
      MultiHAFelem &elem=**cur;
      if(!elem.isFullyEmitted()) return false;
    }
  return true;
}



template<class FLT_TYP>
bool MultiHAF<FLT_TYP>::needToInsertIntoRef()
{
  Vector<MultiHAFelem*>::iterator cur=elements.begin(), 
    end=elements.end();
  ++cur;
  for(; cur!=end ; ++cur)
    {
      MultiHAFelem &elem=**cur;
      if(!elem.isFullyEmitted() && elem.getMatchType()==SECOND_UNMATCHED) 
	return true;
    }
  return false;

}



template<class FLT_TYP>
void MultiHAF<FLT_TYP>::performAlignment()
{
  // First, translate the reference sequence, since all comparisons will
  // be in amino acid space
  int numElements=elements.size();
  if(numElements<1) 
    throw "MultiHAF::performAlignment() -- too few sequences to align!";
  MultiHAFelem &refElem=*elements[0];
  const Sequence *refSeq=refElem.getSequence();
  refSeq=ProteinTrans::translate(*refSeq);
  refElem.replaceSequence(refSeq);

  // Now compute all pairwise alignments with the reference sequence
  for(int i=1 ; i<numElements ; ++i)
    {
      MultiHAFelem &elem=*elements[i];
      const Sequence *nucSeq=elem.getSequence();
      Sequence *aminoSeqs[3], *S2_composite;
      for(int frame=0 ; frame<3 ; ++frame)
	aminoSeqs[frame]=ProteinTrans::translate(*nucSeq,frame);
      HAF aligner(*refSeq,M,
		  *aminoSeqs[0],*aminoSeqs[1],*aminoSeqs[2],
		  gapPenalty,bandWidth,frameshiftPenalty);
      Alignment *alignment=aligner.go(S2_composite);
      for(int frame=0 ; frame<3 ; ++frame) delete aminoSeqs[frame];
      elem.setAlignment(alignment);
      elem.replaceSequence(S2_composite);
    }

  // Now merge the pairwise alignments into a textual representation of the
  // full multi-alignment
  while(!finishedEmitting())
    {
      // First, handle all SECOND_UNMATCHEDs (insertion in reference seq)
      while(needToInsertIntoRef())
	{
	  refElem.emitGap();
	  for(int i=1 ; i<numElements ; ++i)
	    {
	      MultiHAFelem &elem=*elements[i];
	      if(!elem.isFullyEmitted() && 
		 elem.getMatchType()==SECOND_UNMATCHED)
		{
		  elem.emitBase();
		  elem.advanceSeqBookmark();
		  elem.advanceAlignmentBookmark();
		}
	      else
		elem.emitGap();
	    }
	}
      
      // Now handle all MATCHes and FIRST_UNMATCHEDs
      if(!refElem.isFullyEmitted())
	{
	  refElem.emitBase();
	  refElem.advanceSeqBookmark();
	}
      else refElem.emitGap();
      for(int i=1 ; i<numElements ; ++i)
	{
	  MultiHAFelem &elem=*elements[i];
	  if(!elem.isFullyEmitted() && elem.getMatchType()==MATCH)
	    {
	      elem.emitBase();
	      elem.advanceSeqBookmark();
	    }
	  else
	    elem.emitGap();
	  elem.advanceAlignmentBookmark();
	}      
    }
}



template<class FLT_TYP>
int MultiHAF<FLT_TYP>::numTracks() const
{
  return elements.size();
}



template<class FLT_TYP>
const String &MultiHAF<FLT_TYP>::getIthTrack(int i) const
{
  return elements[i]->getTrack();
}



#endif
