/****************************************************************
 SumLogProbs.H
 william.majoros@duke.edu

 For adding probabilities in log space.  Uses a table of log values,
 as suggested in (Durbin et al., 1998), to speed up the computation.

 This is open source software distributed under the GPL.
 ****************************************************************/
#ifndef INCL_SumLogProbs_H
#define INCL_SumLogProbs_H
#include "String.H"
#include "Vector.H"
#include "Array1D.H"
#include <fstream>
#include "Constants.H"

namespace BOOM {
    
template <class REAL>
REAL sumLogProbs(REAL a,REAL b);
    
template <class REAL>
REAL sumLogProbs_ordered(REAL larger,REAL smaller);
    
template <class REAL>
REAL sumLogProbs(const Vector<REAL> &);
    
template <class REAL>
REAL sumLogProbs(const Array1D<REAL> &);



//====================================================================    


template <class REAL>
REAL sumLogProbs(REAL logP,REAL logQ)
{
  if(logP>logQ) return sumLogProbs_ordered(logP,logQ);
  return sumLogProbs_ordered(logQ,logP);
}



template <class REAL>
REAL sumLogProbs_ordered(REAL largerValue,REAL smallerValue)
{
  if(smallerValue==NEGATIVE_INFINITY) return largerValue;// necessary!!!!
  return largerValue+log(1+exp(smallerValue-largerValue));
}



template <class REAL>
REAL sumLogProbs(const Vector<REAL> &x) {
    int n=x.size();

    // Pull out the largest value
    int largestI=0;
    REAL largestValue=x[0];
    for(int i=1 ; i<n ; ++i) {
        REAL v=x[i];
        if(v>largestValue) {
            largestValue=v;
            largestI=i;
        }
    }

    // Handle the case of all zeros separately
    if(largestValue==NEGATIVE_INFINITY) return NEGATIVE_INFINITY;

    // Apply the Kingsbury-Raynor formula
    REAL sum=(REAL) 0.0;
    for(int i=0 ; i<n ; ++i) {
        REAL v=x[i];
        if(v==NEGATIVE_INFINITY) continue;
        sum+=exp(v-largestValue);
    }
    return largestValue+log(sum);
}
    


template <class REAL>
REAL sumLogProbs(const Array1D<REAL> &x) {
    int n=x.size();

    // Pull out the largest value
    int largestI=0;
    REAL largestValue=x[0];
    for(int i=1 ; i<n ; ++i) {
        REAL v=x[i];
        if(v>largestValue) {
            largestValue=v;
            largestI=i;
        }
    }

    // Handle the case of all zeros separately
    if(largestValue==NEGATIVE_INFINITY) return NEGATIVE_INFINITY;

    // Apply the Kingsbury-Raynor formula
    REAL sum=(REAL) 0.0;
    for(int i=0 ; i<n ; ++i) {
        REAL v=x[i];
        if(v==NEGATIVE_INFINITY) continue;
        sum+=exp(v-largestValue);
    }
    return largestValue+log(sum);
}


}
#endif
