/****************************************************************
 prune-by-index.C
 william.majoros@duke.edu

 This program loads an unpruned tree generated by build-tree and 
 prunes it to the specified size.  The pruning threshold is 
 specified by index.  The entropies of all nodes are sorted in 
 increasing order, and adjacent pairs are averaged.  The index 
 specified on the command line refers to the associated average, 
 which is then used as the pruning threshold.

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/

#include <iostream>
#include <stdlib.h>
#include <fstream>
#include "BOOM/Exceptions.H"
#include "BOOM/String.H"
#include "BOOM/CommandLine.H"
#include "BOOM/Stack.H"
#include "BOOM/Time.H"
#include "BOOM/Vector.H"
#include "BOOM/Stacktrace.H"
#include "../common/NamesFile.H"
#include "../common/TrainingSet.H"
#include "../common/Schema.H"
#include "TreeNode.H"
#include "../common/Entropy.H"
#include "../common/AttrRangeTable.H"
#include "../common/AttrValuePair.H"
#include "../common/AttributeIterator.H"
#include "DecisionTree.H"

class Application
{
  float pruneThreshold;
  NamesFile namesFile;
  Schema *schema;
  TreeNode *root;
  int pruneIndex;
  Vector<float> thresholds;
  String infile, outfile, nameFile;

  void getPruneThresholds(DecisionTree &);
  TreeNode *prune(TreeNode *);
  BitSet *buildUniversalSet(TrainingSet &);
  Distribution *computeDistribution(BitSet &,TrainingSet &);
  void computeDistribution(BitSet &set,
			   Distribution &distribution,
			   TrainingSet &);
public:
  Application(String nameFile,String infile,String outfile,int pruneIndex);
  void go();
};


void usage()
{
  cerr << "\
prune-by-index [-c] <tree-in> <tree-out> <prune-index> <names-file>\n\
    -c = count number of possible prune points in tree\n\
"
       << endl;  
}


int main(int argc,char *argv[])
{
  CommandLine cmd(argc,argv,"tc");
  if(cmd.numArgs()!=4) 
    {
      usage();
      return -1;
    }

  try
    {
      Time stopwatch;
      stopwatch.startCounting();

      bool dashC=cmd.option('c');
      String infile=cmd.arg(0);
      String outfile=cmd.arg(1);
      int pruneIndex=(dashC ? -1 : cmd.arg(2).asInt());
      String nameFile=cmd.arg(3);
      Application app(nameFile,infile,outfile,pruneIndex);
      app.go();

      stopwatch.stopCounting();
      cout << stopwatch.elapsedTime() << endl;
    }
  catch(RootException &e)
    {
      cout << e.getMessage() << endl;
      return -1;
    }
  catch(const string &msg)
    {
      cout << msg.c_str() << endl;
      return -1;
    }
  catch(const char *msg)
    {
      cout << msg << endl;
      return -1;
    }
  catch(...)
    {
      cout << "unknown exception caught in main()" << endl;
      return -1;
    }

  return 0;
}



Application::Application(String nameFile,String infile,String outfile,
			 int pruneIndex)
  : pruneIndex(pruneIndex),
    infile(infile),
    outfile(outfile),
    nameFile(nameFile)
{
  // ctor
}



void Application::go()
{
  // Load schema
  schema=namesFile.processFile(nameFile);

  // Load unpruned tree
  DecisionTree tree(NULL);
  tree.load(infile);

  // Get pruning threshold
  getPruneThresholds(tree);
  if(pruneIndex<0 || pruneIndex>thresholds.size())
    throw String("Prune index must be between 0 and ")+(thresholds.size()-1);
  pruneThreshold=thresholds[pruneIndex-1];
  cout << "pruning with threshold " << pruneIndex << " out of "
       << thresholds.size() << " (" << pruneThreshold << ")" << endl;

  // Prune tree
  TreeNode *root=tree.getRoot();
  root=prune(root);

  // Write into output file
  DecisionTree newTree(root);
  newTree.save(outfile);
}



BitSet *Application::buildUniversalSet(TrainingSet &cases)
{
  int n=cases.numCases();
  BitSet *set=new BitSet(n);
  int i;
  for(i=0 ; i<n ; ++i)
    set->addMember(i);
  return set;
}



Distribution *Application::computeDistribution(BitSet &set,
					       TrainingSet &trainingSet)
{
  Distribution &distribution=
    *new Distribution(schema->getNumCategories());
  computeDistribution(set,distribution,trainingSet);

  return &distribution;
}



void Application::computeDistribution(BitSet &set,
				      Distribution &distribution,
				      TrainingSet &trainingSet)
{
  distribution.setAllTo(0.0);

  int n=set.getMaxSize();
  int cardinality=0;
  for(int i=0 ; i<n ; ++i)
    if(set.isMember(i))
      {
	TrainingCase &c=trainingSet.getCase(i);
	int category=c.getCategory();
	distribution[category]+=1;
	++cardinality;
      }

  int numCat=schema->getNumCategories();
  if(cardinality>0)
    for(int i=0 ; i<numCat ; ++i)
      distribution[i]/=cardinality;
}



TreeNode *Application::prune(TreeNode *node)
{
  if(node->getType()==LEAF_NODE) return node;

  PredicateNode *pred=(PredicateNode *) node;
  if(pred->getOriginalH()<pruneThreshold)//####
    {
      LeafNode *leaf=new LeafNode(pred->getMostPopularCategory());
      delete pred;
      return leaf;
    }
  else
    {
      pred->attachLeftChild(prune(pred->getLeftChild()));
      pred->attachRightChild(prune(pred->getRightChild()));
      return pred;
    }
}



void Application::getPruneThresholds(DecisionTree &tree)
{
  Vector<float> entropies;
  tree.getEntropies(entropies);
  sort(entropies.begin(),entropies.end());
  
  int i, n=entropies.size();
  for(i=0 ; i<n-1 ; ++i)
    {
      if(entropies[i]==entropies[i+1]) continue;
      float threshold=(entropies[i]+entropies[i+1])/2;
      thresholds.push_back(threshold);
    }
}




