
// demes.C : genetic programming on parallel computers

#include "Framework/Tableau.H"
#include "Framework/IndividualSchema.H"
#include "Framework/SymbolSet.H"
#include "Framework/CrosserOver.H"
#include "Framework/RouletteWheel.H"
#include "Framework/Mutator.H"
#include "Framework/SubtreeMutator.H"
#include "Framework/ClassifiedAR.H"
#include "Framework/PopulationBuilder.H"
#include "Framework/FitnessNormalizer.H"
#include "Framework/Stats.H"
#include "Framework/Tournament.H"
#include "/home/bmajoros/ml/NamesFile.H"
#include "/home/bmajoros/ml/TrainingSet.H"
#include "/home/bmajoros/stlplus/StringTokenizer.H"
#include "/home/bmajoros/stlplus/Random.H"
#include "/home/bmajoros/stlplus/Time.H"
#include "/home/bmajoros/stlplus/MPI_Base.H"
#include "/home/bmajoros/stlplus/impistream.H"
#include "/home/bmajoros/stlplus/ompistream.H"
#include "/home/bmajoros/stlplus/Exceptions.H"
#include <string>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <fstream.h>
#include <strstream.h>
#include <unistd.h>

enum SelectionType
{
  FITNESS_PROPORTIONATE,
  TOURNAMENT_SELECTION
};

class Application
{
  MPI_Base mpi;
  bool echo;
  ConfigFile *configFile;
  string filestem;
  SymbolSet symbolSet;
  IndividualBuilder *individualBuilder;
  Population *population;
  CrosserOver crosserOver;
  Mutator *pointMutator;
  SubtreeMutator *subtreeMutator;
  NamesFile namesFile;
  vector<ClassifiedAR*> trainingSet, testSet;
  int numCategories;
  FitnessNormalizer fitnessNormalizer;
  SelectionType selectionType;
  int tournamentSize;
  int initialHeight;
  Individual *bestOfRun;

  void gatherResults(int &best);
  void immigration(Population &newPopulation,int numMigrants);
  void emmigration(Population &population,int numMigrants,RouletteWheel &);
  void rmOldMigrationFiles();
  void processConfig();
  void setup();
  void setupSchema();
  void setupSymbols();
  void loadTrainingSets();
  void convertToARs(TrainingSet *,vector<ClassifiedAR*> &);
  void bigLoop();
  void evaluateFitness(vector<ClassifiedAR*> &,float proportion);
  void exercise(Individual &,vector<ClassifiedAR*> &ARs,float proportion);
  Population *generateNextPop();
  inline Individual *selectIndividual(RouletteWheel &);
  void finish();
public:
  Application(int &argc,char **&argv);
  int go(int argc,char *argv[]);
};


int main(int argc,char *argv[])
{
  try
    {
      Application app(argc,argv);
      return app.go(argc,argv);
    }
  catch(const char *msg)
    {
      cerr << msg << endl;
      return -1;
    }
  catch(string msg)
    {
      cerr << msg.c_str() << endl;
      return -1;
    }
  catch(RootException &e)
    {
      cout << e.getMessage().c_str() << endl;
      return -1;
    }
  catch(...)
    {
      cerr << "Unknown exception caught in main()" << endl;
      return -1;
    }
}



Application::Application(int &argc,char **&argv)
  : mpi(argc,argv)
{
  // ctor

  echo=(mpi.getProcessRank()==0);
}



int Application::go(int argc,char *argv[])
{
  // Process the command line
  if(argc!=2)
    throw string("demes <filestem>");
  filestem=argv[1];

  // Create initial population, etc.
  setup();

  // Loop through the generations
  bigLoop();

  // Report results
  finish();
  if(mpi.getProcessRank()==0) 
    cout << "\a" << endl; // beep!
}



void Application::setup()
{
  // Process the configuration file
  processConfig();
  initialHeight=Tableau::getGlobal().getInitialHeight();

  // Create schema describing individuals
  setupSchema();

  // Create terminal & nonterminal sets
  setupSymbols();
  individualBuilder=new IndividualBuilder(symbolSet);

  // Load training & test sets
  loadTrainingSets();

  // Create an initial population
  PopulationBuilder populationBuilder(symbolSet);
  const int M=Tableau::getGlobal().getM();
  population=populationBuilder.buildPopulation(M,2,initialHeight);

  // Other misc. initialization
  pointMutator=new Mutator(symbolSet);
  subtreeMutator=new SubtreeMutator(individualBuilder->getTreeBuilder());
  rmOldMigrationFiles();
}



void Application::loadTrainingSets()
{
  string namesFilename=filestem+".names";
  Schema *schema=namesFile.processFile(namesFilename);
  numCategories=schema->getNumCategories();

  string trainFilename=filestem+".data";
  TrainingSet *train=new TrainingSet(*schema,trainFilename);
  convertToARs(train,trainingSet);
  delete train;

  string testFilename=filestem+".test";
  TrainingSet *test=new TrainingSet(*schema,testFilename);
  convertToARs(test,testSet);
  delete test;

  delete schema;
}



void Application::convertToARs(TrainingSet *ts,vector<ClassifiedAR*> &ARs)
{
  int i, n=ts->numCases();
  for(i=0 ; i<n ; ++i)
    {
      TrainingCase &tc=ts->getCase(i);
      ARs.push_back(new ClassifiedAR(tc));
    }
}



void Application::setupSchema()
{
  // Create schema describing individuals

  Tableau &tableau=Tableau::getGlobal();
  const int maxCall=tableau.getMaxCall();
  IndividualSchema::allocateGlobal(maxCall);
  string adfArities=tableau.getAdfArities();
  string entryArities=tableau.getEntryArities();
  
  // First, the ADF's
  StringTokenizer adfTokenizer(adfArities,",;-");
  while(adfTokenizer.hasMoreTokens())
    {
      const char *token=adfTokenizer.nextToken();
      int arity=atoi(token);
      IndividualSchema::getGlobal().addADF(arity);
    }

  // Now the entry point(s)
  StringTokenizer entryTokenizer(entryArities," \t,;-");
  while(entryTokenizer.hasMoreTokens())
    {
      const char *token=entryTokenizer.nextToken();
      int arity=atoi(token);
      IndividualSchema::getGlobal().addEntryPoint(arity);
    }
}


void Application::bigLoop()
{
  bestOfRun=NULL;

  Stopwatch stopwatch;
  stopwatch.start();
  const int G=Tableau::getGlobal().getG();
  int generation, almostG=int(0.75*G);
  float trainPortion=Tableau::getGlobal().getTrainPortion();
  for(generation=0 ; generation<G ; ++generation)
    {
      if(echo) 
	{
	  cout << "processing generation " << generation << ": "; 
	  cout.flush();// ###
	}

      // Evaluate fitness of population members
      if(generation==almostG) trainPortion=1.1;
      evaluateFitness(trainingSet,trainPortion);
      Stats stats(*population);
      if(echo) cout << stats << endl;

      // See if a new bestOfRun has been produced
      Individual *winner=population->getMostFit();
      if(!bestOfRun ||
	 winner->getAdjustedFitness() > bestOfRun->getAdjustedFitness())
	bestOfRun=winner->clone();

      // Generate the next population
      Population *newPopulation=generateNextPop();

      // Replace the old with the new
      delete population;
      population=newPopulation;
    }
  stopwatch.stop();
  if(echo) cout << stopwatch.elapsedTime().c_str() << endl;
}



Population *Application::generateNextPop()
{
  // Get relevant parameters
  Tableau &tableau=Tableau::getGlobal();
  const float Pc=tableau.getPc();
  const float Pm=tableau.getPm();
  const float subtreeMutation=tableau.getSubtreeMutation();
  const float migration=tableau.getImmigration();
  const float cloning=tableau.getCloning();
  const int M=tableau.getM();

  // Allocate a new population
  Population *newPopulation=new Population();
  
  // Decide how many crossovers & mutations to perform
  const int numCrossovers=int(Pc*M);
  int extra=numCrossovers%2;
  const int numCrossoverPairs=int(numCrossovers/2);
  const int pointMutations=int(Pm*M+0.5);
  const int subtreeMutations=int(subtreeMutation*M+0.5);
  const int numMigrants=int(migration*M+0.5);
  const int numCopies=int(cloning*M+0.5)+extra;
  int total=2*numCrossoverPairs+pointMutations+subtreeMutations+
    numMigrants+numCopies;
  if(total != M)
    throw string("crossover+mutation+immigration+cloning must equal 1.0");

  // Create a roulette wheel which reflects relative fitnesses
  RouletteWheel rouletteWheel(*population);
  
  // Crossover
  int i;
  for(i=0 ; i<numCrossoverPairs ; ++i)
    {
      Individual *mother=selectIndividual(rouletteWheel);
      Individual *father=selectIndividual(rouletteWheel);
      while(mother==father) 
	father=selectIndividual(rouletteWheel);
      Individual *daughter, *son;
      crosserOver.crossover(mother,father,daughter,son);
      newPopulation->addMember(daughter);
      newPopulation->addMember(son);
    }

  // Point mutation
  for(i=0 ; i<pointMutations ; ++i)
    {
      Individual *lucky=selectIndividual(rouletteWheel);
      if(Tableau::getGlobal().getMaxHeight()>0)
	lucky->getTreeHeight(); // height is cached for later
      Individual *mutant=pointMutator->mutate(lucky);
      newPopulation->addMember(mutant);
    }

  // Subtree mutation
  for(i=0 ; i<subtreeMutations ; ++i)
    {
      Individual *lucky=selectIndividual(rouletteWheel);
      if(Tableau::getGlobal().getMaxHeight()>0)
	lucky->getTreeHeight(); // height is cached for later
      Individual *mutant=subtreeMutator->mutate(lucky);
      newPopulation->addMember(mutant);
    }

  // Migration
  immigration(*newPopulation,numMigrants);
  emmigration(*population,numMigrants,rouletteWheel);

  // Exploitation (copy)
  for(i=0 ; i<numCopies ; ++i)
    {
      Individual *lucky=selectIndividual(rouletteWheel);
      if(Tableau::getGlobal().getMaxHeight()>0)
	lucky->getTreeHeight(); // height is cached for later
      newPopulation->addMember(lucky->clone());
    }

  return newPopulation;
}



Individual *Application::selectIndividual(RouletteWheel &rouletteWheel)
{
  switch(selectionType)
    {
    case FITNESS_PROPORTIONATE:
      return population->getMember(rouletteWheel.spin());
    case TOURNAMENT_SELECTION:
      {
	Tournament tournament(*population,tournamentSize);
	return tournament.pickWinner();
      }
    }
}



void Application::processConfig()
{  
  string configFilename=filestem+".gp";
  configFile=new ConfigFile(configFilename);

  Tableau::processConfig(*configFile);

  int seed=Tableau::getGlobal().getSeed();
  if(seed) SeedRandomizer(unsigned(seed));
  else randomize();

  selectionType=
    Tableau::getGlobal().useTournament() ? TOURNAMENT_SELECTION :
    FITNESS_PROPORTIONATE;
  tournamentSize=Tableau::getGlobal().getTournamentSize();
  if(selectionType==TOURNAMENT_SELECTION && tournamentSize<=0)
    throw string("set tournament-size to a positive integer");
}



void Application::setupSymbols()
{
  // Nonterminals
  string symbols=Tableau::getGlobal().getNonterminals();
  StringTokenizer tokenizer(symbols,",");
  while(tokenizer.hasMoreTokens())
    {
      const char *token=tokenizer.nextToken();
      if(!strcmp(token,"+")) 
	symbolSet.addNonterminal(new PlusNode());
      else if(!strcmp(token,"-")) 
	symbolSet.addNonterminal(new MinusNode());
      else if(!strcmp(token,"*")) 
	symbolSet.addNonterminal(new TimesNode());
      else if(!strcmp(token,"/")) 
	symbolSet.addNonterminal(new DivideNode());
      else if(!strcmp(token,"sin")) 
	symbolSet.addNonterminal(new SinNode());
      else if(!strcmp(token,"cos")) 
	symbolSet.addNonterminal(new CosNode());
      else if(!strcmp(token,"exp")) 
	symbolSet.addNonterminal(new ExpNode());
      else if(!strcmp(token,"log")) 
	symbolSet.addNonterminal(new LogNode());
      else if(!strcmp(token,"if")) 
	symbolSet.addNonterminal(new IfNode());
      else if(!strcmp(token,"call")) 
	symbolSet.addNonterminal(new CallNode());
      else if(!strcmp(token,"min")) 
	symbolSet.addNonterminal(new MinNode());
      else if(!strcmp(token,"max")) 
	symbolSet.addNonterminal(new MaxNode());
      else if(!strcmp(token,"fmod")) 
	symbolSet.addNonterminal(new FmodNode());
      else if(!strcmp(token,"abs")) 
	symbolSet.addNonterminal(new AbsNode());
      else if(!strcmp(token,"ceil")) 
	symbolSet.addNonterminal(new CeilNode());
      else if(!strcmp(token,"hypot")) 
	symbolSet.addNonterminal(new HypotNode());
      else if(!strcmp(token,"tanh")) 
	symbolSet.addNonterminal(new TanhNode());
      else if(!strcmp(token,"sinh")) 
	symbolSet.addNonterminal(new SinhNode());
      else if(!strcmp(token,"cosh")) 
	symbolSet.addNonterminal(new CoshNode());
      else if(!strcmp(token,"atanh")) 
	symbolSet.addNonterminal(new AtanhNode());
      else if(!strcmp(token,"asinh")) 
	symbolSet.addNonterminal(new AsinhNode());
      else if(!strcmp(token,"acosh")) 
	symbolSet.addNonterminal(new AcoshNode());
      else if(!strcmp(token,"log2")) 
	symbolSet.addNonterminal(new Log2Node());
      else if(!strcmp(token,"log10")) 
	symbolSet.addNonterminal(new Log10Node());
      else if(!strcmp(token,"cbrt")) 
	symbolSet.addNonterminal(new CbrtNode());
      else if(!strcmp(token,"sqrt")) 
	symbolSet.addNonterminal(new CallNode());
      else if(!strcmp(token,"atan")) 
	symbolSet.addNonterminal(new AtanNode());
      else if(!strcmp(token,"asin")) 
	symbolSet.addNonterminal(new AsinNode());
      else if(!strcmp(token,"acos")) 
	symbolSet.addNonterminal(new AcosNode());
      else if(!strcmp(token,"tan")) 
	symbolSet.addNonterminal(new TanNode());
      else if(!strcmp(token,"^")) 
	symbolSet.addNonterminal(new PowNode());
      else if(!strcmp(token,"%")) 
	symbolSet.addNonterminal(new ModNode());
      else if(!strcmp(token,"&&")) 
	symbolSet.addNonterminal(new AndNode());
      else if(!strcmp(token,"||")) 
	symbolSet.addNonterminal(new OrNode());
      else if(!strcmp(token,"!=")) 
	symbolSet.addNonterminal(new NotEqualsNode());
      else if(!strcmp(token,">=")) 
	symbolSet.addNonterminal(new GreaterEqualsNode());
      else if(!strcmp(token,"<=")) 
	symbolSet.addNonterminal(new LessEqualsNode());
      else if(!strcmp(token,"!")) 
	symbolSet.addNonterminal(new NotNode());
      else if(!strcmp(token,"<")) 
	symbolSet.addNonterminal(new LessNode());
      else if(!strcmp(token,">")) 
	symbolSet.addNonterminal(new GreaterNode());
      else
	throw string("nonterminal symbol \"")+token+"\" is not recognized";
    }

  // Terminals
  const float minConst=Tableau::getGlobal().getMinConst();
  const float maxConst=Tableau::getGlobal().getMaxConst();
  symbols=Tableau::getGlobal().getTerminals();
  tokenizer.tokenizeThis(symbols.c_str());
  while(tokenizer.hasMoreTokens())
    {
      const char *token=tokenizer.nextToken();
      if(!strcmp(token,"const")) 
	symbolSet.addTerminal(new ConstantNode(minConst,maxConst));
      else if(!strcmp(token,"var")) 
	symbolSet.addTerminal(new VariableNode());
      else
	throw string("terminal symbol \"")+token+"\" is not recognized";
    }
}



void Application::finish()
{
  // Get best individual from entire run
  evaluateFitness(trainingSet,1.1);
  Individual *winner=population->removeMostFit();
  if(!bestOfRun ||
     winner->getAdjustedFitness() > bestOfRun->getAdjustedFitness())
    {
      delete bestOfRun;
      bestOfRun=winner;
    }

  // Test the winner on the test set
  delete population;
  population=new Population();
  population->addMember(bestOfRun);
  evaluateFitness(testSet,1.1);

  // Report results
  int testSetSize=testSet.size();
  float fitness=bestOfRun->getStandardizedFitness();
  int accuracy=int(100*(1-fitness)+0.5);
  if(mpi.getProcessRank()==0) 
    {
      gatherResults(accuracy);
      cout << "accuracy of winner: " << accuracy << "%" << endl;
      ofstream os((filestem+".gp.tree").c_str());
      os << *bestOfRun << endl;
    }
  else
    {
      ompistream os(0);
      os << accuracy;
      os.flush(0);
    }
}



void Application::gatherResults(int &best)
{
  int i, n=mpi.getNumProcesses(); 
  for(i=1 ; i<n ; ++i)
    {
      int accuracy;
      impistream is;
      is >> accuracy;
      if(accuracy>best) best=accuracy;
    }
}



void Application::evaluateFitness(vector<ClassifiedAR*> &ARs,
				  float proportion)
{
  int i, n=population->size();
  for(i=0 ; i<n ; ++i)
    {
      Individual *ind=population->getMember(i);
      exercise(*ind,ARs,proportion);
    }

  fitnessNormalizer.normalize(*population);
}



void Application::exercise(Individual &ind,vector<ClassifiedAR*> &ARs,
			   float proportion)
{
  int errors=0;

  // Get the result-producing branch, so we can call it
  int resultBranch=Tableau::getGlobal().getResultBranch();
  TreeNode *tree=ind.getEntryPoint(resultBranch);

  // Iterate through the test cases
  int i, n=ARs.size(), num=0;
  for(i=0 ; i<n ; ++i)
    {
      // Exercise only on a random sample of cases
      if(Random0to1()>=proportion) continue;

      // Get the next test case
      ClassifiedAR *ar=ARs[i];

      // Call the evolved procedure
      float value=tree->eval(ar->getAR(),ind);

      // Discretize the result into a category
      int category=int(value+0.5);
      if(category<0) category=0;
      else if(category>numCategories-1) 
	category=numCategories-1;

      // Score the result
      if(category!=ar->getCategory())
	++errors;
      ++num;
    }

  ind.setStandardizedFitness(errors/float(num));
  ind.setAdjustedFitness(1.0/(1.0+errors));
}



void Application::immigration(Population &newPopulation,int numMigrants)
{
  // Build filename
  int rank=mpi.getProcessRank();
  char filename1[50], filename2[20];
  ostrstream os(filename1,sizeof(filename1));
  os << "deme" << rank << ".imm.ready" << ends;
  
  // See whether .ready file exists
  ifstream isReady(filename1);
  if(isReady.good())
    {
      // Open migrants file
      ostrstream os(filename2,sizeof(filename2));
      os << "deme" << rank << ".imm" << ends;
      ifstream isMigrants(filename2);
      if(!isMigrants) 
	throw string("cannot open immigration file ")+filename2;

      // Load migrants
      int i;
      for(i=0 ; i<numMigrants ; ++i)
	{
	  Individual *migrant=new Individual();
	  migrant->load(isMigrants);
	  newPopulation.addMember(migrant);
	}

      // Delete files
      isReady.close();
      isMigrants.close();
      unlink(filename1);
      unlink(filename2);
    }
  else
    {
      // The migration file is not present -- simulate immigration
      // by generating some random individuals

      int i;
      for(i=0 ; i<numMigrants ; ++i)
	newPopulation.addMember(
	  individualBuilder->buildIndividual(initialHeight));
    }
}



void Application::emmigration(Population &population,int numMigrants,
			      RouletteWheel &rouletteWheel)
{
  // Build filename
  int rank=mpi.getProcessRank();
  int dest=(rank+1)%mpi.getNumProcesses();
  char filename[50];
  ostrstream os(filename,sizeof(filename));
  os << "deme" << dest << ".imm" << ends;
  
  // Make sure previous boatload of migrants has been received
  ifstream isImm(filename);
  if(isImm.good()) return; // file exists - migration not possible

  // Write migrants into file
  ofstream osImm(filename);
  int i;
  for(i=0 ; i<numMigrants ; ++i)
    {
      Individual *migrant=selectIndividual(rouletteWheel);
      migrant->save(osImm);
    }
  osImm.close();

  // Create .ready file to indicate that writing is done
  ostrstream os2(filename,sizeof(filename));
  os2 << "deme" << dest << ".imm.ready" << ends;
  ofstream osReady(filename);
  osReady << rank << endl;
  osReady.close();
}



void Application::rmOldMigrationFiles()
{
  /*
    It is very important that we remove any migration files
    from the previous run, because if the previous run had
    a smaller population size, we will try to load the wrong
    number of migrants from the file (causing a seg fault).
   */

  if(mpi.getProcessRank()==0)
    system("echo 'rm deme*.imm* >& /dev/null' | /bin/csh");
  mpi.syncWithAllProcesses();
}

