/****************************************************************
 Classifier.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include <iostream>
#include "Classifier.H"
#include "BOOM/Constants.H"
using namespace std;
using namespace BOOM;


Classifier::Classifier()
  : model(NULL)
{
  // ctor
}



Classifier::Classifier(Model *m,TrainingSet &s)
  : model(m)
{
  computeCentroids(s);
}



Classifier::~Classifier()
{
  delete model;
  int n=centroids.size();
  for(int i=0 ; i<n ; ++i) delete centroids[i];
}



int Classifier::operator()(const Case &c)
{
  GSL::Vector casePoint;
  model->mapToPoint(c,casePoint);
  int n=centroids.size();
  GSL::Vector diffVector;
  double smallestDistance=BOOM::POSITIVE_INFINITY;
  int bestCategory=-1;
  for(int i=0 ; i<n ; ++i)
    {
      GSL::Vector &centroid=*centroids[i];
      casePoint.sub(centroid,diffVector);
      double distance=diffVector.norm();
      if(distance<smallestDistance)
	{
	  bestCategory=i;
	  smallestDistance=distance;
	}
    }
  return bestCategory;
}



void Classifier::printOn(ostream &os) const
{
  os<<*model<<endl;
  int n=centroids.size();
  os<<n<<endl;
  for(int i=0 ; i<n ; ++i)
    os<<*centroids[i]<<endl;
}



void Classifier::readFrom(istream &is)
{
  delete model;
  int n=centroids.size();
  for(int i=0 ; i<n ; ++i) delete centroids[i];

  model=new Model;
  is>>*model;

  is>>n;
  for(int i=0 ; i<n ; ++i)
    {
      GSL::Vector *v=new GSL::Vector;
      is>>*v;
      centroids.push_back(v);
    }
}



void Classifier::computeCentroids(TrainingSet &s)
{
  // Initialize a set of null vectors for summing the training cases
  int numCategories=s.getSchema().getNumCategories();
  BOOM::Array1D<GSL::Vector*> sums(numCategories);
  BOOM::Array1D<int> sampleSizes(numCategories);
  sampleSizes.setAllTo(0);
  int modelDim=model->getDim();
  for(int i=0 ; i<numCategories ; ++i)
    {
      GSL::Vector *v=sums[i]=new GSL::Vector(modelDim);
      v->setAllTo(0.0);
    }

  // Sum the training cases (by category)
  int n=s.numCases();
  GSL::Vector casePoint;
  for(int i=0 ; i<n ; ++i)
    {
      TrainingCase &c=s.getCase(i);
      int cat=c.getCategory();
      model->mapToPoint(c,casePoint);
      sums[cat]->accumulate(casePoint);
      ++sampleSizes[cat];
    }

  // Normalize the sums to produce centroids
  for(int i=0 ; i<numCategories ; ++i)
    {
      GSL::Vector *v=sums[i];
      double factor=1.0/sampleSizes[i];
      v->scale(factor);
      centroids.push_back(v);
    }
}



ostream &operator<<(ostream &os,const Classifier &c)
{
  c.printOn(os);
  return os;
}



istream &operator>>(istream &is,Classifier &c)
{
  c.readFrom(is);
  return is;
}



Model &Classifier::getModel()
{
  return *model;
}


