/****************************************************************
 train-lda.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include <iostream>
#include <math.h>
#include <fstream>
#include "BOOM/String.H"
#include "BOOM/CommandLine.H"
#include "BOOM/GSL/Matrix.H"
#include "BOOM/GSL/Vector.H"
#include "../common/Schema.H"
#include "../common/TrainingSet.H"
#include "../common/NamesFile.H"
#include "Discriminant.H"
#include "Model.H"
#include "Classifier.H"
using namespace std;
using namespace BOOM;


class Application
{
  Schema *schema;
  NamesFile namesFile;
  int numCategories, numVariables, numCases;
  GSL::Vector means; // indexed by variable
  GSL::Matrix groupMeans; // indexed by category, then variable 
  GSL::Vector casesPerCategory;
  GSL::Matrix T,B,W; // B=between-groups, W=within groups, T=B+W
  TrainingSet *trainingSet;
  Classifier *classifier;
  BOOM::String outfile;
  int maxNumDiscrim;

  void buildModel();
  void computeT();
  void computeW();
  void generateOutput();
  void getMeans();
  void getGroupMeans();
public:
  Application();
  int main(int argc,char *argv[]);
};


int main(int argc,char *argv[])
{
  try
    {
      Application app;
      return app.main(argc,argv);
    }
  catch(const char *p)
    {
      cerr << p << endl;
    }
  catch(const string &msg)
    {
      cerr << msg.c_str() << endl;
    }
  catch(const exception &e)
    {
      cerr << "STL exception caught in main:\n" << e.what() << endl;
    }
  catch(...)
    {
      cerr << "Unknown exception caught in main" << endl;
    }
  return -1;
}



Application::Application()
{
  // ctor
}



int Application::main(int argc,char *argv[])
{
  // Process command line
  CommandLine cmd(argc,argv,"d:");
  if(cmd.numArgs()!=3) throw 
    "train-lda [-d #discrim] <train-file> <names-file> <out-file>";
  String trainFile=cmd.arg(0);
  String namesFilename=cmd.arg(1);
  outfile=cmd.arg(2);
  maxNumDiscrim=(cmd.option('d') ? cmd.optParm('d').asInt() : 0);
  
  // Load configuration
  schema=namesFile.processFile(namesFilename);
  numCategories=schema->getNumCategories();
  numVariables=schema->numAttributes();
  
  // Load training data
  trainingSet=new TrainingSet(*schema,trainFile);
  numCases=trainingSet->numCases();
  
  // Build model
  buildModel();
  
  // Generate output
  generateOutput();

  // Classify the training cases and evaluate accuracy
  int right=0;
  for(int i=0 ; i<numCases ; ++i)
    {
      TrainingCase &c=trainingSet->getCase(i);
      int computed=(*classifier)(c);
      int expected=c.getCategory();
      if(computed==expected) ++right;
    }
  double accuracy=int(right/double(numCases)*100+5/9);
  cout<<"accuracy on training set: "<<accuracy<<endl;
  
  return 0;
}



void Application::getMeans()
{
  // First, compute sums
  means.setAllTo(0.0);
  for(int i=0 ; i<numCases ; ++i)
    {
      TrainingCase &c=trainingSet->getCase(i);
      for(int j=0 ; j<numVariables ; ++j)
	means[j]+=c.getContinuous(j);
    }

  // Divide by numCases to get the means
  for(int j=0 ; j<numVariables ; ++j)
    means[j]/=numCases;
}



void Application::getGroupMeans()
{
  // First, compute sums
  casesPerCategory.resize(numCategories);
  casesPerCategory.setAllTo(0.0);
  groupMeans.setAllTo(0.0);
  for(int i=0 ; i<numCases ; ++i)
    {
      TrainingCase &c=trainingSet->getCase(i);
      int category=c.getCategory();
      casesPerCategory[category]+=1.0;
      for(int j=0 ; j<numVariables ; ++j)
	groupMeans(category,j)+=c.getContinuous(j);
    }

  // Divide by n_sub_j to get the means
  for(int i=0 ; i<numCategories ; ++i)
    for(int j=0 ; j<numVariables ; ++j)
      groupMeans(i,j)/=casesPerCategory[i];
}



void Application::computeT()
{
  // First, compute residuals: Xijr-mean(Xr)
  GSL::Matrix residuals(numCases,numVariables);
  for(int i=0 ; i<numCases ; ++i)
    {
      TrainingCase &c=trainingSet->getCase(i);
      for(int j=0 ; j<numVariables ; ++j)
	residuals(i,j)=c.getContinuous(j)-means[j];
    }

  // Now populate T
  T.resize(numVariables,numVariables);
  T.setAllTo(0.0);
  for(int i=0 ; i<numVariables ; ++i)
    for(int j=0 ; j<numVariables ; ++j)
      for(int k=0 ; k<numCases ; ++k)
	T(i,j)+=residuals(k,i)*residuals(k,j);
}



void Application::computeW()
{
  // First, compute residuals: Xijr-mean(Xr)
  GSL::Matrix residuals(numCases,numVariables);
  for(int i=0 ; i<numCases ; ++i)
    {
      TrainingCase &c=trainingSet->getCase(i);
      int cat=c.getCategory();
      for(int j=0 ; j<numVariables ; ++j)
	residuals(i,j)=c.getContinuous(j)-groupMeans(cat,j);
    }

  // Now populate W
  W.resize(numVariables,numVariables);
  W.setAllTo(0.0);
  for(int i=0 ; i<numVariables ; ++i)
    for(int j=0 ; j<numVariables ; ++j)
      for(int k=0 ; k<numCases ; ++k)
	W(i,j)+=residuals(k,i)*residuals(k,j);
}



void Application::buildModel()
{
  // Get the means within and across groups
  means.resize(numVariables);
  groupMeans.resize(numCategories,numVariables);
  getMeans();
  getGroupMeans();

  // Compute the B, T, and W matrices
  computeT();
  computeW();
  T.subtract(W,B); // B=T-W;

  //B.scale(1.0/(numCases-1));
  //W.scale(1.0/(numCases-numCategories));

  GSL::Matrix Winv;
  if(!W.invert(Winv))
    throw "within-groups sums-of-squares-and-cross-products matrix is "
          "non-invertible!";
  GSL::Matrix WinvB;
  Winv.times(B,WinvB);

  // Perform eigenvector decomposition
  GSL::Matrix eigenvectors;
  GSL::Vector eigenvalues, imaginaryParts;
  WinvB.getEigenVectors(eigenvectors,eigenvalues,imaginaryParts);

  GSL::Matrix::sortEigen(eigenvectors,eigenvalues);
  cout<<"rounded eigenvalues: ";
  for(int i=0 ; i<numVariables ; ++i)
    {
      double e=eigenvalues[i];
      if(e<1e-14) e=0;
      cout<<int(100*e+5/9.0)/100.0;
      if(i<numVariables-1) cout<<", ";
    }
  cout<<endl;

  // Decide how many discriminants to use
  int numDiscrim=min(numVariables,numCategories-1);
  if(numDiscrim>maxNumDiscrim && maxNumDiscrim>0) numDiscrim=maxNumDiscrim;
  cout<<"using "<<numDiscrim<<" discriminant function"
      <<(numDiscrim>1?"s":"")<<endl;
  
  // Instantiate discriminants into a model
  Model *model=new Model;
  GSL::Vector eigenvector;
  for(int i=0 ; i<numDiscrim ; ++i)
    {
      eigenvectors.getColumn(i,eigenvector); //### SHOULD BE COLUMN...

      //### scaling the eigenvectors...not sure if this is right...
      /*
      int n=eigenvector.getDim();
      double sum=0;
      for(int j=0 ; j<n ; ++j) sum+=eigenvector[j]*eigenvector[j];
      eigenvector.scale(1/sqrt(sum));
      */
      //###

      Discriminant *d=new Discriminant(eigenvector,eigenvalues[i]);
      model->addDiscriminant(d);
    }

  // Compute centroids and instantate a classifier
  classifier=new Classifier(model,*trainingSet);
}



void Application::generateOutput()
{
  ofstream os(outfile.c_str());
  os<<*classifier<<endl;
}
