/****************************************************************
 pca.C
 william.majoros@duke.edu

 Principal Components Analysis

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include <iostream>
#include <math.h>
#include <fstream>
#include "BOOM/String.H"
#include "BOOM/CommandLine.H"
#include "BOOM/GSL/Matrix.H"
#include "BOOM/GSL/Vector.H"
#include "../common/Schema.H"
#include "../common/TrainingSet.H"
#include "../common/NamesFile.H"
using namespace std;
using namespace BOOM;


class Application
{
  Schema *schema;
  NamesFile namesFile;
  int numCategories, numVariables, numCases, numComponents;
  GSL::Vector means; // indexed by variable
  GSL::Vector sigmas;
  GSL::Matrix T;
  TrainingSet *trainingSet;
  BOOM::String outfile;
  bool shouldNormalize;

  void PCA();
  void computeT();
  void getMeans();
  void normalizeVariables();
public:
  Application();
  int main(int argc,char *argv[]);
};


int main(int argc,char *argv[])
{
  try
    {
      Application app;
      return app.main(argc,argv);
    }
  catch(const char *p)
    {
      cerr << p << endl;
    }
  catch(const string &msg)
    {
      cerr << msg.c_str() << endl;
    }
  catch(const exception &e)
    {
      cerr << "STL exception caught in main:\n" << e.what() << endl;
    }
  catch(...)
    {
      cerr << "Unknown exception caught in main" << endl;
    }
  return -1;
}



Application::Application()
{
  // ctor
}



int Application::main(int argc,char *argv[])
{
  // Process command line
  CommandLine cmd(argc,argv,"r");
  if(cmd.numArgs()!=4) throw 
"\npca <train-file> <names-file> <outfile> <num-components>\n\
     where -r supresses normalization of variables\n";
  String trainFile=cmd.arg(0);
  String namesFilename=cmd.arg(1);
  outfile=cmd.arg(2);
  numComponents=cmd.arg(3).asInt();
  shouldNormalize=false; //!cmd.option('r');
  
  // Load configuration
  schema=namesFile.processFile(namesFilename);
  numCategories=schema->getNumCategories();
  numVariables=schema->numAttributes();
  
  // Load training data
  trainingSet=new TrainingSet(*schema,trainFile);
  numCases=trainingSet->numCases();
  
  // Perform PCA
  PCA();
  
  return 0;
}



void Application::getMeans()
{
  GSL::Vector &sumX=means, &sumXX=sigmas;

  // First, compute sums
  sumX.setAllTo(0.0);
  for(int i=0 ; i<numCases ; ++i)
    {
      TrainingCase &c=trainingSet->getCase(i);
      for(int j=0 ; j<numVariables ; ++j)
	{
	  double x=c.getContinuous(j);
	  sumX[j]+=x;
	  sumXX[j]+=x*x;
	}
    }

  // Compute standard deviations
  if(shouldNormalize)
    for(int j=0 ; j<numVariables ; ++j)
      sigmas[j]=(sumXX[j]-sumX[j]*sumX[j]/numCases)/(numCases-1.0);
  else
    sigmas.setAllTo(1.0);

  // Divide by numCases to get the means
  for(int j=0 ; j<numVariables ; ++j)
    means[j]=sumX[j]/numCases;
}



void Application::computeT()
{
  // First, compute residuals: Xijr-mean(Xr)
  GSL::Matrix residuals(numCases,numVariables);
  for(int i=0 ; i<numCases ; ++i)
    {
      TrainingCase &c=trainingSet->getCase(i);
      for(int j=0 ; j<numVariables ; ++j)
	residuals(i,j)=(c.getContinuous(j)-means[j])/sigmas[j];
    }

  // Now populate T
  T.resize(numVariables,numVariables);
  T.setAllTo(0.0);
  for(int i=0 ; i<numVariables ; ++i)
    for(int j=0 ; j<numVariables ; ++j)
      for(int k=0 ; k<numCases ; ++k)
	T(i,j)+=residuals(k,i)*residuals(k,j);
}



void Application::PCA()
{
  // Get the means & std devs
  means.resize(numVariables);
  sigmas.resize(numVariables);
  getMeans();

  // Compute covariance matrix
  computeT();
  T.scale(1.0/(numCases-1));

  // Perform eigenvector decomposition
  GSL::Matrix eigenvectors;
  GSL::Vector eigenvalues, imaginaryParts;
  T.getEigenVectors(eigenvectors,eigenvalues,imaginaryParts);
  GSL::Matrix::sortEigen(eigenvectors,eigenvalues);
  cout<<"rounded eigenvalues: ";
  for(int i=0 ; i<numVariables ; ++i)
    {
      double e=eigenvalues[i];
      cout<<int(e);
      if(i<numVariables-1) cout<<", ";
    }
  cout<<endl;

  // Get the principal components & write model file
  ofstream os(outfile.c_str());
  os<<numComponents<<endl;
  GSL::Vector eigenvector;
  for(int i=0 ; i<numComponents ; ++i)
    {
      eigenvectors.getColumn(i,eigenvector);
      cout<<"component "<<i<<": "<<eigenvector<<endl;
      os<<eigenvector<<endl;
    }
  os<<means<<endl;
  os<<sigmas<<endl;
}


