/****************************************************************
 Individual.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "Individual.H"
#include <iostream>
#include "IndividualSchema.H"
#include "Tableau.H"
#include "BOOM/Stack.H"
using namespace std;
using namespace BOOM;


Individual::Individual()
  : trees(IndividualSchema::getGlobal().getNumADFs()+
	  IndividualSchema::getGlobal().getNumEntryPoints()),
    callCount(IndividualSchema::getGlobal().getMaxCallCount()),
    height(0)
{
  // ctor
}



Individual::~Individual()
{
  // dtor

  int i, n=trees.size();
  for(i=0 ; i<n ; ++i)
    {
      delete trees[i];
    }
}



void Individual::decrementCallCount()
{
  --callCount;
}



bool Individual::canCall()
{
  return callCount>0;
}



void Individual::resetCallCount()
{
  callCount=IndividualSchema::getGlobal().getMaxCallCount();
}



TreeNode *&Individual::getADF(int adfNum)
{
  return trees[adfNum];
}



TreeNode *&Individual::getEntryPoint(int entryPointNum)
{
  return trees[IndividualSchema::getGlobal().getNumADFs()+entryPointNum];
}



int Individual::getNumTrees() const
{
  return trees.size();
}



TreeNode *Individual::getTree(int i)
{
  return trees[i];
}



Individual *Individual::clone() const
{
  Individual *twin=new Individual();
  twin->height=height;

  int i, n=trees.size();
  for(i=0 ; i<n ; ++i)
    twin->trees[i]=trees[i]->deepCopy();

  twin->fitness=fitness;

  return twin;
}



void Individual::printOn(ostream &os) const
{
  int numADFs=IndividualSchema::getGlobal().getNumADFs();

  int n=trees.size();
  for(int i=0 ; i<n ; ++i)
    {
      TreeNode *tree=trees[i];
      if(i<numADFs)
	os << "double ADF" << i << "() { return " << *tree << ";}"<<endl;
      else if(i-numADFs>0)
	os << "double mainFunction" << (i-numADFs) << "() { return "
	   << *tree << ";}" << endl;
      else
	os << "double mainFunction() { return " << *tree << ";}"<<endl;
      //os << "[" << *tree << "]" << endl;
    }
}



ostream &operator<<(ostream &os,const Individual &individual)
{
  individual.printOn(os);
  return os;
}



int Individual::getCachedHeight()
{
  return height;
}



int Individual::getTreeHeight()
{
  int maxHeight=0;
  int i, n=trees.size();
  for(i=0 ; i<n ; ++i)
    {
      int height=trees[i]->getTreeHeight();
      if(height>maxHeight) maxHeight=height;
    }
  height=maxHeight;
  return maxHeight;
}



void Individual::save(ostream &os) const
{
  // Save the trees!

  int i, n=trees.size();
  os << n << endl;
  for(i=0 ; i<n ; ++i)
    {
      os << trees[i]->countNodes() << endl;
      trees[i]->save(os);
    }
}



void Individual::load(istream &is)
{
  int i, numTrees;
  is >> numTrees;
  //cout << "numTrees=" << numTrees << endl; // ###

  for(i=0 ; i<numTrees ; ++i)
    {
      //cout << "loading tree #" << i << " of " << numTrees << endl;//###
      trees[i]=loadTree(is);
      //cout << "tree loaded." << endl; // ###
    }
}



TreeNode *Individual::loadTree(istream &is)
{
  Stack<TreeNode *> s;

  int nodesToRead;
  is >> nodesToRead;
  while(nodesToRead)
    {
      // Read the next node's definition from the stream
      //cout << "\tread next node's definition" << endl; // ###
      NodeType nodeType;
      int iNodeType, arity, numChildren;
      is >> iNodeType >> arity >> numChildren;
      nodeType=(NodeType) iNodeType;

      // Allocate the node object
      //cout << "\tallocate node object" << endl; // ###
      TreeNode *node=createNode(nodeType,arity,is);

      // Pop its children from the stack
      //cout << "\tpop " << numChildren << " children" << endl; // ###
      int i;
      Stack<TreeNode *> children;
      for(i=0 ; i<numChildren ; ++i)
	{
	  //cout << "\tpop!" << endl; // ###
	  children.push(s.top());
	  s.pop();
	}
      for(i=0 ; i<numChildren ; ++i)
	{
	  //cout << "\taddChild()" << endl; // ###
	  node->addChild(children.top());
	  children.pop();
	}

      // Push the new node onto the stack
      //cout << "\tpush new node" << endl;//###
      s.push(node);
      
      --nodesToRead;
    }

  // The last remaining node on the stack is the resulting tree
  //cout << "return stack top" << endl; // ###
  return s.top();
}



TreeNode *Individual::createNode(NodeType nodeType,int arity,istream &is)
{
  switch(nodeType)
    {
    case TN_MIN: 
      return new MinNode();
    case TN_MAX: 
      return new MaxNode();
    case TN_FMOD: 
      return new FmodNode();
    case TN_ABS: 
      return new AbsNode();
    case TN_CEIL: 
      return new CeilNode();
    case TN_HYPOT: 
      return new HypotNode();
    case TN_TANH: 
      return new TanhNode();
    case TN_SINH: 
      return new SinhNode();
    case TN_COSH: 
      return new CoshNode();
    case TN_ATANH: 
      return new AtanhNode();
    case TN_ASINH: 
      return new AsinhNode();
    case TN_ACOSH: 
      return new AcoshNode();
    case TN_LOG2: 
      return new Log2Node();
    case TN_LOG10: 
      return new Log10Node();
    case TN_CBRT: 
      return new CbrtNode();
    case TN_SQRT: 
      return new SqrtNode();
    case TN_ATAN: 
      return new AtanNode();
    case TN_ASIN: 
      return new AsinNode();
    case TN_ACOS: 
      return new AcosNode();
    case TN_TAN: 
      return new TanNode();
    case TN_POW: 
      return new PowNode();
    case TN_MOD: 
      return new ModNode();
    case TN_AND: 
      return new AndNode();
    case TN_OR: 
      return new OrNode();
    case TN_NOTEQUALS: 
      return new NotEqualsNode();
    case TN_GREATEREQUALS: 
      return new GreaterEqualsNode();
    case TN_LESSEQUALS: 
      return new LessEqualsNode();
    case TN_NOT: 
      return new NotNode();
    case TN_LESS: 
      return new LessNode();
    case TN_GREATER: 
      return new GreaterNode();
    case TN_PLUS: 
      return new PlusNode();
    case TN_MINUS: 
      return new MinusNode();
    case TN_TIMES: 
      return new TimesNode();
    case TN_DIVIDE: 
      return new DivideNode();
    case TN_SIN: 
      return new SinNode();
    case TN_COS: 
      return new CosNode();
    case TN_EXP: 
      return new ExpNode();
    case TN_CONSTANT: 
      {
	float value;
	is >> value;
	return new ConstantNode(value);
      }
    case TN_VARIABLE: 
      {
	int whichParm;
	is >> whichParm;
	return new VariableNode(whichParm);
      }
    case TN_IF: 
      return new IfNode();
    case TN_CALL: 
      {
	int whichADF;
	is >> whichADF;
	return new CallNode(whichADF,arity);
      }
    case TN_LOG: 
      return new LogNode();
    case TN_EQUALS: 
      return new EqualsNode();
    default:
      throw string("unknown node type in Individual::createNode()");
    }
}


