/****************************************************************
 anneal.C
 william.majoros@duke.edu

 Simulated annealing.

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "Framework/Tableau.H"
#include "Framework/IndividualSchema.H"
#include "Framework/SymbolSet.H"
#include "Framework/CrosserOver.H"
#include "Framework/RouletteWheel.H"
#include "Framework/Mutator.H"
#include "Framework/SubtreeMutator.H"
#include "Framework/ClassifiedAR.H"
#include "Framework/PopulationBuilder.H"
#include "Framework/FitnessNormalizer.H"
#include "Framework/Stats.H"
#include "Framework/Tournament.H"
#include "../common/NamesFile.H"
#include "../common/TrainingSet.H"
#include "BOOM/StrTokenizer.H"
#include "BOOM/Random.H"
#include "BOOM/Time.H"
#include "BOOM/CommandLine.H"
#include "BOOM/ConfigFile.H"
#include <iostream>
#include <math.h>
#include <stdlib.h>
#include <fstream>
using namespace std;
using namespace BOOM;

int numAttributes;
NamesFile namesFile;


class Application
{
  ConfigFile *configFile;
  String configFilename, trainFile, outfile, namesFilename;
  SymbolSet symbolSet;
  IndividualBuilder *individualBuilder;
  Mutator *pointMutator;
  SubtreeMutator *subtreeMutator;
  Vector<ClassifiedAR*> trainingSet;
  int numCategories;
  int initialHeight, maxHeight;
  Individual *bestOfRun;
  double temperature, temperatureDecay, finalTemperature;
  double K;

  void processConfig();
  void setup();
  void setupSchema();
  void setupSymbols();
  void loadTrainingSet();
  void convertToARs(TrainingSet *,Vector<ClassifiedAR*> &);
  void simulatedAnnealing();
  void exercise(Individual &,Vector<ClassifiedAR*> &ARs,double proportion);
  void finish();
  bool boltzmann(double deltaFitness);
public:
  int go(int argc,char *argv[]);
};


int main(int argc,char *argv[])
{
  try
    {
      Application app;
      return app.go(argc,argv);
    }
  catch(const String &msg)
    {
      cerr << msg.c_str() << endl;
      return -1;
    }
  catch(const char *msg)
    {
      cerr << msg << endl;
      return -1;
    }
  catch(...)
    {
      cerr<<"unknown exception caught in main()"<<endl;
      return -1;
    }
}



int Application::go(int argc,char *argv[])
{
  // Process the command line
  CommandLine cmd(argc,argv,"tc:");
  if(cmd.numArgs()!=4)
    throw "\nanneal <*.config> <*.names> <train> <outfile>\n\n";
  configFilename=cmd.arg(0);
  namesFilename=cmd.arg(1);
  trainFile=cmd.arg(2);
  outfile=cmd.arg(3);

  // Various initialization
  setup();

  // Perform annealing
  simulatedAnnealing();

  // Report results
  finish();
}



void Application::setup()
{
  // Process the configuration file
  processConfig();
  Tableau &tableau=Tableau::getGlobal();
  initialHeight=tableau.getInitialHeight();
  maxHeight=tableau.getMaxHeight();
  if(maxHeight<initialHeight) throw "max-height < initial-height !!!";
  temperature=tableau.getInitialTemperature();
  temperatureDecay=tableau.getTemperatureDecay();
  finalTemperature=tableau.getFinalTemperature();
  K=tableau.getK();

  // Create schema describing individuals
  setupSchema();

  // Create terminal & nonterminal sets
  setupSymbols();
  individualBuilder=new IndividualBuilder(symbolSet);

  // Load training & test sets
  loadTrainingSet();

  // Other misc. initialization
  pointMutator=new Mutator(symbolSet);
  subtreeMutator=new SubtreeMutator(individualBuilder->getTreeBuilder());
}



void Application::loadTrainingSet()
{
  Schema *schema=namesFile.processFile(namesFilename);
  numCategories=schema->getNumCategories();
  numAttributes=schema->numAttributes();

  TrainingSet *train=new TrainingSet(*schema,trainFile    );
  convertToARs(train,trainingSet);
  delete train;
}



void Application::convertToARs(TrainingSet *ts,Vector<ClassifiedAR*> &ARs)
{
  int i, n=ts->numCases();
  for(i=0 ; i<n ; ++i)
    {
      TrainingCase &tc=ts->getCase(i);
      ARs.push_back(new ClassifiedAR(tc));
    }
}



void Application::setupSchema()
{
  // Create schema describing individuals

  Tableau &tableau=Tableau::getGlobal();
  IndividualSchema::allocateGlobal(0);
  IndividualSchema::getGlobal().addEntryPoint(0);
}



void Application::simulatedAnnealing()
{
  Time stopwatch;
  stopwatch.startCounting();
  Tableau &tableau=Tableau::getGlobal();
  double trainPortion=tableau.getTrainPortion();
  Individual *current=individualBuilder->buildIndividual(initialHeight);
  exercise(*current,trainingSet,trainPortion);
  bestOfRun=current->clone();
  const double probSubtreeMutation=tableau.getSubtreeMutation();
  while(temperature>finalTemperature)
    {
      // Evaluate fitness of current individual
      cout<<"T="<<temperature<<", accuracy="<<current->getFitness()<<endl;

      // See if a new bestOfRun has been produced
      if(current->getFitness() > bestOfRun->getFitness())
	{
	  delete bestOfRun;
	  bestOfRun=current->clone();
	}

      // Generate a new mutant via either point mutation or subtree mutation
      // (but subject to tree-height constraints)
      Individual *mutant=NULL;
      while(!mutant || mutant->getTreeHeight()>maxHeight)
	if(Random0to1()<=probSubtreeMutation)
	  mutant=subtreeMutator->mutate(current);
	else
	  mutant=pointMutator->mutate(current);

      // Accept mutant if it's superior, or (with some probability, 
      // according to Boltzmann distribution) even if it's inferior
      exercise(*mutant,trainingSet,trainPortion);
      double deltaFitness=mutant->getFitness()-current->getFitness();
      if(deltaFitness>0 || boltzmann(deltaFitness))
	{
	  delete current;
	  current=mutant;
	}
      else delete mutant;

      // Decrease the temperature
      temperature*=temperatureDecay;
    }
  stopwatch.stopCounting();
  cout << stopwatch.elapsedTime() << endl;
}



void Application::processConfig()
{  
  configFile=new ConfigFile(configFilename);
  Tableau::processConfig(*configFile);

  int seed=Tableau::getGlobal().getSeed();
  if(seed) SeedRandomizer(unsigned(seed));
  else randomize();
}



void Application::setupSymbols()
{
  // Nonterminals
  String symbols=Tableau::getGlobal().getNonterminals();
  StrTokenizer tokenizer(symbols,",");
  while(tokenizer.hasMoreTokens())
    {
      const char *token=tokenizer.nextToken();
      if(!strcmp(token,"+")) 
	symbolSet.addNonterminal(new PlusNode());
      else if(!strcmp(token,"-")) 
	symbolSet.addNonterminal(new MinusNode());
      else if(!strcmp(token,"*")) 
	symbolSet.addNonterminal(new TimesNode());
      else if(!strcmp(token,"/")) 
	symbolSet.addNonterminal(new DivideNode());
      else if(!strcmp(token,"sin")) 
	symbolSet.addNonterminal(new SinNode());
      else if(!strcmp(token,"cos")) 
	symbolSet.addNonterminal(new CosNode());
      else if(!strcmp(token,"exp")) 
	symbolSet.addNonterminal(new ExpNode());
      else if(!strcmp(token,"log")) 
	symbolSet.addNonterminal(new LogNode());
      else if(!strcmp(token,"if")) 
	symbolSet.addNonterminal(new IfNode());
      else if(!strcmp(token,"min")) 
	symbolSet.addNonterminal(new MinNode());
      else if(!strcmp(token,"max")) 
	symbolSet.addNonterminal(new MaxNode());
      else if(!strcmp(token,"fmod")) 
	symbolSet.addNonterminal(new FmodNode());
      else if(!strcmp(token,"abs")) 
	symbolSet.addNonterminal(new AbsNode());
      else if(!strcmp(token,"ceil")) 
	symbolSet.addNonterminal(new CeilNode());
      else if(!strcmp(token,"hypot")) 
	symbolSet.addNonterminal(new HypotNode());
      else if(!strcmp(token,"tanh")) 
	symbolSet.addNonterminal(new TanhNode());
      else if(!strcmp(token,"sinh")) 
	symbolSet.addNonterminal(new SinhNode());
      else if(!strcmp(token,"cosh")) 
	symbolSet.addNonterminal(new CoshNode());
      else if(!strcmp(token,"atanh")) 
	symbolSet.addNonterminal(new AtanhNode());
      else if(!strcmp(token,"asinh")) 
	symbolSet.addNonterminal(new AsinhNode());
      else if(!strcmp(token,"acosh")) 
	symbolSet.addNonterminal(new AcoshNode());
      else if(!strcmp(token,"log2")) 
	symbolSet.addNonterminal(new Log2Node());
      else if(!strcmp(token,"log10")) 
	symbolSet.addNonterminal(new Log10Node());
      else if(!strcmp(token,"cbrt")) 
	symbolSet.addNonterminal(new CbrtNode());
      else if(!strcmp(token,"sqrt")) 
	symbolSet.addNonterminal(new CallNode());
      else if(!strcmp(token,"atan")) 
	symbolSet.addNonterminal(new AtanNode());
      else if(!strcmp(token,"asin")) 
	symbolSet.addNonterminal(new AsinNode());
      else if(!strcmp(token,"acos")) 
	symbolSet.addNonterminal(new AcosNode());
      else if(!strcmp(token,"tan")) 
	symbolSet.addNonterminal(new TanNode());
      else if(!strcmp(token,"^")) 
	symbolSet.addNonterminal(new PowNode());
      else if(!strcmp(token,"%")) 
	symbolSet.addNonterminal(new ModNode());
      else if(!strcmp(token,"&&")) 
	symbolSet.addNonterminal(new AndNode());
      else if(!strcmp(token,"||")) 
	symbolSet.addNonterminal(new OrNode());
      else if(!strcmp(token,"!=")) 
	symbolSet.addNonterminal(new NotEqualsNode());
      else if(!strcmp(token,">=")) 
	symbolSet.addNonterminal(new GreaterEqualsNode());
      else if(!strcmp(token,"<=")) 
	symbolSet.addNonterminal(new LessEqualsNode());
      else if(!strcmp(token,"!")) 
	symbolSet.addNonterminal(new NotNode());
      else if(!strcmp(token,"<")) 
	symbolSet.addNonterminal(new LessNode());
      else if(!strcmp(token,">")) 
	symbolSet.addNonterminal(new GreaterNode());
      else if(!strcmp(token,">0")) 
	symbolSet.addNonterminal(new GreaterThanZeroNode());
      else if(!strcmp(token,">1")) 
	symbolSet.addNonterminal(new GreaterThanOneNode());
      else if(!strcmp(token,"=0")) 
	symbolSet.addNonterminal(new EqualsZeroNode());
      else
	throw string("nonterminal symbol \"")+token+"\" is not recognized";
    }
  
  // Terminals
  const double minConst=Tableau::getGlobal().getMinConst();
  const double maxConst=Tableau::getGlobal().getMaxConst();
  symbols=Tableau::getGlobal().getTerminals();
  tokenizer.tokenizeThis(symbols.c_str());
  while(tokenizer.hasMoreTokens())
    {
      const char *token=tokenizer.nextToken();
      if(!strcmp(token,"const")) 
	symbolSet.addTerminal(new ConstantNode(minConst,maxConst));
      else if(!strcmp(token,"var")) 
	symbolSet.addTerminal(new VariableNode());
      else
	throw string("terminal symbol \"")+token+"\" is not recognized";
    }
}



void Application::finish()
{
  double fitness=bestOfRun->getFitness();
  double accuracy=int(1000*fitness+5/9.0)/10.0;
  cout << "final accuracy on training set: " << accuracy << "%" << endl;
  ofstream os(outfile.c_str());
  os << *bestOfRun << endl;
}



void Application::exercise(Individual &ind,Vector<ClassifiedAR*> &ARs,
			   double proportion)
{
  // Get the result-producing branch, so we can call it
  TreeNode *tree=ind.getEntryPoint(0);

  // Iterate through the test cases
  int correct=0, n=ARs.size(), num=0;
  for(int i=0 ; i<n ; ++i)
    {
      // Exercise only on a random sample of cases
      if(Random0to1()>=proportion) continue;

      // Get the next test case
      ClassifiedAR *ar=ARs[i];

      // Call the evolved procedure
      double value=tree->eval(ar->getAR(),ind);

      // Discretize the result into a category
      int category=int(value+5/9.0);
      if(category<0) category=0;
      else if(category>numCategories-1) 
		  category=numCategories-1;

      // Score the result
      int actualCategory=ar->getCategory();
      if(category==actualCategory) ++correct;
      ++num;
    }

  double accuracy=correct/double(num);
  ind.setFitness(accuracy);
}


 
bool Application::boltzmann(double deltaFitness)
{
  return Random0to1()<=exp(deltaFitness/(K*temperature));
}
