/****************************************************************
 train-bayes-net.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include <iostream>
#include <fstream>
#include "BOOM/String.H"
#include "BOOM/CommandLine.H"
#include "BOOM/EdgeWeightsMatrix.H"
#include "BOOM/MinimalSpanningTree.H"
#include "BOOM/AdjacencyGraph.H"
#include "BOOM/Chi2IndepTest.H"
#include "../common/Schema.H"
#include "../common/TrainingSet.H"
#include "../common/NamesFile.H"
#include "BayesNetClassifier.H"
#include "Discretizer.H"
using namespace std;
using namespace BOOM;


int NUM_BUCKETS; // for discretization of continuous attributes


class Application
{
public:
  Application();
  int main(int argc,char *argv[]);
private:
  NamesFile namesFile;
  Schema *schema;
  int numCases, numAttributes, numCategories;
  TrainingSet *trainingData;
  Chi2Table chiTable;
  Array1D<Discretizer*> discretizers;
  Array1D<TrainingSet*> casesByCategory;
  BayesNetClassifier *trainClassifier();
  BayesNetwork *trainNetwork(int category);
  void allocateDiscretizers();
  void computeEdgeWeights(EdgeWeightsMatrix &,int category);
  void fillContingencyTable(int row,int col,Array2D<int> &table,
			    int category);
  void getPredictorStats(int i,float &min,float &max);
  int numDiscreteValues(int attribute);
  void computeNetwork(BayesNetwork &,Graph &mst,int category,int root);
  void computeNetwork(BayesNetwork &,Graph &mst,int parent,int child,
		      int category);
  void estimateProbabilities(BayesNetwork &,int parent,int child,
			     int category);
};


int main(int argc,char *argv[])
  {
    try
      {
	Application app;
	return app.main(argc,argv);
      }
    catch(const char *p)
      {
	cerr << p << endl;
      }
    catch(const string &msg)
      {
	cerr << msg.c_str() << endl;
      }
    catch(const exception &e)
      {
	cerr << "STL exception caught in main:\n" << e.what() << endl;
      }
    catch(...)
      {
	cerr << "Unknown exception caught in main" << endl;
      }
    return -1;
  }



Application::Application()
  {
    // ctor
  }



int Application::main(int argc,char *argv[])
  {
    // Process command line
    CommandLine cmd(argc,argv,"");
    if(cmd.numArgs()!=4)
      throw String(
"train-bayes-net <*.names> <*.data> <outfile> <#discretization-bins>");
    String namesFilename=cmd.arg(0);
    String trainFile=cmd.arg(1);
    String outfile=cmd.arg(2);
    NUM_BUCKETS=cmd.arg(3).asInt();
    
    // Load configuration
    schema=namesFile.processFile(namesFilename);
    numAttributes=schema->numAttributes();
    numCategories=schema->getNumCategories();

    // Load training data & split by category
    trainingData=new TrainingSet(*schema,trainFile,true);
    numCases=trainingData->numCases();
    casesByCategory.resize(numCategories);
    for(int i=0 ; i<numCategories ; ++i)
      casesByCategory[i]=new TrainingSet(*schema);
    trainingData->splitByCategory(casesByCategory);

    // Tabluate numbers of discrete values for discrete variables,
    // and allocate discretizers for the continuous attributes
    delete [] schema->countDiscrete(*trainingData);
    allocateDiscretizers();
    
    // Build model
    BayesNetClassifier *classifier=trainClassifier();
    
    // Save model into file
    ofstream os(outfile.c_str());
    classifier->save(os);

    // Test on the training set
    int correct=0;
    for(int i=0 ; i<numCases ; ++i)
      {
	TrainingCase &c=trainingData->getCase(i);
	int expected=c.getCategory();
	int computed=classifier->classify(c);
	if(expected==computed) ++correct;
      }
    float accuracy=int(100*correct/float(numCases)+5/9.0);
    cout<<"Accuracy on the training set: "<<accuracy<<"%"<<endl;
    
    return 0;
  }



BayesNetClassifier *Application::trainClassifier()
{
  BayesNetClassifier *classifier=
    new BayesNetClassifier(numAttributes,numCategories);
  classifier->estimatePriors(*trainingData);
  for(int i=0 ; i<numCategories ; ++i)
    {
      BayesNetwork *net=trainNetwork(i);
      classifier->setNetwork(i,net);
    }
  return classifier;
}



BayesNetwork *Application::trainNetwork(int category)
{
  // First, compute a matrix of edge weights, where each edge weight
  // is the chi-squared statistic from a test-of-independence
  EdgeWeightsMatrix E(numAttributes);
  computeEdgeWeights(E,category);
  
  // Now extract a maximimal spanning tree (MST)
  AdjacencyGraph G(numAttributes);
  for(int i=0 ; i<numAttributes ; ++i)
    for(int j=i+1 ; j<numAttributes ; ++j)
      G.addEdge(i,j);
  MinimalSpanningTree MST(G,E);
  Graph *mst=MST.getTree();

  // Convert the tree to a BayesNetwork
  BayesNetwork *net=new BayesNetwork(discretizers);
  int root=0; // ### arbitrary
  computeNetwork(*net,*mst,category,root);

  // Return the network
  return net;
}



void Application::computeNetwork(BayesNetwork &net,Graph &mst,int category,
				 int root)
{
  // Compute the root specially, since it's not a conditional probability
  ConditionalDensity &table=net.getDensity(root);
  int levels=numDiscreteValues(root);
  table.resize(levels,1);
  table.setAllTo(0.0);
  Discretizer *discretizer=discretizers[root];
  TrainingSet *cases=casesByCategory[category];
  int n=cases->numCases();
  for(int i=0 ; i<n ; ++i)
    {
      Case &c=cases->getCase(i);
      int value=Discretizer::getDiscrete(c,root,discretizer);
      table[value][0]+=1.0;
    }
  for(int i=0 ; i<levels ; ++i)
    table[i][0]/=n;

  // Now recurse through the tree to compute the rest of the probabilities
  computeNetwork(net,mst,-1,root,category);
}



void Application::computeNetwork(BayesNetwork &net,Graph &mst,int parent,
				 int child,int category)
{
  bool shouldDelete;
  Neighborhood *neighbors=mst.getNeighborsOf(child,shouldDelete);
  BOOM::Set<VertexId>::iterator cur=neighbors->begin(), 
    end=neighbors->end();
  for(; cur!=end ; ++cur)
    {
      int neighbor=*cur;
      if(neighbor==parent) continue;
      net.addEdge(child,neighbor);
      estimateProbabilities(net,child,neighbor,category);
      computeNetwork(net,mst,child,neighbor,category);
    }
  if(shouldDelete) delete neighbors;
}



void Application::estimateProbabilities(BayesNetwork &net,int parent,
					int child,int category)
{
  ConditionalDensity &table=net.getDensity(child);
  int childLevels=numDiscreteValues(child);
  int parentLevels=numDiscreteValues(parent);
  table.resize(childLevels,parentLevels);
  table.setAllTo(0.0);
  Discretizer *parentDiscretizer=discretizers[parent];
  Discretizer *childDiscretizer=discretizers[child];
  TrainingSet *cases=casesByCategory[category];
  int n=cases->numCases();
  for(int i=0 ; i<n ; ++i)
    {
      Case &c=cases->getCase(i);
      int childValue=Discretizer::getDiscrete(c,child,childDiscretizer);
      int parentValue=Discretizer::getDiscrete(c,parent,parentDiscretizer);
      table[childValue][parentValue]+=1.0;
    }
  for(int i=0 ; i<parentLevels ; ++i)
    {
      double sum=0.0;
      for(int j=0 ; j<childLevels ; ++j)
	sum+=table[j][i];
      for(int j=0 ; j<parentLevels ; ++j)
	table[j][i]/=sum;
    }
}



void Application::computeEdgeWeights(EdgeWeightsMatrix &e,int category)
{
  Array2D<int> contingencyTable;
  for(int i=0 ; i<numAttributes-1 ; ++i)
    for(int j=i+1 ; j<numAttributes ; ++j)
      {
	fillContingencyTable(i,j,contingencyTable,category);
	Chi2IndepTest X(contingencyTable,chiTable);
	double weight;
	/*
	if(X.getP()<=0.05) 
	  weight=-X.getChiSquared(); // NEGATIVE : want MAX spanning tree!
	else 
	  weight=0;
	*/
	weight=-X.getChiSquared();
	e.setWeight(i,j,weight);
      }
}



void Application::fillContingencyTable(int row,int col,Array2D<int> &table,
				       int category)
{
  TrainingSet *cases=casesByCategory[category];
  int n=cases->numCases();
  Discretizer *rowDiscretizer=discretizers[row];
  Discretizer *colDiscretizer=discretizers[col];
  int rowBuckets=numDiscreteValues(row);
  int colBuckets=numDiscreteValues(col);
  table.resize(rowBuckets,colBuckets);
  table.setAllTo(0);
  for(int i=0 ; i<n ; ++i)
    {
      TrainingCase &c=cases->getCase(i);
      int rowVal=Discretizer::getDiscrete(c,row,rowDiscretizer);
      int colVal=Discretizer::getDiscrete(c,col,colDiscretizer);
      ++table[rowVal][colVal];
    }
}



void Application::allocateDiscretizers()
{
  discretizers.resize(numAttributes);
  float minValue, maxValue;
  for(int i=0 ; i<numAttributes ; ++i)
    if((*schema)[i]->getType()==DISCRETE)
      discretizers[i]=NULL;
    else
      {
	getPredictorStats(i,minValue,maxValue);
	const float range=maxValue-minValue;
	const int numBuckets=NUM_BUCKETS;
	const float bucketSize=range/numBuckets;
	discretizers[i]=
	  new Discretizer(minValue,maxValue,bucketSize,NUM_BUCKETS);
      }
}



void Application::getPredictorStats(int i,float &min,float &max)
{
  // Initialize min and max to values of first case
  Case &firstCase=trainingData->getCase(0);
  min=max=firstCase.getContinuous(i);
  
  // Iterate through all cases, noting values less than min or
  // greater than max
  int n=trainingData->numCases();
  for(int j=0 ; j<n ; ++j)
    {
      Case &c=trainingData->getCase(j);
      float value=c.getContinuous(i);
      if(value<min)
	min=value;
      else if(value>max)
	max=value;
    }
}



int Application::numDiscreteValues(int attribute)
{
  Discretizer *discretizer=discretizers[attribute];
  if(discretizer) return discretizer->getNumBuckets();
  Discrete *descriptor=static_cast<Discrete*>((*schema)[attribute]);
  return descriptor->getN();
}



