/****************************************************************
 ModelBuilder.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "ModelBuilder.H"
using namespace BOOM;

extern int NUM_BUCKETS;


ModelBuilder::ModelBuilder(TrainingSet &dataSet,Schema &schema)
  : dataSet(dataSet), 
    dataSetStats(dataSet), 
    schema(schema),
    numCategories(0), 
    numPredictors(schema.numAttributes())
{
  // ctor
}



Model *ModelBuilder::buildModel()
{
  PublicModel *publicModel=buildPublicModel();
  return new Model(*publicModel,schema);
}



PublicModel *ModelBuilder::buildPublicModel()
{
  // Allocate a model object
  Map<int,int> *category2index=dataSetStats.getX2Index();
  numCategories=category2index->size();
  PublicModel *model=
    new PublicModel(numPredictors,numCategories,*category2index);

  // Identify continuous variables and allocate discretizers
  allocateDiscretizers(model);

  // Iterate through training cases, updating xProportions and
  // yStructure along the way
  processTrainingCases(model);

  return model;
}



void ModelBuilder::allocateDiscretizers(PublicModel *model)
{
  for(int i=0 ; i<numPredictors ; ++i)
    if(!(schema[i]->getType()==DISCRETE))
      {
	float minValue, maxValue;
	dataSetStats.getPredictorStats(i,minValue,maxValue);
	const float range=maxValue-minValue;
	const int numBuckets=NUM_BUCKETS;
	const float bucketSize=range/numBuckets;
	model->yStructure.allocateDiscretizer(i,minValue,
					      bucketSize);
      }
}



void ModelBuilder::processTrainingCases(PublicModel *model)
{
  const int numCases=dataSet.numCases();
  for(int i=0 ; i<numCases ; ++i)
    {
      // Get next case
      TrainingCase &c=dataSet.getCase(i);
      
      // Process category
      int category=c.getCategory();
      const int categoryIndex=model->x2index[category];
      model->xProportions[categoryIndex]+=1.0;

      // Process predictors
      processPredictors(c,model,categoryIndex);
    }

  // Divide elements of yStructure by sizes of 
  // respective categories
  divideYStructure(model);

  // Divide xProportions by # categories
  divideCategoriesBy(model,numCases);
}



void ModelBuilder::processPredictors(TrainingCase &c,PublicModel *model,
				     int categoryIndex)
{
  for(int i=0 ; i<numPredictors ; ++i)
    {
      int discreteValue;
      switch(schema[i]->getType())
	{
	case CONTINUOUS:
	  {
	    Discretizer &d=model->yStructure[i].getDiscretizer();
	    float value=c.getContinuous(i);
	    discreteValue=d.discretize(value);
	    break;
	  }
	case DISCRETE:
	  discreteValue=c.getDiscrete(i);
	  break;
	}
      model->yStructure.lookup(i,discreteValue,categoryIndex)+=1.0;
    }
}



void ModelBuilder::divideYStructure(PublicModel *model)
{
  // Outer loop: iterate through predictor variables
  int i;
  for(i=0 ; i<numPredictors ; ++i)
    {
      YStructureElement &elem=model->yStructure[i];
      Map<int,FloatArray1D*>::iterator cur=elem.getIterator(),
	end=elem.getIteratorEnd();

      // Middle loop: iterate through values of this 
      // predictor variable
      for(; cur!=end ; ++cur)
	{
	  FloatArray1D &vect=*(*cur).second;
	  
	  // Inner loop: iterate through categories in which this
	  // predictor value occurs
	  int j;
	  for(j=0 ; j<numCategories ; ++j)
	    {
	      // ### WARNING: assigning float to int:
	      const int casesInCategory=(int) model->xProportions[j];
	      vect[j]/=casesInCategory;
	    }
	}
    }
}



void ModelBuilder::divideCategoriesBy(PublicModel *model,int numCases)
{
  const int n=model->xProportions.size();
  for(int i=0 ; i<n ; ++i)
    model->xProportions[i]/=numCases;
}




