/****************************************************************
 AttrRangeTable.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "AttrRangeTable.H"
#include <iostream>
#include "BOOM/Map.H"
using namespace BOOM;


AttrRangeTable::AttrRangeTable(TrainingSet &t,int granularity)
  : trainingSet(t), schema(t.getSchema()),
    sets(t.getSchema().numAttributes()),
    arrays(t.getSchema().numAttributes()),
    granularity(granularity)
{
  // ctor

  computeTable();
}



void AttrRangeTable::computeTable()
{
  /*
    First, we compute an array of sets.  There is a set 
    for each attribute.  We will iterate through all 
    training cases and for each attribute of each case, 
    we will hash that attribute's value into the setfor 
    that attribute.  Thus, each value is stored only once.
   */
  initSets();
  computeSets();

  /*
    Next, we copy the elements from the sets into a normal 
    array, so they are easier to work with.
   */
  computeArrays();

  /*
    Now we eliminate some of the continuous values.
    There may be many more unique floats than ints,
    so we reduce the number of floats so there are
    no more of them than the average number of
    discrete values for any one attribute.
   */
  compressContinuous();
}



void AttrRangeTable::computeSets()
{
  int nAttributes=schema.numAttributes();
  int nCases=trainingSet.numCases();
  for(int i=0 ; i<nCases ; ++i)
    {
      TrainingCase &c=trainingSet.getCase(i);
      for(int j=0 ; j<nAttributes ; ++j)
	{
	  ValueSet &valueSet=*sets[j];
	  switch(schema[j]->getType())
	    {
	    case CONTINUOUS:
	      static_cast<DoubleSet&>(valueSet).values.
		insert(c.getContinuous(j));
	      break;
	    case DISCRETE:
	      static_cast<IntSet&>(valueSet).values.
		insert(c.getDiscrete(j));
	      break;
	    }
	}
    }
}



void AttrRangeTable::initSets()
{
  int i, n=schema.numAttributes();
  for(i=0 ; i<n ; ++i) 
    switch(schema[i]->getType())
      {
      case DISCRETE:   sets[i]=new IntSet;    break;
      case CONTINUOUS: sets[i]=new DoubleSet; break;
      }
}



void AttrRangeTable::computeArrays()
{
  int n=schema.numAttributes();
  for(int j=0 ; j<n ; ++j)
    {
      switch(schema[j]->getType())
	{
	case DISCRETE:
	  {
	    IntSet &valueSet=static_cast<IntSet&>(*sets[j]);
	    int numValues=valueSet.values.size();
	    IntArray *array=new IntArray(numValues);
	    Set<int>::iterator cur=valueSet.values.begin(), 
	      end=valueSet.values.end();
	    for(int i=0 ; i<numValues ; ++i, ++cur)
	      array->values[i]=*cur;
	    delete &valueSet;
	    arrays[j]=array;
	  }
	  break;
	case CONTINUOUS:
	  {
	    DoubleSet &valueSet=static_cast<DoubleSet&>(*sets[j]);
	    int numValues=valueSet.values.size();
	    DoubleArray *array=new DoubleArray(numValues);
	    Set<double>::iterator cur=valueSet.values.begin(), 
	      end=valueSet.values.end();
	    for(int i=0 ; i<numValues ; ++i, ++cur)
	      array->values[i]=*cur;
	    delete &valueSet;
	    arrays[j]=array;
	  }
	  break;
	}
    }
  sets.resize(0);
}



void AttrRangeTable::compressContinuous()
{
  /*
    Since there can be very many unique values
    of a continuous attribute, this method
    samples them so there are on average the 
    same number as in a discrete attribute.
   */

  int aveSize=aveNumDiscreteVals();
  if(aveSize<granularity) aveSize=granularity;

  int i, n=schema.numAttributes();
  for(i=0 ; i<n ; ++i)
    if(schema[i]->getType()==CONTINUOUS)
      {
	DoubleArray &array=static_cast<DoubleArray&>(*arrays[i]);
	int originalSize=array.values.size();
	if(originalSize>aveSize)
	  {
	    arrays[i]=
	      compressContinuous(array,originalSize,aveSize);
	    delete &array;
	  }
      }
}



int AttrRangeTable::aveNumDiscreteVals()
{
  /*
    This method counts the number of values
    for each discrete attribute, and returns
    the average
   */

  int sum=0;
  int i, n=schema.numAttributes(), denom=0;
  for(i=0 ; i<n ; ++i)
    if(schema[i]->getType()==DISCRETE)
      {
	ValueArray &array=*arrays[i];
	sum+=array.getSize();
	++denom;
      }

  return int(sum/float(denom));
}



DoubleArray *AttrRangeTable::compressContinuous(DoubleArray &array,
					       int size,
					       int newSize)
{
  DoubleArray &newArray=*new DoubleArray(newSize);
  const float p=newSize/float(size);
  float accum=0.0;
  int j=0;
  for(int i=0 ; i<size ; ++i)
    {
      accum+=p;
      if(accum>=1.0)
	{
	  newArray.values[j++]=array.values[i];
	  accum-=1.0;
	}
    }

  if(j<newSize)
    newArray.values[j]=array.values[size-1];

  return &newArray;
}



Schema &AttrRangeTable::getSchema()
{
  return schema;
}



Array1D<ValueArray*> &AttrRangeTable::getArrays()
{
  return arrays;
}



