/****************************************************************
 TrainingSet.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "TrainingSet.H"
#include <iostream>
#include <fstream>
#include <math.h>
#include "BOOM/StrTokenizer.H"
#include <stdlib.h>
using namespace BOOM;


TrainingSet::TrainingSet(Schema &schema,String filename,
			 bool requireCategory)
  : schema(schema)
{
  // ctor

  load(filename,requireCategory);
}



TrainingSet::TrainingSet(Schema &schema)
  : schema(schema)
{
  // ctor
}



void TrainingSet::load(String &filename,bool requireCategory)
{
  int numAttributes=schema.numAttributes();

  ifstream is(filename.c_str());
  if(!is.good()) throw String("can't open file \"")+filename+"\"";
  while(!is.eof())
    {
      // Load line containing next training case
      String line;
      getline(is,line);
      if(line.length()==0 || is.eof()) break;

      // Parse attribute values
      TrainingCase *newCase=new TrainingCase(numAttributes);
      StrTokenizer tokenizer(line," \t\n\r");
      for(int i=0 ; i<numAttributes ; ++i)
	{
	  if(!tokenizer.hasMoreTokens())
	    throw String("Incomplete case");
	  const char *token=tokenizer.nextToken();

	  switch(schema[i]->getType())
	    {
	    case DISCRETE:
	      newCase->setDiscrete(i,atoi(token));
	      break;
	    case CONTINUOUS:
	      newCase->setContinuous(i,atof(token));
	      break;
	    }
	}

      // Parse the category
      if(tokenizer.hasMoreTokens())
	{
	  const char *token=tokenizer.nextToken();
	  newCase->setCategory(atoi(token));
	  if(tokenizer.hasMoreTokens())
	    throw String("Case has too many fields");
	}
      else if(requireCategory) throw String("Incomplete case");

      // Add training case to theVector
      theVector.push_back(newCase);
    }
}



TrainingCase &TrainingSet::getCase(int i)
{
  return *theVector[i];
}



int TrainingSet::numCases()
{
  return theVector.size();
}



Schema &TrainingSet::getSchema()
{
  return schema;
}



void TrainingSet::addCase(TrainingCase *newCase)
{
  theVector.push_back(newCase);
}



TrainingSet *TrainingSet::getCategory(int category)
{
  TrainingSet &newSet=*new TrainingSet(schema);

  int i, n=numCases();
  for(i=0 ; i<n ; ++i)
    {
      TrainingCase &thisCase=getCase(i);
      if(thisCase.getCategory()==category)
	newSet.addCase(&thisCase);
    }

  return &newSet;
}



TrainingSet *TrainingSet::getNotCategory(int category)
{
  TrainingSet &newSet=*new TrainingSet(schema);

  int i, n=numCases();
  for(i=0 ; i<n ; ++i)
    {
      TrainingCase &thisCase=getCase(i);
      if(thisCase.getCategory()!=category)
	newSet.addCase(&thisCase);
    }

  return &newSet;
}



void TrainingSet::deleteCases()
{
  int n=theVector.size();
  for(int i=0 ; i<n ; ++i)
    delete theVector[i];
}



void TrainingSet::splitByCategory(Array1D<TrainingSet*> &A)
{
  int n=theVector.size();
  for(int i=0 ; i<n ; ++i)
    {
      TrainingCase *c=theVector[i];
      int cat=c->getCategory();
      A[cat]->addCase(c);
    }
}


