/****************************************************************
 f-ratio.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include <iostream>
#include <stdlib.h>
#include <fstream>
#include <math.h>
#include <unistd.h>
#include "BOOM/Exceptions.H"
#include "BOOM/String.H"
#include "BOOM/Stack.H"
#include "BOOM/Stacktrace.H"
#include "BOOM/BitSet.H"
#include "BOOM/Vector.H"
#include "BOOM/VectorSorter.H"
#include "BOOM/Random.H"
#include "BOOM/Array1D.H"
#include "BOOM/Array2D.H"
#include "BOOM/CommandLine.H"
#include "BOOM/InversionMatrix.H"
#include "BOOM/GaussJordan.H"
#include "BOOM/Time.H"
#include "BOOM/BoundedSortedArray.H"
#include "BOOM/Constants.H"
#include "../common/NamesFile.H"
#include "../common/TestSet.H"
#include "../common/Schema.H"
#include "FeatureF.H"
using namespace std;
using namespace BOOM;

class Application
{
  int numCategories;
  String trainFile, nameFile;
  NamesFile namesFile;
  TrainingSet *trainingSet;
  BitSet *features; // Which attributes are included in model
  Vector<Vector<TrainingCase*>*> groups; // each grp = vector(TrainingCase)
  Vector<FeatureF*> f_values; // vector of FeatureF
  Schema *schema;

  double computeFRatio(int attrNum);
  void computeFratios();
  void setUpGroups();
  int groupSize(int groupNum);
public:
  Application(String nameFile,String trainFile);
  void go();
};


void usage();


int main(int argc,char *argv[])
{
  try
    {
      randomize();
      CommandLine cmd(argc,argv,"wsm");
      if(cmd.numArgs()!=2) usage();
      String namesFile=cmd.arg(0);
      String trainFile=cmd.arg(1);

      Application app(namesFile,trainFile);
      app.go();
    }
  catch(RootException &e)
    {
      cerr << e.getMessage() << endl;
      return -1;
    }
  catch(String s)
    {
      cerr << s.c_str() << endl;
      return -1;
    }
  catch(const char *s)
    {
      cerr << s << endl;
      return -1;
    }

  return 0;
}



void usage()
{
  throw String("\nf-ratio <names-file> <train-file>\n\n");
}



Application::Application(String nameFile,String trainFile)
  : trainFile(trainFile),
    nameFile(nameFile),
    trainingSet(NULL)
{
  // ctor
}



void Application::go()
{
  schema=namesFile.processFile(nameFile);
  numCategories=schema->getNumCategories();
  features=new BitSet(schema->numAttributes());
  features->addAll();
  trainingSet=new TrainingSet(*schema,trainFile);
  int numTrainingCases=trainingSet->numCases();
  setUpGroups();
  computeFratios();
}



double Application::computeFRatio(int attrNum)
{
  /*
    Computes MS(groups)/MS(error) = between-group 
    variability vs. within-group variability = a measure of
    how well this variable separates groups.
   */

  double sumX=0.0;
  int k=schema->getNumCategories();
  int i, N=0;
  double totalSS=0.0, groupSS=0.0;

  for(i=0 ; i<k ; ++i)
    {
      Vector<TrainingCase*> *group=groups[i];
      int groupSize=group->size();
      if(groupSize==0) continue;
      N+=groupSize;

      double localSumX=0.0;
      int j;
      for(j=0 ; j<groupSize ; ++j)
	{
	  TrainingCase *thisCase=(*group)[j];
	  double x=thisCase->getContinuous(attrNum);
	  localSumX+=x;
	  totalSS+=(x*x);
	}
      sumX+=localSumX;
      groupSS+=(localSumX/groupSize*localSumX);
    }

  double C=sumX/N*sumX;
  totalSS-=C;
  groupSS-=C;
  double errorSS=totalSS-groupSS;
  double totalDF=N-1;
  double groupDF=k-1;
  double errorDF=totalDF-groupDF;
  double groupMS=groupSS/groupDF;
  double errorMS=errorSS/errorDF;
  
  double ratio;
  if(errorMS==0)
    {
      ratio=POSITIVE_INFINITY;
      cerr << "WARNING!  F ratio=INF (" << ratio << ")" << endl; // ###
    }
  else
    ratio=groupMS/errorMS;
  return ratio;
}



void Application::setUpGroups()
{
  int nGroups=schema->getNumCategories();
  int i;
  for(i=0 ; i<nGroups ; ++i)
    groups.push_back(new Vector<TrainingCase*>);

  int n=trainingSet->numCases();
  for(i=0 ; i<n ; ++i)
    {
      TrainingCase &c=trainingSet->getCase(i);
      int cat=c.getCategory();
      if(cat>=nGroups) cat=nGroups-1;
      Vector<TrainingCase*> *group=groups[cat];
      group->push_back(&c);
    }
}



int Application::groupSize(int groupNum)
{
  return groups[groupNum]->size();
}



void Application::computeFratios()
{
  int numFeatures=schema->numAttributes();
  for(int i=0 ; i<numFeatures ; ++i)
    {
      double F=computeFRatio(i);
      cout << "F(" << namesFile.getName(i) << ")=" << F << endl;
    }
}



