/****************************************************************
 knn.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include <iostream>
#include <stdlib.h>
#include <fstream>
#include <math.h>
#include <unistd.h>
#include "BOOM/Exceptions.H"
#include "BOOM/String.H"
#include "BOOM/Stack.H"
#include "BOOM/Stacktrace.H"
#include "BOOM/BitSet.H"
#include "BOOM/Vector.H"
#include "BOOM/VectorSorter.H"
#include "BOOM/Random.H"
#include "BOOM/Array1D.H"
#include "BOOM/Array2D.H"
#include "BOOM/CommandLine.H"
#include "BOOM/InversionMatrix.H"
#include "BOOM/GaussJordan.H"
#include "BOOM/Time.H"
#include "BOOM/BoundedSortedArray.H"
#include "BOOM/Constants.H"
#include "../common/NamesFile.H"
#include "../common/TestSet.H"
#include "../common/Schema.H"
#include "FeatureF.H"
using namespace std;
using namespace BOOM;

class Application
{
  int numCategories;
  String trainFile, testFile, outfile, nameFile;
  NamesFile namesFile;
  TrainingSet *trainingSet, *testSet;
  BitSet *features; // Which attributes are included in model
  BitSet *trainingTraining;// For cross-validation
  Vector<Vector<TrainingCase*>*> groups; // each grp = vector(TrainingCase)
  Vector<FeatureF*> f_values; // vector of FeatureF
  bool stepwise, weighted, useMahalanobis;
  DblArray1D weights;
  DblMatrix *inverseCovMatrix;

  Schema *schema;
  double categoryWeight;
  int K;

  void classify();
  double distance(TrainingCase &,TrainingCase &);
  void getNeighbors(TrainingCase &,
		    BoundedSortedArray<TrainingCase*,double> &);
  int predictCategory(BoundedSortedArray<TrainingCase*,double> &);
  double computeFRatio(int attrNum);
  void setUpGroups();
  int groupSize(int groupNum);
  void featureSelection();
  void computeFratios();
  double crossVal_classify();
  void crossVal_getNeighbors(TrainingCase &,
			     BoundedSortedArray<TrainingCase*,double> &);
  void computeCovMatrix();
  double mahalanobis(TrainingCase &,TrainingCase &);
public:
  Application(int K,String nameFile,String trainFile,String testFile,
	      String outfile,bool stepwise,bool weighted,
	      bool useMahalanobis);
  void go();
};


void usage();


int main(int argc,char *argv[])
{
  try
    {
      Time stopwatch;
      stopwatch.startCounting();
      randomize();

      CommandLine cmd(argc,argv,"wsm");
      if(cmd.numArgs()!=5) usage();
      bool stepwise=cmd.option('s');
      bool weighted=cmd.option('w');
      bool useMahalanobis=cmd.option('m');
      int K=cmd.arg(0).asInt();
      String namesFile=cmd.arg(1);
      String trainFile=cmd.arg(2);
      String testFile=cmd.arg(3);
      String outfile=cmd.arg(4);

      Application app(K,namesFile,trainFile,testFile,outfile,
		      stepwise,weighted,useMahalanobis);
      app.go();
      stopwatch.stopCounting();
      cout << stopwatch.elapsedTime() << endl;
    }
  catch(RootException &e)
    {
      cerr << e.getMessage() << endl;
      return -1;
    }
  catch(String s)
    {
      cerr << s.c_str() << endl;
      return -1;
    }
  catch(const char *s)
    {
      cerr << s << endl;
      return -1;
    }

  return 0;
}



void usage()
{
  throw String("\
knn [-wsm] <K> <names-file> <train-file> <test-file> <out-file>\n\
   -w : weight variables by F ratio (between-group MS/within-group MS)\n\
   -s : stepwise - drop variables with low F ratio\n\
   -m : mahalanobis - account for multicollinearity\n\
\n");
}



Application::Application(int K,String nameFile,String trainFile,
			 String testFile,String outfile,bool stepwise,
			 bool weighted,bool useMahalanobis)
  : trainFile(trainFile),
    nameFile(nameFile),
    testFile(testFile),
    outfile(outfile),
    trainingSet(NULL), 
    weighted(weighted),
    categoryWeight(1.0), 
    K(K), 
    weights(0),
    stepwise(stepwise), 
    useMahalanobis(useMahalanobis)
{
  // ctor
}



void Application::go()
{
  schema=namesFile.processFile(nameFile);
  numCategories=schema->getNumCategories();
  features=new BitSet(schema->numAttributes());
  features->addAll();
  weights.resize(schema->numAttributes());

  trainingSet=new TrainingSet(*schema,trainFile);
  int numTrainingCases=trainingSet->numCases();
  trainingTraining=new BitSet(numTrainingCases); // ### ?
  setUpGroups();
  computeFratios();
  if(useMahalanobis) computeCovMatrix();
  if(stepwise) featureSelection();

  testSet=new TestSet(*schema,testFile);

  classify();
}



void Application::classify()
{
  ofstream os(outfile.c_str());
  double lastPercent=0.0;
  int n=testSet->numCases();
  for(int i=0 ; i<n ; ++i)
    {
      double percentDone=100.0 * i / n;
      if(percentDone>=lastPercent+5)
	{
	  cout << int(percentDone) << "% "; cout.flush();
	  lastPercent=percentDone;
	}
      TrainingCase &thisCase=testSet->getCase(i);
      
      BoundedSortedArray<TrainingCase*,double> neighbors(K,KEEP_LOW_VALUES);
      getNeighbors(thisCase,neighbors);
      int predicted=predictCategory(neighbors);
      os<<predicted<<endl;
    }
}



void Application::getNeighbors(TrainingCase &thisCase,
		BoundedSortedArray<TrainingCase*,double> &neighbors)
{
  int i;
  for(i=0 ; i<trainingSet->numCases() ; ++i)
    {
      TrainingCase &thatCase=trainingSet->getCase(i);
      double d=distance(thisCase,thatCase);
      neighbors.add(&thatCase,d);
    }
}



int Application::predictCategory(
        BoundedSortedArray<TrainingCase*,double> &neighbors)
{
  DblArray1D distances(numCategories);
  distances.setAllTo(0.0);
  BitSet touched(numCategories);

  for(int i=0 ; i<K ; ++i)
    {
      TrainingCase &c=*neighbors[i];
      int category=c.getCategory();
      if(category>=numCategories) 
	throw String("Illegal category: ")+category;
      double distance=neighbors.getValue(i);
      distances[category]+=1/(distance*distance);
      touched.addMember(category);
    }

  int lowestIndex=-1;
  double lowestDist=0;
  for(int i=0 ; i<numCategories ; ++i)
    if(touched.isMember(i))
       {
	 double d=distances[i];
	 if(d >lowestDist)
	   {
	     lowestIndex=i;
	     lowestDist=d;
	   }
       }

  return lowestIndex;
}



double Application::distance(TrainingCase &case1,TrainingCase &case2)
{
  // Computes euclidean distance

  if(useMahalanobis) return mahalanobis(case1,case2);

  double sum=0;

  int i, n=schema->numAttributes();
  for(i=0 ; i<n ; ++i)
    {
      // Make sure this feature hasn't been dropped
      // (due to a low F value)
      if(!features->isMember(i)) continue;

      switch((*schema)[i]->getType())
	{
	case DISCRETE:
	  if(case1.getDiscrete(i)!=case2.getDiscrete(i))
	    sum+=weights[i];
	  break;
	case CONTINUOUS:
	  {
	    double delta=
	      case1.getContinuous(i)-case2.getContinuous(i);
	    if(weighted)
	      sum+=delta*delta*weights[i];
	    else
	      sum+=delta*delta;
	  }
	  break;
	}
    }

  return sqrt(sum);
}



double Application::computeFRatio(int attrNum)
{
  /*
    Computes MS(groups)/MS(error) = between-group 
    variability vs. within-group variability = a measure of
    how well this variable separates groups.
   */

  double sumX=0.0;
  int k=schema->getNumCategories();
  int i, N=0;
  double totalSS=0.0, groupSS=0.0;

  for(i=0 ; i<k ; ++i)
    {
      Vector<TrainingCase*> *group=groups[i];
      int groupSize=group->size();
      if(groupSize==0) continue;
      N+=groupSize;

      double localSumX=0.0;
      int j;
      for(j=0 ; j<groupSize ; ++j)
	{
	  TrainingCase *thisCase=(*group)[j];
	  double x=thisCase->getContinuous(attrNum);
	  localSumX+=x;
	  totalSS+=(x*x);
	}
      sumX+=localSumX;
      groupSS+=(localSumX/groupSize*localSumX);
    }

  double C=sumX/N*sumX;
  totalSS-=C;
  groupSS-=C;
  double errorSS=totalSS-groupSS;
  double totalDF=N-1;
  double groupDF=k-1;
  double errorDF=totalDF-groupDF;
  double groupMS=groupSS/groupDF;
  double errorMS=errorSS/errorDF;
  
  double ratio;
  if(errorMS==0)
    {
      ratio=POSITIVE_INFINITY;
      cerr << "WARNING!  F ratio=INF (" << ratio << ")" << endl; // ###
    }
  else
    ratio=groupMS/errorMS;
  return ratio;
}



void Application::setUpGroups()
{
  int nGroups=schema->getNumCategories();
  int i;
  for(i=0 ; i<nGroups ; ++i)
    groups.push_back(new Vector<TrainingCase*>);

  int n=trainingSet->numCases();
  for(i=0 ; i<n ; ++i)
    {
      TrainingCase &c=trainingSet->getCase(i);
      int cat=c.getCategory();
      if(cat>=nGroups) cat=nGroups-1;
      Vector<TrainingCase*> *group=groups[cat];
      group->push_back(&c);
    }
}



int Application::groupSize(int groupNum)
{
  return groups[groupNum]->size();
}



void Application::computeFratios()
{
  int numFeatures=schema->numAttributes();
  int i;
  for(i=0 ; i<numFeatures ; ++i)
    {
      double F=computeFRatio(i);
      //cerr << "F(" << i << ")=" << F << endl;
      
      f_values.push_back(new FeatureF(i,F));
      weights[i]=1/F;
    }
}



void Application::featureSelection()
{
  cerr << "Performing feature selection..." << endl; 

  int numFeatures=schema->numAttributes();

  // Sort the F values in increasing order
  FeatureComparator cmp;
  VectorSorter<FeatureF*> sorter(f_values,cmp);
  sorter.sortAscendInPlace();

  // Split training set for cross-validation
  trainingTraining->purge();
  int i, trainingSetSize=trainingSet->numCases();
  for(i=0 ; i<trainingSetSize ; ++i)
    if(Random0to1()<0.66)
      trainingTraining->addMember(i);

  // Compute cross-val accuracy with all features left
  // in the model
  double bestAccuracy=crossVal_classify();
  cerr << "acc=" << bestAccuracy << endl; // ###

  // Try removing some features to see if accuracy is
  // improved
  for(i=0 ; i<numFeatures ; ++i)
    {
      FeatureF *ff=f_values[i];
      features->removeMember(ff->featureNum);
      
      double accuracy=crossVal_classify();
      if(accuracy<bestAccuracy)
	{
	  cerr << "Removing feature " << ff->featureNum
	       << " reduces acc to " << accuracy
	       << ", so it will not be removed." << endl; // ###
	  features->addMember(ff->featureNum);
	  break;
	}
      bestAccuracy=accuracy;
      cerr << "removing feature " << ff->featureNum
	   << ".  Now acc=" << bestAccuracy << endl; // ###
    }
}



double Application::crossVal_classify()
{
  /*
    This is the cross-validation version of classify()
   */

  int numCorrect=0, totalError=0;

  int n=trainingSet->numCases();
  for(int i=0 ; i<n ; ++i)
    if(!trainingTraining->isMember(i))
      {
	TrainingCase &thisCase=trainingSet->getCase(i);
	
	BoundedSortedArray<TrainingCase*,double> 
	  neighbors(K,KEEP_LOW_VALUES);
	crossVal_getNeighbors(thisCase,neighbors);
	int predictedCategory=predictCategory(neighbors);
	
	if(thisCase.getCategory()==predictedCategory)
	  ++numCorrect;
      }

  double accuracy;
  accuracy=100.0 * numCorrect/(double) n;

  return accuracy;
}



void Application::crossVal_getNeighbors(TrainingCase &thisCase,
	       BoundedSortedArray<TrainingCase*,double> &neighbors)
{
  /*
    This is the cross-validation version of getNeighbors()
   */

  for(int i=0 ; i<trainingSet->numCases() ; ++i)
    if(trainingTraining->isMember(i))
      {
	TrainingCase &thatCase=trainingSet->getCase(i);
	double d=distance(thisCase,thatCase);
	neighbors.add(&thatCase,d);
      }
}



void Application::computeCovMatrix()
{
  int numAttr=schema->numAttributes();
  DblArray2D sumXY(numAttr,numAttr);
  DblArray1D sumX(numAttr);

  sumXY.setAllTo(0.0);
  sumX.setAllTo(0.0);

  int numCases=trainingSet->numCases();
  for(int i=0 ; i<numCases ; ++i)
    {
      TrainingCase &thisCase=trainingSet->getCase(i);
      for(int j=0 ; j<numAttr ; ++j)
	{
	  double x=thisCase.getContinuous(j);
	  sumX[j]+=x;
	  for(int k=0 ; k<numAttr ; ++k)
	    {
	      double y=thisCase.getContinuous(k);
	      sumXY[j][k]+=x*y;
	    }
	}
    }
  
  DblMatrix covMatrix(numAttr,numAttr);
  for(int j=0 ; j<numAttr ; ++j)
    for(int k=0 ; k<numAttr ; ++k)
      {
	covMatrix(j,k)=
	  (sumXY[j][k]-sumX[j]*sumX[k]/numCases)/
	  (numCases-1.0);
      }

  inverseCovMatrix=new DblMatrix(numAttr,numAttr);
  if(!GaussJordan::invert(covMatrix,*inverseCovMatrix))
    {
      cerr << "WARNING: cov matrix not invertible"
	   << "\n         using euclidean distance" << endl;
      useMahalanobis=false;
    }
}



double Application::mahalanobis(TrainingCase &case1,
			       TrainingCase &case2)
{
  int numAttr=schema->numAttributes();
  DblMatrix left(1,numAttr), right(numAttr,1);

  int i;
  for(i=0 ; i<numAttr ; ++i)
    {
      double d=case1.getContinuous(i)-case2.getContinuous(i);
      if(weighted)
	left(0,i)=right(i,0)=d*sqrt(weights[i]);
      else
	left(0,i)=right(i,0)=d;
    }

  DblMatrix temp1(1,numAttr);
  multiply(left,*inverseCovMatrix,temp1);

  DblMatrix temp2(1,1);
  multiply(temp1,right,temp2);

  double d=temp2(0,0);
  return d;
}


