/****************************************************************
 train-logistic.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/

/*

              Quote from some web page I found:

"For improved precision, the independent variables are temporarily 
converted to "standard scores" ( value - Mean ) / StdDev. 

The Null Model is used as the starting guess for the iterations -- all 
parameter coefficients are zero, and the intercept is the logarithm of 
the ratio of the number of cases with y=1 to the number with y=0. 

The quantity -2*Ln(Likelihood) is displayed for the null model, for 
each step of the iteration, and for the final (converged model)."

*/

#include <iostream>
#include <fstream>
#include "BOOM/String.H"
#include "BOOM/CommandLine.H"
#include "BOOM/GSL/Optimizer.H"
#include "BOOM/Random.H"
#include "Classifier.H"
#include "../Schema.H"
#include "../TrainingSet.H"
#include "../NamesFile.H"
#include "LogLikelihood.H"
using namespace std;
using namespace BOOM;


class Application
{
public:
  Application();
  int main(int argc,char *argv[]);

private:
  Schema *schema;
  NamesFile namesFile;
  int numCategories, numVariables, numCases;
  TrainingSet *trainingSet;
  Classifier *classifier;
  double stepSize, threshold;
  int maxIterations, numRuns;
  GSL::OptimizerType optimizerType;

  void trainBinary();
  LogisticFunction trainBinary(TrainingSet &class0,TrainingSet &class1);
  void trainMulticlass();
  double evaluate(Classifier &,TrainingSet &);
};


int main(int argc,char *argv[])
  {
    try
      {
	Application app;
	return app.main(argc,argv);
      }
    catch(const char *p)
      {
	cerr << p << endl;
      }
    catch(const string &msg)
      {
	cerr << msg.c_str() << endl;
      }
    catch(const exception &e)
      {
	cerr << "STL exception caught in main:\n" << e.what() << endl;
      }
    catch(...)
      {
	cerr << "Unknown exception caught in main" << endl;
      }
    return -1;
  }



Application::Application()
  {
    // ctor
  }



int Application::main(int argc,char *argv[])
{
  // Process command line
  CommandLine cmd(argc,argv,"i:r:t:s:g:");
  if(cmd.numArgs()!=3)
    throw String(
"\n\
train-logistic [options] <*.names> <*.data> <outfile>\n\
     where:\n\
       -i <N> : use N iterations of gradient ascent\n\
       -r <N> : randomly restart N-1 times and take the best\n\
       -t <T> : quit when error<T (threshold)\n\
       -s <s> : use stepsize s for the gradient ascent\n\
       -a <G> : use optimization algorithm G\n\
");
  String namesFilename=cmd.arg(0);
  String trainFile=cmd.arg(1);
  String outfile=cmd.arg(2);
  
  // Load configuration
  schema=namesFile.processFile(namesFilename);
  numCategories=schema->getNumCategories();
  numVariables=schema->numAttributes();
  
  // Load training data
  trainingSet=new TrainingSet(*schema,trainFile);
  numCases=trainingSet->numCases();
  
  // Set default optimization meta-parameters
  stepSize=0.1;
  threshold=0.0001;
  maxIterations=50;
  numRuns=5;
  optimizerType=GSL::BFGS;

  // Override defaults using command-line options
  if(cmd.option('i')) maxIterations=cmd.optParm('i').asInt();
  if(cmd.option('r')) numRuns=cmd.optParm('r').asInt();
  if(cmd.option('t')) threshold=cmd.optParm('t').asDouble();
  if(cmd.option('s')) stepSize=cmd.optParm('s').asDouble();
  if(cmd.option('a')) 
    optimizerType=GSL::stringToOptimizerType(cmd.optParm('a'));

  // Build model
  randomize();
  if(numCategories==2)
    trainBinary();
  else
    trainMulticlass();
  
  // Generate output
  ofstream os(outfile.c_str());
  classifier->save(os);
  os.close();

  // Evaluate on training set
  double acc=evaluate(*classifier,*trainingSet);
  acc=int(acc*100+5/9.0);
  cout<<acc<<"% accuracy on training set"<<endl;
 
  return 0;
}



void Application::trainBinary()
{
  TrainingSet *s=trainingSet->getCategory(0);
  TrainingSet *t=trainingSet->getCategory(1);
  double bestAcc=0;
  this->classifier=NULL;
  for(int r=0 ; r<numRuns ; ++r)
    {
      LogisticFunction f=trainBinary(*s,*t);
      delete s;
      delete t;
      Classifier *classifier=new BinaryClassifier(f);
      double acc=evaluate(*classifier,*trainingSet);
      if(this->classifier==NULL || acc>bestAcc)
	{
	  delete this->classifier;
	  this->classifier=classifier;
	  bestAcc=acc;
	}
    }
}



LogisticFunction Application::trainBinary(TrainingSet &class0,
					  TrainingSet &class1)
{
  GSL::Vector initialPoint(numVariables+1);
  initialPoint.setAllTo(0);
  initialPoint[0]=log(class0.numCases()/(double)class1.numCases());
  bool success=false;
  GSL::Vector optimalPoint;
  LogLikelihood L(class0,class1);
  while(!success)
    {
      L.resetOK();
      for(int i=1 ; i<numVariables+1 ; ++i)initialPoint[i]=Random0to1()-0.5;
      GSL::Optimizer optimizer(optimizerType,L,initialPoint,stepSize,
			       GSL::BY_EITHER,threshold,maxIterations);
      success=optimizer.run();
      if(!L.isOK()) {success=false; continue;}
      cout<<optimizer.iterationsUsed()<<" iterations"<<endl;
      optimalPoint=optimizer.getOptimalPoint();
    }
  LogisticFunction f(numVariables);
  L.initFunc(f,optimalPoint);
  return f;
}



void Application::trainMulticlass()
{
  MulticlassClassifier *classifier=new MulticlassClassifier;
  for(int i=0 ; i<numCategories ; ++i)
    {
      TrainingSet *s=trainingSet->getCategory(i);
      TrainingSet *t=trainingSet->getNotCategory(i);
      LogisticFunction f=trainBinary(*s,*t);
      delete s;
      delete t;
      classifier->addFunction(f);
    }
  this->classifier=classifier;
}




double Application::evaluate(Classifier &classifier,TrainingSet &T)
{
  int numCorrect=0;
  int n=T.numCases();
  for(int i=0 ; i<n ; ++i)
    {
      TrainingCase &c=T.getCase(i);
      int predicted=classifier.classify(c);
      int actual=c.getCategory();
      if(predicted==actual) ++numCorrect;
    }
  return numCorrect/(double)n;
}


