/****************************************************************
 Backprop.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "Backprop.H"
#include "Neuron.H"
#include <iostream>
#include <stdlib.h>
#include <fstream>
#include <math.h>
#include "BOOM/Array1D.H"
#include "BOOM/Random.H"



Backpropagator::Backpropagator()
  : synapticWeights(NULL), bestError(-1), noiseFactor(0.0)
{
  // ctor
}



void Backpropagator::train(NeuralNetwork &nn,
  Vector<NeuralCase*> &cases,float learningRate,
  TransferFunction &f,int maxIterations,ostream &log,
  float noiseFactor)
{
  /*
    This is the top-level function for the 
    back-propagation algorithm.
    */

  this->noiseFactor=noiseFactor;

  prepareToTrain(nn,learningRate,log,f,cases);
  trainLoop(nn,maxIterations);
  recallBestWeights(nn);
}



void Backpropagator::prepareToTrain(NeuralNetwork &nn,
				    float learningRate,
				    ostream &log,
				    TransferFunction &f,
				    Vector<NeuralCase*> &cases)
{
  this->learningRate=learningRate;
  this->log=&log;
  this->f=&f;
  this->cases=&cases;

  synapticWeights=new FloatArray1D(2*nn.countSynapses());
  nn.copySynapsesOut(*synapticWeights);

  // Topologically sort the neurons, so that we 
  // can process them in a back-to-front manner
  indices=nn.topologicalSort();
  numNeurons=indices->size();
}



float Backpropagator::iterate(NeuralNetwork &nn,int j)
{
  /*
    This method runs all training cases through the network
    once.  It returns the average error.
   */

  float totalError=0;
  
  // Process each training case
  int i;
  for(i=0 ; i<cases->size() ; ++i)
    {
      NeuralCase *c=(*cases)[i];
      float err;
      processTrainingCase(*c,nn,learningRate,
			  *f,*indices,numNeurons,err);
      totalError+=err;
    }
  
  int numOutputNeurons=nn.getNumOutputs();
  int numCases=cases->size();
  float averageError=
    totalError/(numCases*numOutputNeurons);
  writeToLog(*log,j,averageError);
  
  return averageError;
}



void Backpropagator::trainLoop(NeuralNetwork &nn,
			       int maxIterations)
{
  int j;
  for(j=0 ; j<maxIterations ; ++j)
    {
      updateNoiseFactor(j,maxIterations);

      cerr << "iteration  #" << j << endl;
      float aveErr=iterate(nn,j);

      if(aveErr<bestError || bestError<0)
	{
	  nn.copySynapsesOut(*synapticWeights);
	  bestError=aveErr;
	}
    }
}



void Backpropagator::writeToLog(ostream &log,int j,
				float averageError)
{
  //log << "iter #" << j << ", ave err=" << averageError << endl;
}



void Backpropagator::processTrainingCase(NeuralCase &c,
  NeuralNetwork &nn,float learningRate,TransferFunction &f,
  IntVector &indices,int numNeurons,float &err)
{
  /*
    This function tests the network on a single 
    training case, and modifies the weights of 
    the synapses to modify the output of the network 
    to more closely match the expected outputs.
   */

  // Propagate the inputs through the net to 
  // produce some outputs
  applyCase(c,nn);

  // Allocate an array to hold the delta value 
  // for each neuron
  FloatArray1D delta(numNeurons);
  delta.setAllTo(0.0);//in case the net is recurrent

  // Compute delta values for the output neurons 
  // (because this uses a formula different from 
  // that used for hidden nodes)
  int i;
  IntVector &outputIndices=nn.getOutputIndices();
  err=0.0;
  for(i=0 ; i<outputIndices.size() ; ++i)
    {
      int index=outputIndices[i];
      float expected=c.getOutput(i);
      Neuron &neuron=nn[index];
      float observed=neuron.getActivationValue();
      float x=neuron.getCombinedInputs();
      float dydx=f.derivative(x);

      // Changed the following line to / rather than *
      // on 6/4/99: ###
      float deltaValue=(expected-observed)*dydx;

      delta[index]=deltaValue;
      err+=fabs(expected-observed);
    }

  // Sweep backward through the network
  for(i=0 ; i<numNeurons ; ++i)
    {
      int index=indices[i];
      processNeuron(index,nn,delta,f,learningRate);
    }
}



void Backpropagator::applyCase(NeuralCase &c,
			       NeuralNetwork &nn)
{
  /* 
     This function simply applies the inputs of 
     the training case to the input neurons and 
     then propagates the network to compute its 
     outputs.
     */

  // Apply inputs to input neurons
  IntVector &inputIndices=nn.getInputIndices();
  int i;
  for(i=0 ; i<inputIndices.size() ; ++i)
    {
      int index=inputIndices[i];
      float value=c.getInput(i);

      addNoise(value);

      nn.setInput(index,value);
    }

  // Propagate the system
  nn.propagate();
}



void Backpropagator::processNeuron(int index,
   NeuralNetwork &nn,FloatArray1D &delta,
   TransferFunction &f,float learningRate)
{
  /*
    This function modifies the weights of the 
    synapses leading into this neuron.
    */

  // First, compute the delta value for this neuron
  computeDelta(index,nn,delta,f);

  // Now iterate through the synapses leading into 
  // this neuron, so that we may modify their weights
  Neuron &neuron=nn[index];
  int inDegree=neuron.getInDegree();
  int i;
  for(i=0 ; i<inDegree ; ++i)
    {
      Synapse &synapse=neuron.getInput(i);
      float weight=synapse.getWeight();
      int from=synapse.getFrom();
      float oi=nn[from].getActivationValue();
      float deltaValue=delta[index];
      float dWeight=learningRate * deltaValue * oi;
      float newWeight=weight+dWeight;
      synapse.changeWeight(newWeight);
    }
  
}



void Backpropagator::computeDelta(int index,
   NeuralNetwork &nn,FloatArray1D &delta,
   TransferFunction &f)
{
  if(delta[index]!=0.0)// if it has already been computed
    return;

  Neuron &neuron=nn[index];
  int outDegree=neuron.getOutDegree();
  float sum=0.0;
  int i;
  for(i=0 ; i<outDegree ; ++i)
    {
      Synapse &synapse=neuron.getOutput(i);
      int to=synapse.getTo();
      float weight=synapse.getWeight();
      float deltaValue=delta[to];
      sum+=weight*deltaValue;
    }

  float oj=neuron.getActivationValue();
  float x=neuron.getCombinedInputs();
  float deltaValue=sum * f.derivative(x);
  delta[index]=deltaValue;
}



void Backpropagator::addNoise(float &value)
{
  float r=RandomGaussian(-0.5,0.5,5);
  float noise=noiseFactor*r;
  value+=noise;
  if(value<0.0) value=0.0;
  if(value>1.0) value=1.0;
}



void Backpropagator::updateNoiseFactor(int iteration,
				       int maxIter)
{
  if(iteration>=0.25*maxIter)
    noiseFactor*=0.5;
  if(iteration>=0.50*maxIter)
    noiseFactor*=0.5;
  if(noiseFactor>=0.75*maxIter)
    noiseFactor=0.0;
}



void Backpropagator::recallBestWeights(NeuralNetwork &nn)
{
  nn.copySynapsesIn(*synapticWeights);
}


