/****************************************************************
 NeuralNetwork.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include <iostream>
#include <math.h>
#include "BOOM/VectorSorter.H"
#include "BOOM/ListQueue.H"
#include "NeuralNetwork.H"
#include "Neuron.H"
#include "StandardNeuron.H"
using namespace std;


NeuralNetwork::NeuralNetwork()
{
  // ctor
}



NeuralNetwork::~NeuralNetwork()
{
}



bool NeuralNetwork::validIndex(int i)
{
  return i>=0 && i<neurons.size() ? true : false;
}



void NeuralNetwork::addInputNeuron(Neuron *neuron)
{
  int num=getNumNeurons();
  addNeuron(neuron);
  inputs.push_back(num);
}



void NeuralNetwork::addOutputNeuron(Neuron *neuron)
{
  int num=getNumNeurons();
  addNeuron(neuron);
  outputs.push_back(num);
}



Vector<int> &NeuralNetwork::getInputIndices()
{
  return inputs;
}



Vector<int> &NeuralNetwork::getOutputIndices()
{
  return outputs;
}



void NeuralNetwork::addNeuron(Neuron *neuron)
{
  neurons.push_back(neuron);
}



Neuron &NeuralNetwork::operator[](int i) 
{
  if(!validIndex(i))
    throw Stacktrace("index out of bounds");
  return *getNeuron(i);
}



void NeuralNetwork::propagate()
{
  for(int i=0 ; i<neurons.size() ; ++i)
    {
      Neuron &neuron=*getNeuron(i);
      if(neuron.getInDegree()>0)
	neuron.update(*this);
    }
}



void NeuralNetwork::setInput(int whichNode,float value)
{
  if(!validIndex(whichNode))
    throw Stacktrace("index out of bounds");
  
  getNeuron(whichNode)->setActivationValue(value);
}



float NeuralNetwork::getOutput(int whichNode)
{
  if(!validIndex(whichNode))
    throw Stacktrace("index out of bounds");
  
  return getNeuron(whichNode)->getActivationValue();
}



Neuron *NeuralNetwork::getNeuron(int i)
{
  // no index checking -- this function is only 
  // called from within this code; the index checking
  // is done by the operator[], which calls this method

  return (Neuron *) neurons[i];
}



void NeuralNetwork::connect(int from,int to,float weight)
{
  if(!validIndex(from) || !validIndex(to))
    throw Stacktrace("index out of bounds");
  
  Synapse *synapse=new Synapse(weight,from,to);
  getNeuron(from)->addOutput(synapse);
  getNeuron(to)->addInput(synapse);
}



int NeuralNetwork::getNumNeurons() const
{
  return neurons.size();
}


void NeuralNetwork::insertInts(Vector<int> &vector,int value,
			       int num)
{
  for(int i=0 ; i<num ; ++i)
    vector.push_back(value);
}



Vector<int> *NeuralNetwork::topologicalSort()
{
  /*
    This produces a topological sort, with the outputs 
    being earlier in the list, and inputs being later.  
    This must NOT be applied to a recurrent network!
   */

  // Allocate an array to hold the depth of each neuron
  Vector<int> depth;
  insertInts(depth,0,neurons.size());

  // Allocate a queue into which we will put neuron indices
  ListQueue<int> queue;

  // First, enqueue the output neurons
  for(int i=0 ; i<outputs.size() ; ++i)
    queue.enqueue(outputs[i]);

  // Now process each element in the queue, 
  // setting its predecessors to have a depth 
  // equal to its own plus 1
  while(!queue.isEmpty())
    {
      int index=queue.dequeue();
      int thisDepth=depth[index];
      Neuron *neuron=getNeuron(index);
      int inDegree=neuron->getInDegree();
      for(int i=0 ; i<inDegree ; ++i)
	{
	  Synapse &synapse=neuron->getInput(i);
	  int pred=synapse.getFrom();
	  depth[pred]=thisDepth+1;
	  queue.enqueue(pred);
	}
    }

  // Now sort by depth, to achieve the topological sort
  DirectComparator<int> comparator;
  VectorSorter<int> sorter(depth,comparator);
  Vector<int> *indices=sorter.sortAscendByIndex();

  return indices;
}



void NeuralNetwork::dump(ostream &os)
{
  int i;
  for(i=0 ; i<neurons.size() ; ++i)
    {
      Neuron *neuron=getNeuron(i);
      os << "\nNeuron #" << i << endl;
      
      os << "\tInputs:\n\t\t";
      int inDegree=neuron->getInDegree();
      int j;
      for(j=0 ; j<inDegree ; ++j)
	os << neuron->getInput(j).getFrom() << ",";

      os << "\n\tOutputs:\n\t\t";
      int outDegree=neuron->getOutDegree();
      for(j=0 ; j<outDegree ; ++j)
	os << neuron->getOutput(j).getTo() << ",";
    }
}



void NeuralNetwork::saveSelf(ostream &os)
{
  /*
    The file format is as follows:

    -----------------------------------------------
    number of neurons
    number of input neurons
    indices of input neurons (separated by spaces)
    number of output neurons
    indices of output neurons (separated by spaces)
    neuron 0
    neuron 1
    ...
    neuron n
    -----------------------------------------------

    See Neuron.C for a description of the format
    of each neuron in the file.
   */

  // Write # neurons
  const int numNeurons=neurons.size();
  os << numNeurons << endl;

  // Write # input neurons
  const int numInputs=inputs.size();
  os << numInputs << endl;

  // Write indices of input neurons
  int i;
  for(i=0 ; i<numInputs ; ++i)
    os << inputs[i] << ' ';
  os << endl;

  // Write # output neurons
  const int numOutputs=outputs.size();
  os << numOutputs << endl;

  // Write indices of output neurons
  for(i=0 ; i<numOutputs ; ++i)
    os << outputs[i] << ' ';
  os << endl;

  // Finally, instruct each neuron to write itself
  for(i=0 ; i<numNeurons ; ++i)
    getNeuron(i)->saveSelf(os);
}



void NeuralNetwork::loadSelf(istream &is)
{
  // See loadSelf() for the file format

  // Read # neurons
  int numNeurons;
  is >> numNeurons;

  // Allocate that many neurons
  int i;
  for(i=0 ; i<numNeurons ; ++i)
    addNeuron(new StandardNeuron);

  // Read number of input neurons
  int numInputs;
  is >> numInputs;

  // Read indices of input neurons
  for(i=0 ; i<numInputs ; ++i)
    {
      int index;
      is >> index;
      inputs.push_back(index);
    }

  // Read number of output neurons
  int numOutputs;
  is >> numOutputs;

  // Read indices of output neurons
  for(i=0 ; i<numOutputs ; ++i)
    {
      int index;
      is >> index;
      outputs.push_back(index);
    }

  // Finally, instruct each neuron to read itself
  for(i=0 ; i<numNeurons ; ++i)
    getNeuron(i)->loadSelf(is,*this);
}



int NeuralNetwork::getNumInputs() 
{
  return inputs.size();
}



int NeuralNetwork::getNumOutputs() 
{
  return outputs.size();
}



int NeuralNetwork::countSynapses()
{
  int num=0;

  int i, n=neurons.size();
  for(i=0 ; i<n ; ++i)
    {
      Neuron *neuron=(Neuron *) neurons[i];
      num+=neuron->getOutDegree();
    }

  return num;
}



void NeuralNetwork::copySynapsesOut(FloatArray1D &array)
{
  int si=0, n=neurons.size();
  for(int i=0 ; i<n ; ++i)
    {
      Neuron *neuron=neurons[i];

      int inDegree=neuron->getInDegree();
      for(int j=0 ; j<inDegree ; ++j)
		  array[si++]=neuron->getInput(j).getWeight();
		
      int outDegree=neuron->getOutDegree();
      for(int j=0 ; j<outDegree ; ++j)
		  array[si++]=neuron->getOutput(j).getWeight();
    }  
}



void NeuralNetwork::copySynapsesIn(FloatArray1D &array)
{
  int si=0, n=neurons.size();
  for(int i=0 ; i<n ; ++i)
    {
      Neuron *neuron=(Neuron *) neurons[i];

      int inDegree=neuron->getInDegree();
      for(int j=0 ; j<inDegree ; ++j)
		  neuron->getInput(j).changeWeight(array[si++]);
		
      int outDegree=neuron->getOutDegree();
      for(int j=0 ; j<outDegree ; ++j)
		  neuron->getOutput(j).changeWeight(array[si++]);
    }  
}




