/****************************************************************
 Neuron.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "Neuron.H"

Neuron::Neuron(float activationValue,NeuronType &type)
  : activationValue(activationValue),
    type(type)
{
  // ctor
}



Neuron::~Neuron()
{
  // dtor

  for(int i=0 ; i<outputs.size() ; ++i)
    delete outputs[i];
}



void Neuron::addInput(const Synapse *synapse)
{
  inputs.push_back(const_cast<Synapse*>(synapse));
}



void Neuron::addOutput(const Synapse *synapse)
{
  outputs.push_back(const_cast<Synapse*>(synapse));
}



int Neuron::getInDegree() const
{
  return inputs.size();
}



int Neuron::getOutDegree() const
{
  return outputs.size();
}



Synapse &Neuron::getInput(int i)
{
  if(i<0 || i>=getInDegree())
    throw Stacktrace("in Neuron::getInput()");
  return *inputs[i];
}



Synapse &Neuron::getOutput(int i)
{
  if(i<0 || i>=getOutDegree())
    throw Stacktrace("in Neuron::getOutput()");
  return *outputs[i];
}



float Neuron::getActivationValue() const
{
  return activationValue;
}



void Neuron::changeActivationValue(float v)
{
  activationValue=v;
}



void Neuron::update(NeuralNetwork &network)
{
  combinedInputs=
    type.getInputCombiningFunction()(*this,network);
  const float newValue=
    type.getTransferFunction()(combinedInputs);
  activationValue=newValue;
}



void Neuron::setActivationValue(float value)
{
  activationValue=value;
}



float Neuron::getCombinedInputs()
{
  return combinedInputs;
}



void Neuron::saveSelf(ostream &os)
{
  /*
    The file format for a single neuron is as follows:

    -----------------------------
    Activation value
    Number of output synapses
    Output synapse 1
    Output synapse 2
    ...
    Output synapse n
    -----------------------------
   */

  // Write the activation value
  os << activationValue << endl;

  // Write the number of outputs synapses
  const int numOutputs=outputs.size();
  os << numOutputs << endl;

  // Write each output synapse
  int i;
  for(i=0 ; i<numOutputs ; ++i)
    {
      Synapse *synapse=outputs[i];
      synapse->saveSelf(os);
    }
}



void Neuron::loadSelf(istream &is,NeuralNetwork &network)
{
  /*
    See saveSelf() for details of the file format 
    for Neurons
   */

  // Read the activation value
  is >> activationValue;

  // Read the number of output synapses
  int numOutputs;
  is >> numOutputs;

  // Read each output synapse
  Synapse synapse;
  for(int i=0 ; i<numOutputs ; ++i)
    {
      synapse.loadSelf(is);
      network.connect(synapse.getFrom(),synapse.getTo(),
		      synapse.getWeight());
    }
}





