/****************************************************************
 Normalizer.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "Normalizer.H"
#include <iostream>
using namespace std;


Normalizer::Normalizer(int numInputs,int numOutputs)
  : numInputs(numInputs), numOutputs(numOutputs),
    maxInput(numInputs), maxOutput(numOutputs),
    minInput(numInputs), minOutput(numOutputs)
{
  // ctor
}



void Normalizer::normalize(Vector<NeuralCase*> &cases,
			   float maxAdjuster,
			   float minAdjuster)
{
  findMaxima(cases);
  findMinima(cases);

  adjust(maxAdjuster,maxInput,numInputs);
  adjust(maxAdjuster,maxOutput,numOutputs);
  adjust(minAdjuster,minInput,numInputs);
  adjust(minAdjuster,minOutput,numOutputs);

  int n=cases.size();
  for(int i=0 ; i<n ; ++i)
    {
      NeuralCase *trainingCase=cases[i];
      normalize(trainingCase);
    }
}



void Normalizer::normalize(Vector<NeuralCase*> &cases)
{
  int n=cases.size();
  for(int i=0 ; i<n ; ++i)
    {
      NeuralCase *trainingCase=cases[i];
      normalize(trainingCase);
    }
}



void Normalizer::findMaxima(Vector<NeuralCase*> &cases)
{
  NeuralCase *trainingCase=cases[0];

  for(int i=0 ; i<numInputs ; ++i)
    maxInput[i]=trainingCase->getInput(i);

  for(int i=0 ; i<numOutputs ; ++i)
    maxOutput[i]=trainingCase->getOutput(i);

  int n=cases.size();
  for(int i=0 ; i<n ; ++i)
    {
      NeuralCase *trainingCase=cases[i];
      
      for(int j=0 ; j<numInputs ; ++j)
	{
	  float x=trainingCase->getInput(j);
	  if(x>maxInput[j]) maxInput[j]=x;
	}

      for(int j=0 ; j<numOutputs ; ++j)
	{
	  float x=trainingCase->getOutput(j);
	  if(x>maxOutput[j]) maxOutput[j]=x;
	}
    }
}



void Normalizer::findMinima(Vector<NeuralCase*> &cases)
{
  NeuralCase *trainingCase=cases[0];

  for(int i=0 ; i<numInputs ; ++i)
    minInput[i]=trainingCase->getInput(i);

  for(int i=0 ; i<numOutputs ; ++i)
    minOutput[i]=trainingCase->getOutput(i);

  int n=cases.size();
  for(int i=0 ; i<n ; ++i)
    {
      trainingCase=cases[i];      

      for(int j=0 ; j<numInputs ; ++j)
	{
	  float x=trainingCase->getInput(j);
	  if(x<minInput[j])
	    minInput[j]=x;
	}

      for(int j=0 ; j<numOutputs ; ++j)
	{
	  float x=trainingCase->getOutput(j);
	  if(x<minOutput[j])
	    minOutput[j]=x;
	}
    }
}



void Normalizer::normalize(NeuralCase *trainingCase)
{
  for(int j=0 ; j<numInputs ; ++j)
    {
      float x=trainingCase->getInput(j);
      float range=maxInput[j]-minInput[j];
      if(range>0.0)
	x=(x-minInput[j])/range;
      else
	x=0.0;
      trainingCase->setInput(j,x);
    }
  
  for(int j=0 ; j<numOutputs ; ++j)
    {
      float x=trainingCase->getOutput(j);
      float range=maxOutput[j]-minOutput[j];
      if(range>0.0)
	x=(x-minOutput[j])/range;
      else
	x=0.0;
      trainingCase->setOutput(j,x);
    }
}



void Normalizer::adjust(float factor,
			DoubleVector &array,
			int n)
{
  int i;
  for(i=0 ; i<n ; ++i)
    array[i]=array[i]*factor;
}



float Normalizer::denormalizeOutput(float value,
				    int index)
{
  float range=maxOutput[index]-minOutput[index];
  return value*range+minOutput[index];
}



void Normalizer::printOn(ostream &os)
{
  os<<numInputs<<"\t"<<numOutputs<<endl;
  os<<maxInput<<endl;
  os<<maxOutput<<endl;
  os<<minInput<<endl;
  os<<minOutput<<endl;
}



void Normalizer::readFrom(istream &is)
{
  is>>numInputs>>numOutputs;
  is>>maxInput>>maxOutput>>minInput>>minOutput;
}



ostream &operator<<(ostream &os,Normalizer &n)
{
  n.printOn(os);
  return os;
}



istream &operator>>(istream &is,Normalizer &n)
{
  n.readFrom(is);
  return is;
}

