/****************************************************************
 net-classify.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/

#include <iostream>
#include <fstream>
#include <math.h>
#include <stdlib.h>
#include "BOOM/CommandLine.H"
#include "BOOM/Random.H"
#include "BOOM/ConfigFile.H"
#include "BOOM/Stacktrace.H"
#include "BOOM/Vector.H"
#include "BOOM/LinRegressor.H"
#include "BOOM/SummaryStats.H"
#include "BOOM/Vector.H"
#include "BOOM/Time.H"
#include "BOOM/Exceptions.H"
#include "NeuralNetwork.H"
#include "StandardNeuron.H"
#include "NeuralCase.H"
#include "Backprop.H"
#include "NetworkBuilder.H"
#include "Normalizer.H"
#include "../common/NamesFile.H"
using namespace std;

class Application
{
  String nameFile, netFile, outfile, dataFile;
  Vector<NeuralCase*> testCases;
  NamesFile namesFile;
  Schema *schema;
  Normalizer *normalizer;
  NeuralNetwork *nn;
  int numInputs, numOutputs, numCategories;

  bool loadData(String &filename,Vector<NeuralCase*> &);
  void testNetwork();
  void buildNeuralCases(Vector<NeuralCase*> &);
  void testNetwork(IntVector &outputIndices,ostream &);
  void applyCase(NeuralCase &c);
  void processConfigFile();
public:
  Application();
  void go(int argc,char *argv[]);
};



int main(int argc,char *argv[])
{
  try
    {
      Time stopwatch;
      stopwatch.startCounting();

      Application app;
      app.go(argc,argv);

      stopwatch.stopCounting();
      cout << stopwatch.elapsedTime() << endl;
      return 0;
    }
  catch(RootException &e)
    {
      cerr << e.getMessage() << endl;
      return -1;
    }
  catch(const char *msg)
    {
      cerr << msg << endl;
      return -1;
    }
  catch(char *msg)
    {
      cerr << msg << endl;
      return -1;
    }
  catch(string s)
    {
      cerr << s.c_str() << endl;
      return -1;
    }
  catch(...)
    {
      cerr << "unknown exception caught in main()" << endl;
      return -1;
    }
}



Application::Application()
  : normalizer(NULL) 
{
  // ctor
}



void Application::go(int argc,char *argv[])
{
  // Process the command line
  CommandLine cmd(argc,argv,"");
  if(cmd.numArgs()!=4) 
    throw "net-classify <model> <data> <names> <output>";
  netFile=cmd.arg(0);
  dataFile=cmd.arg(1);
  nameFile=cmd.arg(2);
  outfile=cmd.arg(3);

  // Load network
  schema=namesFile.processFile(nameFile);
  numCategories=schema->getNumCategories();
  nn=new NeuralNetwork();
  ifstream is(netFile.c_str());
  normalizer=new Normalizer();
  is>>*normalizer;
  nn->loadSelf(is);
  numInputs=nn->getNumInputs();
  numOutputs=nn->getNumOutputs();

  // Load data
  loadData(dataFile,testCases);
  normalizer->normalize(testCases);
  cout<<testCases.size()<<" cases loaded"<<endl;

  // Make predictions
  testNetwork();
}



bool Application::loadData(String &filename,Vector<NeuralCase*> &cases)
{
  String line;
  DoubleVector vect;
  ifstream is(filename.c_str());
  if(!is.good()) return false;
  while(!is.eof())
    {
      line.getline(is);
      if(line.isEmpty()) break;
      vect.clear();
      StrTokenizer tokenizer(line,", \t");
      while(tokenizer.hasMoreTokens())
	{
	  const char *token=tokenizer.nextToken();
	  double value=atof(token);
	  vect.push_back(value);
	}
      if(vect.isEmpty()) break;
      NeuralCase *c=new NeuralCase(numInputs,numCategories);
      for(int i=0 ; i<numInputs ; ++i)
	{
	  double value=vect[i];
	  c->setInput(i,(float) value);
	}
      testCases.push_back(c);
    }
  return true;
}



void Application::testNetwork()
{
  ofstream os(outfile.c_str());
  IntVector &outputIndices=nn->getOutputIndices();
  testNetwork(outputIndices,os);
}



void Application::testNetwork(IntVector &outputIndices,ostream &os)
{
  for(int i=0 ; i<testCases.size() ; ++i)
    {
      NeuralCase *c=(NeuralCase *) testCases[i];
      applyCase(*c);

      double largestValue=nn->getOutput(outputIndices[0]);
      int largestIndex=0;
      for(int j=1 ; j<numCategories ; ++j)
	{
	  double value=nn->getOutput(outputIndices[j]);
	  if(value>largestValue)
	    {
	      largestValue=value;
	      largestIndex=j;
	    }
	}
      os<<largestIndex<<endl;
    }
}



void Application::applyCase(NeuralCase &c)
{
  /* 
     This function simply applies the inputs of the 
     training case to the input neurons and then 
     propagates the network to compute its outputs.
     */

  // Apply inputs to input neurons
  IntVector &inputIndices=nn->getInputIndices();
  for(int i=0 ; i<inputIndices.size() ; ++i)
    {
      int index=inputIndices[i];
      float value=c.getInput(i);
      nn->setInput(index,value);
    }

  // Propagate the system
  nn->propagate();
}

