/****************************************************************
 train-net.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/

#include <iostream>
#include <fstream>
#include <math.h>
#include <stdlib.h>
#include "BOOM/CommandLine.H"
#include "BOOM/Random.H"
#include "BOOM/ConfigFile.H"
#include "BOOM/Stacktrace.H"
#include "BOOM/Vector.H"
#include "BOOM/LinRegressor.H"
#include "BOOM/SummaryStats.H"
#include "BOOM/Vector.H"
#include "BOOM/Time.H"
#include "BOOM/Exceptions.H"
#include "NeuralNetwork.H"
#include "StandardNeuron.H"
#include "NeuralCase.H"
#include "Backprop.H"
#include "NetworkBuilder.H"
#include "Normalizer.H"
#include "../common/NamesFile.H"
using namespace std;

class Application
{
  String dataFile, nameFile, configFile, outfile;
  Vector<NeuralCase*> trainingCases;
  DoubleVector outputRange;
  float noiseFactor;
  NamesFile namesFile;
  Schema *schema;

  bool shouldRandomize;
  Normalizer *normalizer;
  NeuralNetwork *nn;
  int maxIterations;
  float learningRate;
  int numLayers;
  int neuronsPerLayer;
  int numInputNodes;
  float maxAdj, minAdj;
  int numCategories;

  bool loadData(String &filename,Vector<NeuralCase*> &);
  void buildNetwork();
  void trainNetwork();
  void buildNeuralCases(Vector<NeuralCase*> &);
  void applyCase(NeuralCase &c);
  void processConfigFile();
public:
  Application();
  void go(int argc,char *argv[]);
};



int main(int argc,char *argv[])
{
  try
    {
      Time stopwatch;
      stopwatch.startCounting();

      Application app;
      app.go(argc,argv);

      stopwatch.stopCounting();
      cout << stopwatch.elapsedTime() << endl;
      return 0;
    }
  catch(RootException &e)
    {
      cerr << e.getMessage() << endl;
      return -1;
    }
  catch(const char *msg)
    {
      cerr << msg << endl;
      return -1;
    }
  catch(char *msg)
    {
      cerr << msg << endl;
      return -1;
    }
  catch(string s)
    {
      cerr << s.c_str() << endl;
      return -1;
    }
  catch(...)
    {
      cerr << "unknown exception caught in main()" << endl;
      return -1;
    }
}



Application::Application()
  : normalizer(NULL) 
{
  // ctor
}



void Application::go(int argc,char *argv[])
{
  CommandLine cmd(argc,argv,"");
  if(cmd.numArgs()!=4) 
    throw "train-net <data> <names> <config> <outfile>";
  dataFile=cmd.arg(0);
  nameFile=cmd.arg(1);
  configFile=cmd.arg(2);
  outfile=cmd.arg(3);

  schema=namesFile.processFile(nameFile);
  numCategories=schema->getNumCategories();

  processConfigFile();

  if(!loadData(dataFile,trainingCases)) throw "can't load training data";

  if(shouldRandomize) randomize();

  buildNetwork();

  trainNetwork();

  ofstream os(outfile.c_str());
  os<<*normalizer<<endl;
  nn->saveSelf(os);
}



bool Application::loadData(String &filename,
			   Vector<NeuralCase*> &cases)
{
  String line;
  DoubleVector vector;

  ifstream is(filename.c_str());
  if(!is.good()) return false;
  while(!is.eof())
    {
      line.getline(is);
      if(line.isEmpty()) break;

      vector.clear();

      StrTokenizer tokenizer(line,", \t");
      while(tokenizer.hasMoreTokens())
	{
	  const char *token=tokenizer.nextToken();
	  double value=atof(token);
	  vector.push_back(value);
	}
      if(vector.isEmpty()) break;

      int numInputs=vector.size()-1;

      NeuralCase *c=new NeuralCase(numInputs,numCategories);

      for(int i=0 ; i<numInputs ; ++i)
	{
	  double value=vector[i];
	  c->setInput(i,(float) value);
	}

      int category=int(vector[numInputs]);
      for(int i=0 ; i<numCategories ; ++i)
	c->setOutput(i,i==category ? 0.9 : 0.1);
      c->setCategory(category);

      cases.push_back(c);

      numInputNodes=numInputs;
    }

  return true;
}



void Application::processConfigFile()
{
  ConfigFile configFile(this->configFile);
  maxIterations=configFile.getIntOrDie("maxIterations");
  learningRate=configFile.getFloatOrDie("learningRate");
  numLayers=configFile.getIntOrDie("numLayers");
  neuronsPerLayer=configFile.getIntOrDie("neuronsPerLayer");
  minAdj=configFile.getFloatOrDie("min-adj");
  maxAdj=configFile.getFloatOrDie("max-adj");
  shouldRandomize=(configFile.getIntOrDie("randomize") ? true : false);
  noiseFactor=configFile.getFloatOrDie("noise-factor");
}




void Application::buildNetwork()
{
  nn=NetworkBuilder::buildNetwork(numInputNodes,
     numCategories,numLayers,neuronsPerLayer);
}



void Application::trainNetwork()
{
  Backpropagator trainer;

  normalizer=new Normalizer(numInputNodes,1); // ### 1 was 0
  normalizer->normalize(trainingCases,maxAdj,minAdj);

  LogisticFunction transferFunction;
  trainer.train(*nn,trainingCases,learningRate,transferFunction,
		maxIterations,cerr,noiseFactor);
}




void Application::applyCase(NeuralCase &c)
{
  /* 
     This function simply applies the inputs of the 
     training case to the input neurons and then 
     propagates the network to compute its outputs.
     */

  // Apply inputs to input neurons
  IntVector &inputIndices=nn->getInputIndices();
  int i;
  for(i=0 ; i<inputIndices.size() ; ++i)
    {
      int index=inputIndices[i];
      float value=c.getInput(i);

      nn->setInput(index,value);
    }

  // Propagate the system
  nn->propagate();
}

