/****************************************************************
 TreeNode.C
 william.majoros@duke.edu

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "TreeNode.H"
#include <iostream>
#include "BOOM/Random.H"
#include "IndividualSchema.H"
#include "ActivationRecord.H"
#include "Individual.H"
#include "Tableau.H"
#include "../../common/NamesFile.H"
using namespace std;
using namespace BOOM;


extern int numAttributes;
extern NamesFile namesFile;


#define __USE_ISOC9X 1
#include <math.h>

float ConstantNode::maxVal;
float ConstantNode::minVal;



TreeNode::TreeNode(int arity)
  : arity(arity)
{
  // ctor
}



TreeNode::~TreeNode()
{
  // dtor

  int i, numChildren=children.size();
  for(i=0 ; i<numChildren ; ++i)
    delete children[i];
}



int TreeNode::countNodes() const
{
  int num=1;
  int i, numChildren=children.size();
  for(i=0 ; i<numChildren ; ++i)
    num+=children[i]->countNodes();
  return num;
}



void TreeNode::save(NodeType nodeType,ostream &os) const
{
  int i, n=children.size();
  for(i=0 ; i<n ; ++i)
    children[i]->save(os);
  os << nodeType << endl;
  os << arity << endl;
  os << n << endl;
}



int TreeNode::getArity() const
{
  return arity;
}



void TreeNode::changeArity(int a)
{
  arity=a;
}



int TreeNode::getNumChildren() const
{
  return children.size();
}



TreeNode *TreeNode::getIthChild(int i) const
{
  return i<children.size() ? children[i] : NULL;
}



void TreeNode::setIthChild(int i,TreeNode *child)
{
  int n=children.size();
  if(i>=n)
    {
      int j;
      for(j=n ; j<=i ; ++j)
	children.push_back(NULL);
    }
  children[i]=child;
}



void TreeNode::addChild(TreeNode *child)
{
  children.push_back(child);
}



float TreeNode::leftVal(ActivationRecord &ar,Individual &ind)
{
  TreeNode *left=getIthChild(0);
  return left ? left->eval(ar,ind) : 0;
}



float TreeNode::rightVal(ActivationRecord &ar,Individual &ind)
{
  TreeNode *right=getIthChild(1);
  return right ? right->eval(ar,ind) : 0;
}



TreeNode *TreeNode::clone() const
{
  return instantiate();
}



TreeNode *TreeNode::deepCopy()
{
  TreeNode *twin=clone();

  int i, n=children.size();
  for(i=0 ; i<n ; ++i)
    twin->addChild(children[i]->deepCopy());

  return twin;
}



void TreeNode::printOn(ostream &os) const
{
  os << "TreeNode";
}



int TreeNode::getTreeHeight() const
{
  int numChildren=children.size();
  if(numChildren==0) return 0;

  int i, maxHeight=0;
  for(i=0 ; i<numChildren ; ++i)
    {
      int height=children[i]->getTreeHeight();
      if(height>maxHeight) maxHeight=height;
    }

  return maxHeight+1;
}



ostream &operator<<(ostream &os,const TreeNode &node)
{
  node.printOn(os);
  return os;
}


// class PlusNode ************************************

PlusNode::PlusNode()
  : TreeNode(2)
{
  // ctor
}



float PlusNode::eval(ActivationRecord &ar,Individual &ind)
{
  return leftVal(ar,ind)+rightVal(ar,ind);
}



TreeNode *PlusNode::instantiate() const
{
  return new PlusNode();
}



void PlusNode::printOn(ostream &os) const
{
  int n=getNumChildren();
  if(n==2)
    {
      os << "("<< *getIthChild(0) << "+" << *getIthChild(1) <<")";
    }
  else
    {
      os << "+(";
      for(int i=0 ; i<n ; ++i)
	{
	  os << *getIthChild(i);
	  if(i+1<n) os << ',';
	}
      os << ")";
    }
}



// class MinusNode ************************************

MinusNode::MinusNode()
  : TreeNode(2)
{
  // ctor
}



float MinusNode::eval(ActivationRecord &ar,Individual &ind)
{
  return leftVal(ar,ind)-rightVal(ar,ind);
}



TreeNode *MinusNode::instantiate() const
{
  return new MinusNode();
}



void MinusNode::printOn(ostream &os) const
{
  int n=getNumChildren();
  if(n==2)
    {
      os <<"("<< *getIthChild(0) << "-" << *getIthChild(1)<<")";
    }
  else
    {
      os << "-(";
      for(int i=0 ; i<n ; ++i)
	{
	  os << *getIthChild(i);
	  if(i+1<n) os << ',';
	}
      os << ")";
    }
}



// class TimesNode ************************************

TimesNode::TimesNode()
  : TreeNode(2)
{
  // ctor
}



float TimesNode::eval(ActivationRecord &ar,Individual &ind)
{
  return leftVal(ar,ind)*rightVal(ar,ind);
}



TreeNode *TimesNode::instantiate() const
{
  return new TimesNode();
}



void TimesNode::printOn(ostream &os) const
{
  int n=getNumChildren();
  if(n==2)
    {
      os <<"("<< *getIthChild(0) << "*" << *getIthChild(1)<<")";
    }
  else
    {
      os << "*(";
      for(int i=0 ; i<n ; ++i)
	{
	  os << *getIthChild(i);
	  if(i+1<n) os << ',';
	}
      os << ")";
    }
}



// class DivideNode ************************************

DivideNode::DivideNode()
  : TreeNode(2)
{
  // ctor
}



float DivideNode::eval(ActivationRecord &ar,Individual &ind)
{
  float denom=rightVal(ar,ind);
  return denom==0.0 ? leftVal(ar,ind) : leftVal(ar,ind)/denom;
}



TreeNode *DivideNode::instantiate() const
{
  return new DivideNode();
}



void DivideNode::printOn(ostream &os) const
{
  int n=getNumChildren();
  if(n==2)
    {
      os <<"("<< *getIthChild(0) << "/" << *getIthChild(1)<<")";
    }
  else
    {
      os << "/(";
      for(int i=0 ; i<n ; ++i)
	{
	  os << *getIthChild(i);
	  if(i+1<n) os << ',';
	}
      os << ")";
    }
}



// class SinNode ************************************

SinNode::SinNode()
  : TreeNode(1)
{
  // ctor
}



float SinNode::eval(ActivationRecord &ar,Individual &ind)
{
  return sin(leftVal(ar,ind));
}



TreeNode *SinNode::instantiate() const
{
  return new SinNode();
}



void SinNode::printOn(ostream &os) const
{
  os << "sin(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class CosNode ************************************

CosNode::CosNode()
  : TreeNode(1)
{
  // ctor
}



float CosNode::eval(ActivationRecord &ar,Individual &ind)
{
  return cos(leftVal(ar,ind));
}



TreeNode *CosNode::instantiate() const
{
  return new CosNode();
}



void CosNode::printOn(ostream &os) const
{
  os << "cos(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class ExpNode ************************************

ExpNode::ExpNode()
  : TreeNode(1)
{
  // ctor
}



float ExpNode::eval(ActivationRecord &ar,Individual &ind)
{
  return exp(leftVal(ar,ind));
}



TreeNode *ExpNode::instantiate() const
{
  return new ExpNode();
}



void ExpNode::printOn(ostream &os) const
{
  os << "exp(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class ConstantNode ************************************

ConstantNode::ConstantNode(float value)
  : TreeNode(0), value(value)
{
  // ctor
}



ConstantNode::ConstantNode()
  : TreeNode(0), value(RandomFloat(maxVal-minVal)+minVal)
{
  // ctor
}



ConstantNode::ConstantNode(float maxVal,float minVal)
  : TreeNode(0), value(0)
{
  // ctor

  ConstantNode::maxVal=maxVal;
  ConstantNode::minVal=minVal;
}



float ConstantNode::eval(ActivationRecord &,Individual &)
{
  return value;
}



TreeNode *ConstantNode::instantiate() const
{
  return new ConstantNode();
}



TreeNode *ConstantNode::clone() const
{
  return new ConstantNode(value);
}



void ConstantNode::printOn(ostream &os) const
{
  os << value;
}



// class VariableNode ************************************

VariableNode::VariableNode(int whichParm)
  : TreeNode(0), whichParm(whichParm)
{
  // ctor
}



VariableNode::VariableNode()
  : TreeNode(0), 
    whichParm(RandomNumber(LARGEST_RANDOM_NUMBER))
{
  // ctor
}



float VariableNode::eval(ActivationRecord &ar,Individual &)
{
  return ar[whichParm%ar.numVariables()];
}



TreeNode *VariableNode::instantiate() const
{
  return new VariableNode();
}



TreeNode *VariableNode::clone() const
{
  return new VariableNode(whichParm);
}



void VariableNode::printOn(ostream &os) const
{
  os << namesFile.getName(whichParm%numAttributes);
  //os << "var #" << whichParm;
}


// class IfNode ************************************

IfNode::IfNode()
  : TreeNode(3)
{
  // ctor
}



float IfNode::eval(ActivationRecord &ar,Individual &ind)
{
  // The comparison used to be: if(x>0) then y else z;
  // but it was changed subsequently to writing the paper.
  //  if(leftVal(ar,ind)>0)

  if(leftVal(ar,ind)!=0.0)
    return rightVal(ar,ind);
  
  TreeNode *then=getIthChild(2);
  return then ? then->eval(ar,ind) : 0;
}



TreeNode *IfNode::instantiate() const
{
  return new IfNode();
}



void IfNode::printOn(ostream &os) const
{
  os << "if(";
  if(getIthChild(0)) os << *getIthChild(0); else os << "NULL";
  os<< ") then ";
  if(getIthChild(1)) os << *getIthChild(1); else os << "NULL";
  os << " else ";
  if(getIthChild(2)) os << *getIthChild(2); else os << "NULL";
}


// class CallNode ************************************

CallNode::CallNode(int whichADF,int arity)
  : TreeNode(0), whichADF(whichADF)
{
  // ctor

  changeArity(arity);
}



CallNode::CallNode()
  : whichADF(RandomNumber(IndividualSchema::getGlobal().getNumADFs()))
{
  // ctor

  changeArity(IndividualSchema::getGlobal().getADFarity(whichADF));
}



float CallNode::eval(ActivationRecord &ar,Individual &ind)
{
  throw "shoult not happen";
}



TreeNode *CallNode::instantiate() const
{
  return new CallNode();
}



TreeNode *CallNode::clone() const
{
  const int arity=getArity();
  return new CallNode(whichADF,arity);
}



void CallNode::printOn(ostream &os) const
{
  os << "ADF" << whichADF << "()";
}



// class LogNode ************************************

LogNode::LogNode()
  : TreeNode(1)
{
  // ctor
}



float LogNode::eval(ActivationRecord &ar,Individual &ind)
{
  float parm=leftVal(ar,ind);
  return parm>0.0 ? log(parm) : 0.0 /*-FLT_MAX*/;
}



TreeNode *LogNode::instantiate() const
{
  return new LogNode();
}



void LogNode::printOn(ostream &os) const
{
  os << "log(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class MinNode ************************************

MinNode::MinNode()
  : TreeNode(2)
{
  // ctor
}



float MinNode::eval(ActivationRecord &ar,Individual &ind)
{
  float l=leftVal(ar,ind), r=rightVal(ar,ind);
  return l<r ? l : r;
}



TreeNode *MinNode::instantiate() const
{
  return new MinNode();
}



void MinNode::printOn(ostream &os) const
{
  os << "min(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class MaxNode ************************************

MaxNode::MaxNode()
  : TreeNode(2)
{
  // ctor
}



float MaxNode::eval(ActivationRecord &ar,Individual &ind)
{
  float l=leftVal(ar,ind), r=rightVal(ar,ind);
  return l<r ? l : r;
}



TreeNode *MaxNode::instantiate() const
{
  return new MaxNode();
}



void MaxNode::printOn(ostream &os) const
{
  os << "max(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class FmodNode ************************************

FmodNode::FmodNode()
  : TreeNode(2)
{
  // ctor
}



float FmodNode::eval(ActivationRecord &ar,Individual &ind)
{
  return fmod(leftVal(ar,ind),rightVal(ar,ind));
}



TreeNode *FmodNode::instantiate() const
{
  return new FmodNode();
}



void FmodNode::printOn(ostream &os) const
{
  os << "fmod(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class AbsNode ************************************

AbsNode::AbsNode()
  : TreeNode(1)
{
  // ctor
}



float AbsNode::eval(ActivationRecord &ar,Individual &ind)
{
  return fabs(leftVal(ar,ind));
}



TreeNode *AbsNode::instantiate() const
{
  return new AbsNode();
}



void AbsNode::printOn(ostream &os) const
{
  os << "abs(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class CeilNode ************************************

CeilNode::CeilNode()
  : TreeNode(1)
{
  // ctor
}



float CeilNode::eval(ActivationRecord &ar,Individual &ind)
{
  return ceil(leftVal(ar,ind));
}



TreeNode *CeilNode::instantiate() const
{
  return new CeilNode();
}



void CeilNode::printOn(ostream &os) const
{
  os << "ceil(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class HypotNode ************************************

HypotNode::HypotNode()
  : TreeNode(2)
{
  // ctor
}



float HypotNode::eval(ActivationRecord &ar,Individual &ind)
{
  return hypot(leftVal(ar,ind),rightVal(ar,ind));
}



TreeNode *HypotNode::instantiate() const
{
  return new HypotNode();
}



void HypotNode::printOn(ostream &os) const
{
  os << "hypot(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class TanhNode ************************************

TanhNode::TanhNode()
  : TreeNode(1)
{
  // ctor
}



float TanhNode::eval(ActivationRecord &ar,Individual &ind)
{
  return tanh(leftVal(ar,ind));
}



TreeNode *TanhNode::instantiate() const
{
  return new TanhNode();
}



void TanhNode::printOn(ostream &os) const
{
  os << "tanh(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class SinhNode ************************************

SinhNode::SinhNode()
  : TreeNode(1)
{
  // ctor
}



float SinhNode::eval(ActivationRecord &ar,Individual &ind)
{
  return sinh(leftVal(ar,ind));
}



TreeNode *SinhNode::instantiate() const
{
  return new SinhNode();
}



void SinhNode::printOn(ostream &os) const
{
  os << "sinh(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class CoshNode ************************************

CoshNode::CoshNode()
  : TreeNode(1)
{
  // ctor
}



float CoshNode::eval(ActivationRecord &ar,Individual &ind)
{
  return cosh(leftVal(ar,ind));
}



TreeNode *CoshNode::instantiate() const
{
  return new CoshNode();
}



void CoshNode::printOn(ostream &os) const
{
  os << "cosh(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class AtanhNode ************************************

AtanhNode::AtanhNode()
  : TreeNode(1)
{
  // ctor
}



float AtanhNode::eval(ActivationRecord &ar,Individual &ind)
{
  float x=leftVal(ar,ind);
  return fabs(x)>1 ? x : atanh(x);
}



TreeNode *AtanhNode::instantiate() const
{
  return new AtanhNode();
}



void AtanhNode::printOn(ostream &os) const
{
  os << "atanh(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class AsinhNode ************************************

AsinhNode::AsinhNode()
  : TreeNode(1)
{
  // ctor
}



float AsinhNode::eval(ActivationRecord &ar,Individual &ind)
{
  return asinh(leftVal(ar,ind));
}



TreeNode *AsinhNode::instantiate() const
{
  return new AsinhNode();
}



void AsinhNode::printOn(ostream &os) const
{
  os << "asinh(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class AcoshNode ************************************

AcoshNode::AcoshNode()
  : TreeNode(1)
{
  // ctor
}



float AcoshNode::eval(ActivationRecord &ar,Individual &ind)
{
  float x=leftVal(ar,ind);
  return x<1.0 ? x : acosh(x);
}



TreeNode *AcoshNode::instantiate() const
{
  return new AcoshNode();
}



void AcoshNode::printOn(ostream &os) const
{
  os << "acosh(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class Log2Node ************************************

Log2Node::Log2Node()
  : TreeNode(1)
{
  // ctor
}



float Log2Node::eval(ActivationRecord &ar,Individual &ind)
{
  float x=leftVal(ar,ind);
  return x<0.0 ? 0 : log(x)/log(2.0);
}



TreeNode *Log2Node::instantiate() const
{
  return new Log2Node();
}



void Log2Node::printOn(ostream &os) const
{
  os << "log2(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class Log10Node ************************************

Log10Node::Log10Node()
  : TreeNode(1)
{
  // ctor
}



float Log10Node::eval(ActivationRecord &ar,Individual &ind)
{
  return log10(leftVal(ar,ind));
}



TreeNode *Log10Node::instantiate() const
{
  return new Log10Node();
}



void Log10Node::printOn(ostream &os) const
{
  os << "log10(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class CbrtNode ************************************

CbrtNode::CbrtNode()
  : TreeNode(1)
{
  // ctor
}



float CbrtNode::eval(ActivationRecord &ar,Individual &ind)
{
  return cbrt(leftVal(ar,ind));
}



TreeNode *CbrtNode::instantiate() const
{
  return new CbrtNode();
}



void CbrtNode::printOn(ostream &os) const
{
  os << "cbrt(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class SqrtNode ************************************

SqrtNode::SqrtNode()
  : TreeNode(1)
{
  // ctor
}



float SqrtNode::eval(ActivationRecord &ar,Individual &ind)
{
  float x=leftVal(ar,ind);
  return x<0.0 ? 0 : sqrt(x);
}



TreeNode *SqrtNode::instantiate() const
{
  return new SqrtNode();
}



void SqrtNode::printOn(ostream &os) const
{
  os << "sqrt(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class AtanNode ************************************

AtanNode::AtanNode()
  : TreeNode(1)
{
  // ctor
}



float AtanNode::eval(ActivationRecord &ar,Individual &ind)
{
  return atan(leftVal(ar,ind));
}



TreeNode *AtanNode::instantiate() const
{
  return new AtanNode();
}



void AtanNode::printOn(ostream &os) const
{
  os << "atan(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class AsinNode ************************************

AsinNode::AsinNode()
  : TreeNode(1)
{
  // ctor
}



float AsinNode::eval(ActivationRecord &ar,Individual &ind)
{
  float x=leftVal(ar,ind);
  return fabs(x)>1 ? 0 : asin(x);
}



TreeNode *AsinNode::instantiate() const
{
  return new AsinNode();
}



void AsinNode::printOn(ostream &os) const
{
  os << "asin(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class AcosNode ************************************

AcosNode::AcosNode()
  : TreeNode(1)
{
  // ctor
}



float AcosNode::eval(ActivationRecord &ar,Individual &ind)
{
  float x=leftVal(ar,ind);
  return fabs(x)>1.0 ? 0 : acos(x);
}



TreeNode *AcosNode::instantiate() const
{
  return new AcosNode();
}



void AcosNode::printOn(ostream &os) const
{
  os << "acos(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class TanNode ************************************

TanNode::TanNode()
  : TreeNode(1)
{
  // ctor
}



float TanNode::eval(ActivationRecord &ar,Individual &ind)
{
  return tan(leftVal(ar,ind));
}



TreeNode *TanNode::instantiate() const
{
  return new TanNode();
}



void TanNode::printOn(ostream &os) const
{
  os << "tan(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class PowNode ************************************

PowNode::PowNode()
  : TreeNode(2)
{
  // ctor
}



float PowNode::eval(ActivationRecord &ar,Individual &ind)
{
  return pow(leftVal(ar,ind),rightVal(ar,ind));
}



TreeNode *PowNode::instantiate() const
{
  return new PowNode();
}



void PowNode::printOn(ostream &os) const
{
  os << "pow(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class ModNode ************************************

ModNode::ModNode()
  : TreeNode(2)
{
  // ctor
}



float ModNode::eval(ActivationRecord &ar,Individual &ind)
{
  float l=leftVal(ar,ind), r=rightVal(ar,ind);
  int intR=(int) r;

  //if((l!=l) || (l==INFINITY) || (l==-INFINITY)) return 0; // NaN
  switch(fpclassify(l))
    {
    case FP_NAN:
    case FP_INFINITE:
      return 0;
    }

  return ((intR==0) ? l : float(int(l)%intR));
}



TreeNode *ModNode::instantiate() const
{
  return new ModNode();
}



void ModNode::printOn(ostream &os) const
{
  int n=getNumChildren();
  if(n==2)
    {
      os <<"("<< *getIthChild(0) << "%" << *getIthChild(1)<<")";
    }
  else
    {
      os << "%(";
      for(int i=0 ; i<n ; ++i)
	{
	  os << *getIthChild(i);
	  if(i+1<n) os << ',';
	}
      os << ")";
    }
}



// class AndNode ************************************

AndNode::AndNode()
  : TreeNode(2)
{
  // ctor
}



float AndNode::eval(ActivationRecord &ar,Individual &ind)
{
  return int(leftVal(ar,ind)) && int(rightVal(ar,ind));
}



TreeNode *AndNode::instantiate() const
{
  return new AndNode();
}



void AndNode::printOn(ostream &os) const
{
  int n=getNumChildren();
  if(n==2)
    {
      os <<"("<< *getIthChild(0) << "&&" << *getIthChild(1)<<")";
    }
  else
    {
      os << "&&(";
      for(int i=0 ; i<n ; ++i)
	{
	  os << *getIthChild(i);
	  if(i+1<n) os << ',';
	}
      os << ")";
    }
}



// class OrNode ************************************

OrNode::OrNode()
  : TreeNode(2)
{
  // ctor
}



float OrNode::eval(ActivationRecord &ar,Individual &ind)
{
  return int(leftVal(ar,ind)) || int(rightVal(ar,ind));
}



TreeNode *OrNode::instantiate() const
{
  return new OrNode();
}



void OrNode::printOn(ostream &os) const
{
  int n=getNumChildren();
  if(n==2)
    {
      os <<"("<< *getIthChild(0) << "||" << *getIthChild(1)<<")";
    }
  else
    {
      os << "||(";
      for(int i=0 ; i<n ; ++i)
	{
	  os << *getIthChild(i);
	  if(i+1<n) os << ',';
	}
      os << ")";
    }
}



// class NotEqualsNode ************************************

NotEqualsNode::NotEqualsNode()
  : TreeNode(2)
{
  // ctor
}



float NotEqualsNode::eval(ActivationRecord &ar,Individual &ind)
{
  return float(leftVal(ar,ind)!=rightVal(ar,ind));
}



TreeNode *NotEqualsNode::instantiate() const
{
  return new NotEqualsNode();
}



void NotEqualsNode::printOn(ostream &os) const
{
  int n=getNumChildren();
  if(n==2)
    {
      os <<"("<< *getIthChild(0) << "!=" << *getIthChild(1)<<")";
    }
  else
    {
      os << "!=(";
      for(int i=0 ; i<n ; ++i)
	{
	  os << *getIthChild(i);
	  if(i+1<n) os << ',';
	}
      os << ")";
    }
}



// class GreaterEqualsNode ************************************

GreaterEqualsNode::GreaterEqualsNode()
  : TreeNode(2)
{
  // ctor
}



float GreaterEqualsNode::eval(ActivationRecord &ar,Individual &ind)
{
  return leftVal(ar,ind) >= rightVal(ar,ind);
}



TreeNode *GreaterEqualsNode::instantiate() const
{
  return new GreaterEqualsNode();
}



void GreaterEqualsNode::printOn(ostream &os) const
{
  int n=getNumChildren();
  if(n==2)
    {
      os <<"("<< *getIthChild(0) << ">=" << *getIthChild(1)<<")";
    }
  else
    {
      os << ">=(";
      for(int i=0 ; i<n ; ++i)
	{
	  os << *getIthChild(i);
	  if(i+1<n) os << ',';
	}
      os << ")";
    }
}



// class LessEqualsNode ************************************

LessEqualsNode::LessEqualsNode()
  : TreeNode(2)
{
  // ctor
}



float LessEqualsNode::eval(ActivationRecord &ar,Individual &ind)
{
  return leftVal(ar,ind) <= rightVal(ar,ind);
}



TreeNode *LessEqualsNode::instantiate() const
{
  return new LessEqualsNode();
}



void LessEqualsNode::printOn(ostream &os) const
{
  int n=getNumChildren();
  if(n==2)
    {
      os <<"("<< *getIthChild(0) << "<=" << *getIthChild(1)<<")";
    }
  else
    {
      os << "<=(";
      for(int i=0 ; i<n ; ++i)
	{
	  os << *getIthChild(i);
	  if(i+1<n) os << ',';
	}
      os << ")";
    }
}



// class NotNode ************************************

NotNode::NotNode()
  : TreeNode(1)
{
  // ctor
}



float NotNode::eval(ActivationRecord &ar,Individual &ind)
{
  return ! (int) leftVal(ar,ind);
}



TreeNode *NotNode::instantiate() const
{
  return new NotNode();
}



void NotNode::printOn(ostream &os) const
{
  os << "!(";
  int i, n=getNumChildren();
  for(i=0 ; i<n ; ++i)
    {
      os << *getIthChild(i);
      if(i+1<n) os << ',';
    }
  os << ")";
}



// class LessNode ************************************

LessNode::LessNode()
  : TreeNode(2)
{
  // ctor
}



float LessNode::eval(ActivationRecord &ar,Individual &ind)
{
  return leftVal(ar,ind) < rightVal(ar,ind);
}



TreeNode *LessNode::instantiate() const
{
  return new LessNode();
}



void LessNode::printOn(ostream &os) const
{
  int n=getNumChildren();
  if(n==2)
    {
      os <<"("<< *getIthChild(0) << "<" << *getIthChild(1)<<")";
    }
  else
    {
      os << "<(";
      for(int i=0 ; i<n ; ++i)
	{
	  os << *getIthChild(i);
	  if(i+1<n) os << ',';
	}
      os << ")";
    }
}



// class GreaterNode ************************************

GreaterNode::GreaterNode()
  : TreeNode(2)
{
  // ctor
}



float GreaterNode::eval(ActivationRecord &ar,Individual &ind)
{
  return leftVal(ar,ind) > rightVal(ar,ind);
}



TreeNode *GreaterNode::instantiate() const
{
  return new GreaterNode();
}



void GreaterNode::printOn(ostream &os) const
{
  int n=getNumChildren();
  if(n==2)
    {
      os <<"("<< *getIthChild(0) << ">" << *getIthChild(1)<<")";
    }
  else
    {
      os << ">(";
      for(int i=0 ; i<n ; ++i)
	{
	  os << *getIthChild(i);
	  if(i+1<n) os << ',';
	}
      os << ")";
    }
}



// class EqualsNode ************************************

EqualsNode::EqualsNode()
  : TreeNode(2)
{
  // ctor
}



float EqualsNode::eval(ActivationRecord &ar,Individual &ind)
{
  return leftVal(ar,ind) == rightVal(ar,ind);
}



TreeNode *EqualsNode::instantiate() const
{
  return new EqualsNode();
}



void EqualsNode::printOn(ostream &os) const
{
  int n=getNumChildren();
  if(n==2)
    {
      os <<"("<< *getIthChild(0) << "==" << *getIthChild(1)<<")";
    }
  else
    {
      os << "==(";
      for(int i=0 ; i<n ; ++i)
	{
	  os << *getIthChild(i);
	  if(i+1<n) os << ',';
	}
      os << ")";
    }
}


// ***************** MISC VIRTUAL METHODS *********************


void MinNode::save(ostream &os) const
{
  TreeNode::save(TN_MIN,os);
}



void MaxNode::save(ostream &os) const
{
  TreeNode::save(TN_MAX,os);
}



void FmodNode::save(ostream &os) const
{
  TreeNode::save(TN_FMOD,os);
}



void AbsNode::save(ostream &os) const
{
  TreeNode::save(TN_ABS,os);
}



void CeilNode::save(ostream &os) const
{
  TreeNode::save(TN_CEIL,os);
}



void HypotNode::save(ostream &os) const
{
  TreeNode::save(TN_HYPOT,os);
}



void TanhNode::save(ostream &os) const
{
  TreeNode::save(TN_TANH,os);
}



void SinhNode::save(ostream &os) const
{
  TreeNode::save(TN_SINH,os);
}



void CoshNode::save(ostream &os) const
{
  TreeNode::save(TN_COSH,os);
}



void AtanhNode::save(ostream &os) const
{
  TreeNode::save(TN_ATANH,os);
}



void AsinhNode::save(ostream &os) const
{
  TreeNode::save(TN_ASINH,os);
}



void AcoshNode::save(ostream &os) const
{
  TreeNode::save(TN_ACOSH,os);
}



void Log2Node::save(ostream &os) const
{
  TreeNode::save(TN_LOG2,os);
}



void Log10Node::save(ostream &os) const
{
  TreeNode::save(TN_LOG10,os);
}



void CbrtNode::save(ostream &os) const
{
  TreeNode::save(TN_CBRT,os);
}



void SqrtNode::save(ostream &os) const
{
  TreeNode::save(TN_SQRT,os);
}



void AtanNode::save(ostream &os) const
{
  TreeNode::save(TN_ATAN,os);
}



void AsinNode::save(ostream &os) const
{
  TreeNode::save(TN_ASIN,os);
}



void AcosNode::save(ostream &os) const
{
  TreeNode::save(TN_ACOS,os);
}



void TanNode::save(ostream &os) const
{
  TreeNode::save(TN_TAN,os);
}



void PowNode::save(ostream &os) const
{
  TreeNode::save(TN_POW,os);
}



void ModNode::save(ostream &os) const
{
  TreeNode::save(TN_MOD,os);
}



void AndNode::save(ostream &os) const
{
  TreeNode::save(TN_AND,os);
}



void OrNode::save(ostream &os) const
{
  TreeNode::save(TN_OR,os);
}



void NotEqualsNode::save(ostream &os) const
{
  TreeNode::save(TN_NOTEQUALS,os);
}



void GreaterEqualsNode::save(ostream &os) const
{
  TreeNode::save(TN_GREATEREQUALS,os);
}



void LessEqualsNode::save(ostream &os) const
{
  TreeNode::save(TN_LESSEQUALS,os);
}



void NotNode::save(ostream &os) const
{
  TreeNode::save(TN_NOT,os);
}



void LessNode::save(ostream &os) const
{
  TreeNode::save(TN_LESS,os);
}



void GreaterNode::save(ostream &os) const
{
  TreeNode::save(TN_GREATER,os);
}



void PlusNode::save(ostream &os) const
{
  TreeNode::save(TN_PLUS,os);
}



void MinusNode::save(ostream &os) const
{
  TreeNode::save(TN_MINUS,os);
}



void TimesNode::save(ostream &os) const
{
  TreeNode::save(TN_TIMES,os);
}



void DivideNode::save(ostream &os) const
{
  TreeNode::save(TN_DIVIDE,os);
}



void SinNode::save(ostream &os) const
{
  TreeNode::save(TN_SIN,os);
}



void CosNode::save(ostream &os) const
{
  TreeNode::save(TN_COS,os);
}



void ExpNode::save(ostream &os) const
{
  TreeNode::save(TN_EXP,os);
}



void ConstantNode::save(ostream &os) const
{
  TreeNode::save(TN_CONSTANT,os);
  os << value << endl;
}



void VariableNode::save(ostream &os) const
{
  TreeNode::save(TN_VARIABLE,os);
  os << whichParm << endl;
}



void IfNode::save(ostream &os) const
{
  TreeNode::save(TN_IF,os);
}



void CallNode::save(ostream &os) const
{
  TreeNode::save(TN_CALL,os);
  os << whichADF << endl;
}



void LogNode::save(ostream &os) const
{
  TreeNode::save(TN_LOG,os);
}



void EqualsNode::save(ostream &os) const
{
  TreeNode::save(TN_EQUALS,os);
}



// class GreaterThanZeroNode ************************************

GreaterThanZeroNode::GreaterThanZeroNode()
  : TreeNode(1)
{
  // ctor
}



float GreaterThanZeroNode::eval(ActivationRecord &ar,Individual &ind)
{
  return leftVal(ar,ind)>0;
}



TreeNode *GreaterThanZeroNode::instantiate() const
{
  return new GreaterThanZeroNode();
}



void GreaterThanZeroNode::printOn(ostream &os) const
{
  int n=getNumChildren();
  if(n==1)
    {
      os <<"("<< *getIthChild(0) << ">0)";
    }
  else
    {
      os << "greater_than_zero(";
      for(int i=0 ; i<n ; ++i)
	{
	  os << *getIthChild(i);
	  if(i+1<n) os << ',';
	}
      os << ")";
    }
}



void GreaterThanZeroNode::save(ostream &os) const
{
  TreeNode::save(TN_GREATER_ZERO,os);
}



// class GreaterThanOneNode ************************************

GreaterThanOneNode::GreaterThanOneNode()
  : TreeNode(1)
{
  // ctor
}



float GreaterThanOneNode::eval(ActivationRecord &ar,Individual &ind)
{
  return leftVal(ar,ind)>1;
}



TreeNode *GreaterThanOneNode::instantiate() const
{
  return new GreaterThanOneNode();
}



void GreaterThanOneNode::printOn(ostream &os) const
{
  int n=getNumChildren();
  if(n==1)
    {
      os <<"("<< *getIthChild(0) << ">1)";
    }
  else
    {
      os << "greater_than_one(";
      for(int i=0 ; i<n ; ++i)
	{
	  os << *getIthChild(i);
	  if(i+1<n) os << ',';
	}
      os << ")";
    }
}



void GreaterThanOneNode::save(ostream &os) const
{
  TreeNode::save(TN_GREATER_ONE,os);
}



// class EqualsZeroNode ************************************

EqualsZeroNode::EqualsZeroNode()
  : TreeNode(1)
{
  // ctor
}



float EqualsZeroNode::eval(ActivationRecord &ar,Individual &ind)
{
  return leftVal(ar,ind)==0;
}



TreeNode *EqualsZeroNode::instantiate() const
{
  return new EqualsZeroNode();
}



void EqualsZeroNode::printOn(ostream &os) const
{
  int n=getNumChildren();
  if(n==1)
    {
      os <<"("<< *getIthChild(0) << "==0)";
    }
  else
    {
      os << "equals_zero(";
      for(int i=0 ; i<n ; ++i)
	{
	  os << *getIthChild(i);
	  if(i+1<n) os << ',';
	}
      os << ")";
    }
}



void EqualsZeroNode::save(ostream &os) const
{
  TreeNode::save(TN_EQUALS_ZERO,os);
}



