/****************************************************************
 progen.C
 william.majoros@duke.edu

 Random classification-problem generator

 Copyright 2005 William H. Majoros.  This is open-source software,
 governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "Framework/Tableau.H"
#include "Framework/IndividualSchema.H"
#include "Framework/SymbolSet.H"
#include "Framework/ClassifiedAR.H"
#include "Framework/PopulationBuilder.H"
#include "../common/NamesFile.H"
#include "../common/TrainingSet.H"
#include "BOOM/StrTokenizer.H"
#include "BOOM/Random.H"
#include "BOOM/CommandLine.H"
#include "BOOM/ConfigFile.H"
#include "BOOM/Constants.H"
#include "BOOM/Set.H"
#include "BOOM/Array2D.H"
#include "BOOM/Array1D.H"
#include <iostream>
#include <stdlib.h>
#include <fstream>
using namespace std;


int numAttributes;
NamesFile namesFile;


class Application
{
  ConfigFile *configFile;
  String filestem, configFilename;
  SymbolSet symbolSet;
  IndividualBuilder *individualBuilder;
  Population *population;
  Vector<ClassifiedAR*> trainingSet, testSet;
  int numCategories;
  int maxHeight;
  int numUnobservables;
  int numTrainPoints, numTestPoints;

  void setupSymbols();
  void generateUnobservables(BOOM::Array2D<float> &testPoints,
			     int numPoints);
  void partition(int numCases,BOOM::Array1D<int> &);
  void buildARs(BOOM::Array2D<float> &testPoints,
		BOOM::Vector<ClassifiedAR*> &ARs);
  bool applyTree(TreeNode &,Individual &,BOOM::Array2D<float> &points,
		 BOOM::Vector<ClassifiedAR*> &ARs,
		 BOOM::Array1D<int> &partitions,int varIndex,int cat);
  void outputTree(ostream &,TreeNode &);
  void outputData(BOOM::Array2D<float> &points,
		  BOOM::Array1D<int> &partitions,BOOM::String &filename);
  bool isOK(BOOM::Array2D<float> &points,int beginRow,int endRow,
	    int varIndex);
public:
  int go(int argc,char *argv[]);
};


int main(int argc,char *argv[])
{
  try
    {
      Application app;
      return app.go(argc,argv);
    }
  catch(const String &msg)
    {
      cerr << msg.c_str() << endl;
      return -1;
    }
  catch(const char *msg)
    {
      cerr << msg << endl;
      return -1;
    }
  catch(...)
    {
      cerr<<"unknown exception caught in main()"<<endl;
      return -1;
    }
}



int Application::go(int argc,char *argv[])
{
  // Process the command line
  CommandLine cmd(argc,argv,"tc:");
  if(cmd.numArgs()!=7)
    throw "\nprogen <config-file> <filestem> <#categories> <#observables> <#unobservables> <#training-points> <#test-points>\n\n";
  configFilename=cmd.arg(0);
  filestem=cmd.arg(1);
  numCategories=cmd.arg(2).asInt();
  numAttributes=cmd.arg(3).asInt();
  numUnobservables=cmd.arg(4).asInt();
  numTrainPoints=cmd.arg(5).asInt();
  numTestPoints=cmd.arg(6).asInt();

  // Process the configuration file & do various initializations
  configFile=new ConfigFile(configFilename);
  Tableau::processConfig(*configFile);
  randomize();
  maxHeight=Tableau::getGlobal().getMaxHeight();
  IndividualSchema::allocateGlobal(0);
  IndividualSchema::getGlobal().addEntryPoint(0);
  setupSymbols();

  // Generate the unobservables for all train/test points
  int totalVariables=numAttributes+numUnobservables;
  BOOM::Array2D<float> trainPoints(numTrainPoints,totalVariables);
  BOOM::Array2D<float> testPoints(numTestPoints,totalVariables);
  generateUnobservables(trainPoints,numTrainPoints);
  generateUnobservables(testPoints,numTestPoints);

  // Construct activation records for use later during tree evaluation
  BOOM::Vector<ClassifiedAR*> trainARs, testARs;
  buildARs(trainPoints,trainARs);
  buildARs(testPoints,testARs);

  // Partition the training/test sets according to category
  BOOM::Array1D<int> trainPartitionBoundaries(numCategories+1);
  BOOM::Array1D<int> testPartitionBoundaries(numCategories+1);
  partition(numTrainPoints,trainPartitionBoundaries);
  partition(numTestPoints,testPartitionBoundaries);

  // Generate the *.names file
  BOOM::String trainFile=filestem+".data";
  BOOM::String testFile=filestem+".test";
  BOOM::String nameFile=filestem+".names";
  ofstream os(nameFile.c_str());
  os<<numCategories<<" categories\n";
  for(int i=0 ; i<numAttributes ; ++i)
    {
      os<<"var_"<<i<<": continuous\n";
    }
  for(int i=0 ; i<numUnobservables ; ++i)
    namesFile.addName(String("x")+i);

  // Generate the model for each category independently
  BOOM::String modelFile=filestem+".trees";
  ofstream osM(modelFile.c_str());
  individualBuilder=new IndividualBuilder(symbolSet);
  for(int cat=0 ; cat<numCategories ; ++cat)
    {
      osM<<"category "<<cat<<":"<<endl;

      // Generate a tree for each observable variable
      int retries=0;
      for(int i=0 ; i<numAttributes ; ++i)
	{
	  ++retries;
	  if(retries>10000) throw "too many iterations!";

	  // Generate the tree
	  int height=RandomNumber(maxHeight)+1;
	  Individual *individual=individualBuilder->buildIndividual(height);
	  TreeNode *tree=individual->getEntryPoint(0);

	  // Use the tree to generate values for the variable
	  bool ok1=
	    applyTree(*tree,*individual,trainPoints,trainARs,
		      trainPartitionBoundaries,i,cat);
	  bool ok2=
	    applyTree(*tree,*individual,testPoints,testARs,
		      testPartitionBoundaries,i,cat);
	  if(ok1 && ok2)
	    {
	      // Output the tree
	      osM<<"var_"<<i<<" = ";
	      int t=numAttributes;
	      numAttributes=numUnobservables;
	      outputTree(osM,*tree);
	      numAttributes=t;
	      retries=0;
	    }
	  else --i;

	  delete individual;
	}
      osM<<endl;
    }

  // Generate the data files
  outputData(trainPoints,trainPartitionBoundaries,trainFile);
  outputData(testPoints,testPartitionBoundaries,testFile);
}



void Application::buildARs(BOOM::Array2D<float> &points,
			   BOOM::Vector<ClassifiedAR*> &ARs)
{
  int n=points.getFirstDim();
  TrainingCase c(numUnobservables);
  for(int i=0 ; i<n ; ++i)
    {
      BOOM::Array2D<float>::RowIn2DArray<float> row=points[i];
      for(int j=0 ; j<numUnobservables ; ++j)
	c.setContinuous(j,row[j]);
      ClassifiedAR *ar=new ClassifiedAR(c);
      ARs.push_back(ar);
    }
}



void Application::outputTree(ostream &os,TreeNode &tree)
{
  os<<tree<<endl;
}



void Application::outputData(BOOM::Array2D<float> &points,
			     BOOM::Array1D<int> &partitions,
			     BOOM::String &filename)
{
  ofstream os(filename.c_str());
  int v=points.getSecondDim();
  for(int i=0 ; i<numCategories ; ++i)
    {
      int begin=partitions[i], end=partitions[i+1];
      for(int j=begin ; j<end ; ++j)
	{
	  BOOM::Array2D<float>::RowIn2DArray<float> row=points[j];
	  for(int k=numUnobservables ; k<v ; ++k)
	    os<<row[k]<<"\t";
	  os<<i<<endl;
	}
    }
}



bool Application::applyTree(TreeNode &tree,
			    Individual &ind,
			    BOOM::Array2D<float> &points,
			    BOOM::Vector<ClassifiedAR*> &ARs,
			    BOOM::Array1D<int> &partitions,
			    int varIndex,
			    int cat)
{
  int indexInPoint=varIndex+numUnobservables;
  int begin=partitions[cat], end=partitions[cat+1];
  for(int i=begin ; i<end ; ++i)
    {
      ClassifiedAR *ar=ARs[i];
      float value=tree.eval(ar->getAR(),ind);
      points[i][indexInPoint]=value;
    }
  return isOK(points,begin,end,indexInPoint);
}



void Application::setupSymbols()
{
  // Nonterminals
  String symbols=Tableau::getGlobal().getNonterminals();
  StrTokenizer tokenizer(symbols,",");
  while(tokenizer.hasMoreTokens())
    {
      const char *token=tokenizer.nextToken();
      if(!strcmp(token,"+")) 
	symbolSet.addNonterminal(new PlusNode());
      else if(!strcmp(token,"-")) 
	symbolSet.addNonterminal(new MinusNode());
      else if(!strcmp(token,"*")) 
	symbolSet.addNonterminal(new TimesNode());
      else if(!strcmp(token,"/")) 
	symbolSet.addNonterminal(new DivideNode());
      else if(!strcmp(token,"sin")) 
	symbolSet.addNonterminal(new SinNode());
      else if(!strcmp(token,"cos")) 
	symbolSet.addNonterminal(new CosNode());
      else if(!strcmp(token,"exp")) 
	symbolSet.addNonterminal(new ExpNode());
      else if(!strcmp(token,"log")) 
	symbolSet.addNonterminal(new LogNode());
      else if(!strcmp(token,"if")) 
	symbolSet.addNonterminal(new IfNode());
      else if(!strcmp(token,"min")) 
	symbolSet.addNonterminal(new MinNode());
      else if(!strcmp(token,"max")) 
	symbolSet.addNonterminal(new MaxNode());
      else if(!strcmp(token,"fmod")) 
	symbolSet.addNonterminal(new FmodNode());
      else if(!strcmp(token,"abs")) 
	symbolSet.addNonterminal(new AbsNode());
      else if(!strcmp(token,"ceil")) 
	symbolSet.addNonterminal(new CeilNode());
      else if(!strcmp(token,"hypot")) 
	symbolSet.addNonterminal(new HypotNode());
      else if(!strcmp(token,"tanh")) 
	symbolSet.addNonterminal(new TanhNode());
      else if(!strcmp(token,"sinh")) 
	symbolSet.addNonterminal(new SinhNode());
      else if(!strcmp(token,"cosh")) 
	symbolSet.addNonterminal(new CoshNode());
      else if(!strcmp(token,"atanh")) 
	symbolSet.addNonterminal(new AtanhNode());
      else if(!strcmp(token,"asinh")) 
	symbolSet.addNonterminal(new AsinhNode());
      else if(!strcmp(token,"acosh")) 
	symbolSet.addNonterminal(new AcoshNode());
      else if(!strcmp(token,"log2")) 
	symbolSet.addNonterminal(new Log2Node());
      else if(!strcmp(token,"log10")) 
	symbolSet.addNonterminal(new Log10Node());
      else if(!strcmp(token,"cbrt")) 
	symbolSet.addNonterminal(new CbrtNode());
      else if(!strcmp(token,"sqrt")) 
	symbolSet.addNonterminal(new CallNode());
      else if(!strcmp(token,"atan")) 
	symbolSet.addNonterminal(new AtanNode());
      else if(!strcmp(token,"asin")) 
	symbolSet.addNonterminal(new AsinNode());
      else if(!strcmp(token,"acos")) 
	symbolSet.addNonterminal(new AcosNode());
      else if(!strcmp(token,"tan")) 
	symbolSet.addNonterminal(new TanNode());
      else if(!strcmp(token,"^")) 
	symbolSet.addNonterminal(new PowNode());
      else if(!strcmp(token,"%")) 
	symbolSet.addNonterminal(new ModNode());
      else if(!strcmp(token,"&&")) 
	symbolSet.addNonterminal(new AndNode());
      else if(!strcmp(token,"||")) 
	symbolSet.addNonterminal(new OrNode());
      else if(!strcmp(token,"!=")) 
	symbolSet.addNonterminal(new NotEqualsNode());
      else if(!strcmp(token,">=")) 
	symbolSet.addNonterminal(new GreaterEqualsNode());
      else if(!strcmp(token,"<=")) 
	symbolSet.addNonterminal(new LessEqualsNode());
      else if(!strcmp(token,"!")) 
	symbolSet.addNonterminal(new NotNode());
      else if(!strcmp(token,"<")) 
	symbolSet.addNonterminal(new LessNode());
      else if(!strcmp(token,">")) 
	symbolSet.addNonterminal(new GreaterNode());
      else if(!strcmp(token,">0")) 
	symbolSet.addNonterminal(new GreaterThanZeroNode());
      else if(!strcmp(token,">1")) 
	symbolSet.addNonterminal(new GreaterThanOneNode());
      else if(!strcmp(token,"=0")) 
	symbolSet.addNonterminal(new EqualsZeroNode());
      else
	throw string("nonterminal symbol \"")+token+"\" is not recognized";
    }
  
  // Terminals
  const float minConst=Tableau::getGlobal().getMinConst();
  const float maxConst=Tableau::getGlobal().getMaxConst();
  symbols=Tableau::getGlobal().getTerminals();
  tokenizer.tokenizeThis(symbols.c_str());
  while(tokenizer.hasMoreTokens())
    {
      const char *token=tokenizer.nextToken();
      if(!strcmp(token,"const")) 
	symbolSet.addTerminal(new ConstantNode(minConst,maxConst));
      else if(!strcmp(token,"var")) 
	symbolSet.addTerminal(new VariableNode());
      else
	throw string("terminal symbol \"")+token+"\" is not recognized";
    }
}



void Application::generateUnobservables(BOOM::Array2D<float> &testPoints,
					int numPoints)
{
  for(int i=0 ; i<numPoints ; ++i)
    for(int j=0 ; j<numUnobservables ; ++j)
      testPoints[i][j]=RandomFloat(-1,1);  // ### maybe use Gaussian?
}



void Application::partition(int numCases,BOOM::Array1D<int> &a)
{
  float increment=numCases/(float)numCategories, first=0;
  for(int i=0 ; i<numCategories ; ++i)
    {
      a[i]=int(first);
      first+=increment;
    }
  a[numCategories]=numCases;
}



bool Application::isOK(BOOM::Array2D<float> &points,int begin,int end,
		       int varIndex)
{
  // Heuristic: scan the first 10 values and make sure they are all mostly
  // different
  BOOM::Set<float> counter;
  int n=points.getFirstDim();
  if(n>10) n=10;
  for(int i=0 ; i<n ; ++i)
    counter+=points[i][varIndex];
  int minDiversity=int(n-1);
  if(counter.size()<minDiversity)
    return false;

  // Don't allow INF or NaN
  for(int i=begin ; i<end ; ++i)
    if(!isFinite(points[i][varIndex])) 
      return false;

  // Make sure this variable isn't the same as any previous variable
  for(int i=numUnobservables ; i<varIndex ; ++i)
    {
      bool diff=false;
      for(int j=begin ; j<end ; ++j)
	if(points[j][i]!=points[j][varIndex])
	  {
	    diff=true;
	    break;
	  }
      if(!diff) return false;
    }

  // No problems found
  return true;
}


