/****************************************************************
 regress.C
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include <iostream>
#include <stdlib.h>
#include <fstream>
#include "BOOM/Exceptions.H"
#include "BOOM/Stacktrace.H"
#include "BOOM/InversionMatrix.H"
#include "BOOM/GaussJordan.H"
#include "BOOM/Array1D.H"
#include "BOOM/TriangMatrix.H"
#include "BOOM/String.H"
#include "BOOM/CommandLine.H"
#include "BOOM/MultRegress.H"
#include "BOOM/Correlation.H"
#include "../common/NamesFile.H"
#include "../common/TrainingSet.H"
#include "../common/Schema.H"
#include "Discriminator.H"

class Application
{
  const int order;
  MultRegress regression;
  String filestem;
  NamesFile namesFile;
  TrainingSet *trainingSet, *testSet;
  Schema *schema;
  TriangMatrix<Discriminator*> *discriminators;
  int numCategories;

  void square(TrainingSet *&);
  TrainingCase *square(TrainingCase &);
  int countCases(const String &filename);
  void train();
  Discriminator *train(int categoryA,int categoryB);
  void classify(TrainingSet *,int &TP,int &TN,int &FP,int &FN);
  int classify(TrainingCase &);
  void addToMatrices(TrainingSet &,int firstRow,DblMatrix &X,
		     DblMatrix &Y,int yValue);
  bool detectColinearity(DblMatrix &X);
public:
  Application(const String &filestem,int order);
  void go();
};


int main(int argc,char *argv[])
{
  try
    {
      CommandLine cmd(argc,argv,"o:");
      if(cmd.numArgs()!=1) throw "regress [-o order] <filestem>";
      String filestem=cmd.arg(0);
      int order=1;
      if(cmd.option('o')) order=cmd.optParm('o').asInt();
      Application app(filestem,order);
      app.go();
    }
  catch(RootException &e)
    {
      cerr << e.getMessage() << endl;
      return -1;
    }
  catch(const char *msg)
    {
      cerr << msg << endl;
      return -1;
    }
  catch(const String &s)
    {
      cerr << s.c_str() << endl; 
      return -1;
    }

  return 0;
}



Application::Application(const String &filestem,int order)
  : filestem(filestem), trainingSet(NULL), order(order)
{
  // ctor
}



void Application::go()
{
  if(order<1 || order>2) throw "Only orders 1 & 2 are currently supported";

  String namesFilename=filestem+".names";
  String trainFilename=filestem+".data";
  String testFilename=filestem+".test";

  schema=namesFile.processFile(namesFilename);
  numCategories=schema->getNumCategories();
  discriminators=new TriangMatrix<Discriminator*>(numCategories);

  trainingSet=new TrainingSet(*schema,trainFilename);
  testSet=new TrainingSet(*schema,testFilename);
  if(order==2) 
    {
      square(trainingSet);
      square(testSet);

      int numAttr=schema->numAttributes();
      numAttr=numAttr+numAttr*numAttr;
      cout<<"after squaring, num attributes is "<<numAttr<<endl;
      delete schema;
      schema=new Schema(numCategories);
      for(int i=0 ; i<numAttr ; ++i)
	schema->add(new Continuous(String("x")+i));
    }

  train();

  cout << "Accuracy on training set:  ";
  int TP=0, TN=0, FP=0, FN=0;
  classify(trainingSet,TP,TN,FP,FN);

  cout << "Accuracy on test set:  ";
  TP=TN=FP=FN=0;
  classify(testSet,TP,TN,FP,FN);
}



int Application::countCases(const String &filename)
{
  ifstream is(filename.c_str());
  if(!is.good())
    throw String("Cannot open file ")+filename;
  int n=0;
  String line;
  while(!is.eof())
    {
      getline(is,line,'\n');
      if(line=="") break;
      ++n;
    }

  return n;
}



void Application::classify(TrainingSet *testSet,
			   int &TP,int &TN,int &FP,int &FN)
{
  int numCorrect=0;

  float lastPercent=0.0;
  int i, n=testSet->numCases();
  for(i=0 ; i<n ; ++i)
    {
      float percentDone=100.0 * i / n;
      if(percentDone>=lastPercent+5)
	{
	  //cout << int(percentDone) << "% "; cout.flush();
	  lastPercent=percentDone;
	}
      TrainingCase &thisCase=testSet->getCase(i);
      int computed=classify(thisCase);
      int expected=thisCase.getCategory();
      if(computed==expected) ++numCorrect;
      switch(computed)
	{
	case 0:
	  switch(expected)
	    {
	    case 0: ++TN; break;
	    case 1: ++FN; break;
	    }
	  break;
	case 1:
	  switch(expected)
	    {
	    case 0: ++FP; break;
	    case 1: ++TP; break;
	    }
	  break;
	}
    }
  
  float accuracy=100.0 * numCorrect/(float) n;
  cout << accuracy << '%' << endl;
}



void Application::train()
{
  for(int i=0 ; i<numCategories ; ++i)
    for(int j=i+1 ; j<numCategories ; ++j)
      (*discriminators)(i,j)=train(i,j);
}



Discriminator *Application::train(int categoryA,int categoryB)
{
  /*
    This method performs multiple regression on all 
    training cases in categoryA or categoryB, producing
    a Discriminator which can (hopefully) discriminate
    just between those two categories.
   */

  // Get just those training cases belonging to category A or B
  TrainingSet *setA=trainingSet->getCategory(categoryA);
  TrainingSet *setB=trainingSet->getCategory(categoryB);
  const int numA=setA->numCases();
  const int numB=setB->numCases();
  const int numCases=numA+numB;

  // Allocate & initialize the matrices used for regression;
  DblMatrix X(numCases,schema->numAttributes());
  DblMatrix Y(numCases,1);
  DblMatrix A(schema->numAttributes()+1,1);
  addToMatrices(*setA,0,X,Y,0);
  addToMatrices(*setB,numA,X,Y,1);
  detectColinearity(X);

  // Perform regression
  if(!regression.regress(Y,X,A))
    {
      cout<<"ERROR: Cannot invert matrix.  Please remove dependent "
	  <<"variables"<<endl;
      throw "";
    }  

  // Extract a Discriminator
  Discriminator *discriminator=new Discriminator(A);
  cout<<"discriminator: "<<*discriminator<<endl;

  // Clean up 
  delete setA;
  delete setB;
  return discriminator;
}



void Application::addToMatrices(TrainingSet &cases,int nextRow,
				DblMatrix &X,DblMatrix &Y,
				int yValue)
{
  const int numAttr=schema->numAttributes();

  int numCases=cases.numCases();
  for(int i=0 ; i<numCases ; ++i)
    {
      // Get the next case
      TrainingCase &thisCase=cases.getCase(i);
      
      // Define row in Y
      Y(nextRow,0)=yValue;

      // Define row in X
      for(int j=0 ; j<numAttr ; ++j)
	X(nextRow,j)=thisCase.getContinuous(j);

      // Advance to next row
      ++nextRow;
    }
}



int Application::classify(TrainingCase &thisCase)
{
  IntArray1D counts(numCategories);
  counts.setAllTo(0);

  // Apply all discriminators
  for(int i=0 ; i<numCategories ; ++i)
    for(int j=i+1 ; j<numCategories ; ++j)
      {
	Discriminator &discriminator=*(*discriminators)(i,j);
	int d=discriminator(thisCase);
	++counts[d?j:i];
      }

  // See which category was most favored
  int maxIndex=0, maxCount=counts[0];
  for(int i=1 ; i<numCategories ; ++i)
    {
      int countsI=counts[i];
      if(countsI>maxCount)
	{
	  maxIndex=i;
	  maxCount=countsI;
	}
    }
  return maxIndex;
}



TrainingCase *Application::square(TrainingCase &c)
{
  int numAttr=c.getNumAttributes();
  TrainingCase *newCase=new TrainingCase(numAttr+numAttr*numAttr);
  for(int i=0, k=0 ; i<numAttr ; ++i)
    {
      double d=c.getContinuous(i);
      newCase->setContinuous(k,d);
      ++k;
      for(int j=0 ; j<numAttr ; ++j, ++k)
	newCase->setContinuous(k,d*c.getContinuous(j));
    }
  newCase->setCategory(c.getCategory());
  return newCase;
}



void Application::square(TrainingSet *&s)
{
  TrainingSet *newSet=new TrainingSet(s->getSchema());

  int numCases=s->numCases();
  for(int i=0 ; i<numCases ; ++i)
    {
      TrainingCase &c=s->getCase(i);
      TrainingCase *newCase=square(c);
      newSet->addCase(newCase);
    }

  s->deleteCases();
  delete s;
  s=newSet;
}



bool Application::detectColinearity(DblMatrix &X)
{
  int numCases=X.getFirstDim(), numAttributes=X.getSecondDim();
  Vector<double> col1, col2;
  for(int i=0 ; i<numAttributes-1 ; ++i)
    {
      X.getColumn(i,col1);
      for(int j=i+1 ; j<numAttributes ; ++j)
	{
	  X.getColumn(j,col2);
	  Correlation c(col1,col2);
	  double r=c.getR();
	  if(r>=0.8)
	    cout<<"WARNING! "<<namesFile.getName(i)<<" and "
		<<namesFile.getName(j)<<" are highly correlated (r="
		<<r<<")"<<endl;
	}
    }
}



